Package: tm
Title: Text Mining Package
Version: 0.7-18
Date: 2026-02-18
Authors@R: c(person("Ingo", "Feinerer", role = "aut",
                    email = "feinerer@logic.at",
                    comment = c(ORCID = "0000-0001-7656-8338")),
             person("Kurt", "Hornik", role = c("aut", "cre"),
     		    email = "Kurt.Hornik@R-project.org",
                    comment = c(ORCID = "0000-0003-4198-9911")),
             person("Artifex Software, Inc.", role = c("ctb", "cph"),
                    comment = "pdf_info.ps taken from GPL Ghostscript"))
Depends: R (>= 3.4.0), NLP (>= 0.2-0)
Imports: Rcpp, parallel, slam (>= 0.1-37), stats, tools, utils,
        graphics, xml2
LinkingTo: BH, Rcpp
Suggests: antiword, filehash, methods, pdftools, Rcampdf, Rgraphviz,
        Rpoppler, SnowballC, testthat, tm.lexicon.GeneralInquirer
Description: A framework for text mining applications within R.
License: GPL-3
URL: https://tm.r-forge.r-project.org/
Additional_repositories: https://datacube.wu.ac.at
NeedsCompilation: yes
Packaged: 2026-02-18 10:06:28 UTC; hornik
Author: Ingo Feinerer [aut] (ORCID: <https://orcid.org/0000-0001-7656-8338>),
  Kurt Hornik [aut, cre] (ORCID: <https://orcid.org/0000-0003-4198-9911>),
  Artifex Software, Inc. [ctb, cph] (pdf_info.ps taken from GPL
    Ghostscript)
Maintainer: Kurt Hornik <Kurt.Hornik@R-project.org>
Repository: CRAN
Date/Publication: 2026-02-18 11:15:00 UTC

useDynLib("tm", .registration = TRUE)

importFrom("NLP", "content", "content<-", "meta", "meta<-", "words",
           "as.Token_Tokenizer", "is.Span_Tokenizer", "Token_Tokenizer",
           "TaggedTextDocument")
importFrom("Rcpp", "evalCpp")
importFrom("graphics", "abline", "plot")
importFrom("parallel", "parLapply")
importFrom("stats", "coef", "cor", "lm", "setNames")
importFrom("utils", "download.file", "getS3method", "unzip")
importFrom("slam", "as.simple_triplet_matrix", "col_sums",
           "crossapply_simple_triplet_matrix", "read_stm_MC", "rollup",
           "row_sums", "simple_triplet_matrix")
importFrom("xml2",
           "read_xml",
           "xml_contents",
           "xml_find_all",
           "xml_missing",
           "xml_text")

export("as.DocumentTermMatrix",
       "as.TermDocumentMatrix",
       "as.VCorpus",
       "Boost_tokenizer",
       "content_transformer",
       "Corpus",
       "DataframeSource",
       "DirSource",
       "Docs",
       "DocumentTermMatrix",
       "DublinCore",
       "DublinCore<-",
       "eoi",
       "FunctionGenerator",
       "getElem",
       "getMeta",
       "Heaps_plot",
       "findAssocs",
       "findFreqTerms",
       "findMostFreqTerms",
       "getReaders",
       "getSources",
       "getTokenizers",
       "getTransformations",
       "inspect",
       "MC_tokenizer",
       "nDocs",
       "nTerms",
       "PCorpus",
       "pGetElem",
       "PlainTextDocument",
       "read_dtm_Blei_et_al",
       "read_dtm_MC",
       "readDataframe",
       "readDOC",
       "reader",
       "readPlain",
       "readReut21578XML",
       "readReut21578XMLasPlain",
       "readRCV1",
       "readRCV1asPlain",
       "readPDF",
       "readTagged",
       "readXML",
       "removeNumbers",
       "removePunctuation",
       "removeSparseTerms",
       "removeWords",
       "scan_tokenizer",
       "SimpleCorpus",
       "SimpleSource",
       "stemCompletion",
       "stemDocument",
       "stepNext",
       "stopwords",
       "stripWhitespace",
       "TermDocumentMatrix",
       "termFreq",
       "Terms",
       "tm_filter",
       "tm_index",
       "tm_map",
       "tm_parLapply",
       "tm_parLapply_engine",
       "tm_reduce",
       "tm_term_score",
       "URISource",
       "VCorpus",
       "VectorSource",
       "WeightFunction",
       "weightTf",
       "weightTfIdf",
       "weightBin",
       "weightSMART",
       "writeCorpus",
       "XMLSource",
       "XMLTextDocument",
       "Zipf_plot",
       "ZipSource")

S3method("removeNumbers", "character")
S3method("removePunctuation", "character")
S3method("removeWords", "character")
S3method("stemDocument", "character")
S3method("stripWhitespace", "character")
S3method("words", "character")

S3method("[", "DataframeSource")
S3method("[[", "DataframeSource")
S3method("getElem", "DataframeSource")
S3method("getMeta", "DataframeSource")
S3method("pGetElem", "DataframeSource")

S3method("[", "DirSource")
S3method("[[", "DirSource")
S3method("getElem", "DirSource")
S3method("pGetElem", "DirSource")

S3method("[", "DocumentTermMatrix")
S3method("c", "DocumentTermMatrix")
S3method("dimnames<-", "DocumentTermMatrix")
S3method("findAssocs", "DocumentTermMatrix")
S3method("findMostFreqTerms", "DocumentTermMatrix")
S3method("inspect", "DocumentTermMatrix")
S3method("plot", "DocumentTermMatrix")
S3method("print", "DocumentTermMatrix")
S3method("t", "DocumentTermMatrix")
S3method("tm_term_score", "DocumentTermMatrix")

S3method("as.VCorpus", "list")

S3method("tm_term_score", "term_frequency")

S3method("[", "PCorpus")
S3method("[[", "PCorpus")
S3method("[[<-", "PCorpus")
S3method("as.list", "PCorpus")
S3method("content", "PCorpus")
S3method("format", "PCorpus")
S3method("inspect", "PCorpus")
S3method("length", "PCorpus")
S3method("meta", "PCorpus")
S3method("meta<-", "PCorpus")
S3method("names", "PCorpus")
S3method("names<-", "PCorpus")
S3method("print", "PCorpus", .print_via_format)
S3method("TermDocumentMatrix", "PCorpus")
S3method("tm_filter", "PCorpus")
S3method("tm_index", "PCorpus")
S3method("tm_map", "PCorpus")

S3method("as.character", "PlainTextDocument")
S3method("content", "PlainTextDocument")
S3method("content<-", "PlainTextDocument")
S3method("format", "PlainTextDocument")
S3method("meta", "PlainTextDocument")
S3method("meta<-", "PlainTextDocument")
S3method("print", "PlainTextDocument", .print_via_format)
S3method("removeNumbers", "PlainTextDocument")
S3method("removePunctuation", "PlainTextDocument")
S3method("removeWords", "PlainTextDocument")
S3method("stemDocument", "PlainTextDocument")
S3method("stripWhitespace", "PlainTextDocument")
S3method("tm_term_score", "PlainTextDocument")
S3method("words", "PlainTextDocument")

S3method("[", "SimpleCorpus")
S3method("[[", "SimpleCorpus")
S3method("[[<-", "SimpleCorpus")
S3method("as.list", "SimpleCorpus")
S3method("content", "SimpleCorpus")
S3method("format", "SimpleCorpus")
S3method("inspect", "SimpleCorpus")
S3method("length", "SimpleCorpus")
S3method("meta", "SimpleCorpus")
S3method("meta<-", "SimpleCorpus")
S3method("names", "SimpleCorpus")
S3method("print", "SimpleCorpus", .print_via_format)
S3method("TermDocumentMatrix", "SimpleCorpus")
S3method("tm_filter", "SimpleCorpus")
S3method("tm_index", "SimpleCorpus")
S3method("tm_map", "SimpleCorpus")

S3method("close", "SimpleSource")
S3method("eoi", "SimpleSource")
S3method("length", "SimpleSource")
S3method("open", "SimpleSource")
S3method("reader", "SimpleSource")
S3method("stepNext", "SimpleSource")

S3method("c", "TermDocumentMatrix")
S3method("[", "TermDocumentMatrix")
S3method("dimnames<-", "TermDocumentMatrix")
S3method("findAssocs", "TermDocumentMatrix")
S3method("findMostFreqTerms", "TermDocumentMatrix")
S3method("inspect", "TermDocumentMatrix")
S3method("plot", "TermDocumentMatrix")
S3method("print", "TermDocumentMatrix")
S3method("t", "TermDocumentMatrix")
S3method("tm_term_score", "TermDocumentMatrix")

S3method("c", "term_frequency")
S3method("findMostFreqTerms", "term_frequency")

S3method("c", "TextDocument")
S3method("inspect", "TextDocument")

S3method("print", "TextDocumentMeta")

S3method("[", "URISource")
S3method("[[", "URISource")
S3method("getElem", "URISource")
S3method("pGetElem", "URISource")

S3method("[", "VCorpus")
S3method("[[", "VCorpus")
S3method("[[<-", "VCorpus")
S3method("as.list", "VCorpus")
S3method("as.VCorpus", "VCorpus")
S3method("c", "VCorpus")
S3method("content", "VCorpus")
S3method("format", "VCorpus")
S3method("inspect", "VCorpus")
S3method("length", "VCorpus")
S3method("meta", "VCorpus")
S3method("meta<-", "VCorpus")
S3method("names", "VCorpus")
S3method("names<-", "VCorpus")
S3method("print", "VCorpus", .print_via_format)
S3method("TermDocumentMatrix", "VCorpus")
S3method("tm_filter", "VCorpus")
S3method("tm_index", "VCorpus")
S3method("tm_map", "VCorpus")

S3method("[", "VectorSource")
S3method("[[", "VectorSource")
S3method("getElem", "VectorSource")
S3method("pGetElem", "VectorSource")

S3method("getElem", "XMLSource")

S3method("as.character", "XMLTextDocument")
S3method("content", "XMLTextDocument")
S3method("content<-", "XMLTextDocument")
S3method("format", "XMLTextDocument")
S3method("meta", "XMLTextDocument")
S3method("meta<-", "XMLTextDocument")
S3method("print", "XMLTextDocument", .print_via_format)

S3method("close", "ZipSource")
S3method("getElem", "ZipSource")
S3method("open", "ZipSource")
S3method("pGetElem", "ZipSource")

S3method("TermDocumentMatrix", "default")

S3method("as.DocumentTermMatrix", "DocumentTermMatrix")
S3method("as.DocumentTermMatrix", "TermDocumentMatrix")
S3method("as.DocumentTermMatrix", "default")
S3method("as.DocumentTermMatrix", "term_frequency")
S3method("as.DocumentTermMatrix", "textcnt")
S3method("as.TermDocumentMatrix", "TermDocumentMatrix")
S3method("as.TermDocumentMatrix", "DocumentTermMatrix")
S3method("as.TermDocumentMatrix", "default")
S3method("as.TermDocumentMatrix", "term_frequency")
S3method("as.TermDocumentMatrix", "textcnt")

S3method("Docs", "DocumentTermMatrix")
S3method("Docs", "TermDocumentMatrix")
S3method("Terms", "DocumentTermMatrix")
S3method("Terms", "TermDocumentMatrix")
S3method("nDocs", "DocumentTermMatrix")
S3method("nDocs", "TermDocumentMatrix")
S3method("nTerms", "DocumentTermMatrix")
S3method("nTerms", "TermDocumentMatrix")

S3method("findAssocs", "matrix")

What is it?
-----------
This base package contains Perl extensions for Topic Maps

   http://www.isotopicmaps.org/

With it you can

  - manipulate topic maps via an API
  - navigate through TMs via a TMDM-like API (almost TMAPI)
  - read/write AsTMa 1.x and 2.x resources
  - read LTM resources
  - read/write XTM 1/2 resources
  - create virtualized topic maps
  - use a topic map backend store
  - add drivers for your own native store
  - add indices to speed up things
  - query topic maps via TMQL (query language, in the working)
  - and more

Prerequisites
-------------

Of course, TM::* relies heavily on other Perl packages. Most of them
(hopefully all) should be resolved automatically if you use the CPAN
frontend.

Installation
------------

This module should be completely installable via the cpan frontend.
To build the extensions manually, unpack this distribution somewhere,
and exercise the usual incantation:

perl Makefile.PL
make
make test
make install

You will need perl version 5.8 or better to install these modules.

There may also be Debian packages for this generation.

Documentation
-------------

You can learn about Topic Maps in general at

  http://www.isotopicmaps.org/  (standards)
  http://kill.devc.at/taxonomy/term/22 (my blog)

  http://topicmaps.bond.edu.au/ (tutorials, other stuff)
  http://easytopicmaps.com/     (tutorials, lotta links and hints)

  http://astma.it.bond.edu.au/  (shorthand notation used by the authors)

For this particular package I also recommend the
tutorials in

  maps/tutorials.atm

Support
-------

You can either

  - mail to the author (rho@devc.at), or
  - discuss this in one of the topicmap mailing lists, or
  - find me in the freenode #topicmaps IRC channel, or
  - create a bug entry in rt.cpan.org (preferred)

Copyright 200[1-8] Robert Barta, rho@devc.at

This library is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

bibentry(bibtype = "InProceedings",
         key = "Feinerer:2010",
         author = person(given = "Ingo",
                         family = "Feinerer"),
         editor = c(person(given = "Pu-Jen",
                           family = "Cheng"),
                    person(given = "Min-Yen",
                           family = "Kan"),
                    person(given = "Wai",
                           family = "Lam"),
                    person(given = "Preslav",
                           family = "Nakov")),
         title = "Analysis and Algorithms for Stemming Inversion",
         booktitle = "Information Retrieval Technology",
         year = "2010",
         publisher = "Springer Berlin Heidelberg",
         address = "Berlin, Heidelberg",
         pages = "290--299",
         isbn = "978-3-642-17187-1",
         doi = "10.1007/978-3-642-17187-1_28")

bibentry(bibtype = "Misc",
         key = "Lewis:1987",
         doi = "10.24432/C52G6M",
         author = person(given = "David",
                         family = "Lewis"),
         title = "Reuters-21578 Text Categorization Collection",
         publisher = "UCI Machine Learning Repository",
         year = "1987")

bibentry(bibtype = "Article",
         key = "Lewis+Yang+Rose:2004",
         author = c(person(given = c("David", "D."),
                           family = "Lewis"),
                    person(given = "Yiming",
                           family = "Yang"),
                    person(given = c("Tony", "G."),
                           family = "Rose"),
                    person(given = "Fan",
                           family = "Li")),
         title = "{RCV1}: A New Benchmark Collection for Text Categorization Research",
         journal = "Journal of Machine Learning Research",
         year = "2004",
         volume = "5",
         pages = "361--397",
         url = "https://www.jmlr.org/papers/volume5/lewis04a/lewis04a.pdf")

bibentry(bibtype = "Book",
         key = "Manning+Raghavan+Schütze:2008",
         author = c(person(given = c("Christopher", "D."),
                           family = "Manning"),
                    person(given = "Prabhakar",
                           family = "Raghavan"),
                    person(given = "Hinrich",
                           family = "Schütze")),
         title = "Introduction to Information Retrieval",
         publisher = "Cambridge University Press",
         year = "2008",
         doi = "10.1017/CBO9780511809071")

bibentry(bibtype = "Article",
         key = "Salton+Buckley:1988",
         author = c(person(given = "Gerard",
                           family = "Salton"),
                    person(given = "Christopher",
                           family = "Buckley")),
         title = "Term-weighting Approaches in Automatic Text Retrieval",
         journal = "Information Processing and Management",
         year = "1988",
         volume = "24",
         number = "5",
         pages = "513--523",
         doi = "10.1016/0306-4573(88)90021-0")

## Authors: Ingo Feinerer, Kurt Hornik

TermDocumentMatrix_classes <-
    c("TermDocumentMatrix", "simple_triplet_matrix")
DocumentTermMatrix_classes <-
    c("DocumentTermMatrix", "simple_triplet_matrix")

.TermDocumentMatrix <-
function(x, weighting)
{
    x <- as.simple_triplet_matrix(x)
    if (!is.null(dimnames(x)))
        names(dimnames(x)) <- c("Terms", "Docs")
    class(x) <- TermDocumentMatrix_classes

    if (is.null(weighting))
        weighting <- weightTf
    ## <NOTE>
    ## Note that if weighting is a weight function, it already needs to
    ## know whether we have a term-document or document-term matrix.
    ##
    ## Ideally we would require weighting to be a WeightFunction object
    ## or a character string of length 2.  But then
    ##   dtm <- DocumentTermMatrix(crude,
    ##                             control = list(weighting =
    ##                                            function(x)
    ##                                            weightTfIdf(x, normalize =
    ##                                                        FALSE),
    ##                                            stopwords = TRUE))
    ## in example("DocumentTermMatrix") fails [because weightTfIdf() is
    ## a weight function and not a weight function generator ...]
    ## Hence, for now, instead of
    ##   if (inherits(weighting, "WeightFunction"))
    ##      x <- weighting(x)
    ## use
    if (is.function(weighting))
        x <- weighting(x)
    ## and hope for the best ...
    ## </NOTE>
    else if (is.character(weighting) && (length(weighting) == 2L))
        attr(x, "weighting") <- weighting
    x
}

.SimpleTripletMatrix <-
function(i, j, v, terms, corpus)
{
    docs <- as.character(meta(corpus, "id", "local"))
    if (length(docs) != length(corpus)) {
        warning("invalid document identifiers")
        docs <- NULL
    }

    simple_triplet_matrix(i, j, v,
                          nrow = length(terms),
                          ncol = length(corpus),
                          dimnames = list(Terms = terms, Docs = docs))
}

filter_global_bounds <-
function(m, bounds)
{
    m <- as.simple_triplet_matrix(m)

    if (length(bounds) == 2L && is.numeric(bounds)) {
        rs <- row_sums(m > 0)
        m <- m[(rs >= bounds[1]) & (rs <= bounds[2]), ]
    }
    m
}

TermDocumentMatrix <-
function(x, control = list())
    UseMethod("TermDocumentMatrix", x)

TermDocumentMatrix.SimpleCorpus <-
function(x, control = list())
{
    stopifnot(is.list(control))

    if (any(unlist(lapply(control, is.function))))
        warning("custom functions are ignored")

    if (!is.null(control$tokenize) && !identical(control$tokenize, "Boost"))
        warning("custom tokenizer is ignored")

    txt <- content(x)

    ## Conversion to lower case
    if (is.null(control$tolower) || isTRUE(control$tolower))
        txt <- tolower(txt)

    ## Stopword filtering
    .stopwords <- if (isTRUE(control$stopwords)) stopwords(meta(x, "language"))
        else if (is.character(control$stopwords)) control$stopwords
        else character(0)

    .dictionary <- if (is.null(control$dictionary)) character(0)
        else control$dictionary

    ## Ensure local bounds
    bl <- control$bounds$local
    min_term_freq <-
        if (length(bl) == 2L && is.numeric(bl) && bl[1] >= 0) bl[1] else 0L
    max_term_freq <-
        if (length(bl) == 2L && is.numeric(bl) && bl[2] >= 0)
            min(bl[2], .Machine$integer.max) else .Machine$integer.max

    ## Filter out too short or too long terms
    wl <- control$wordLengths
    min_word_length <- if (is.numeric(wl[1]) && wl[1] >= 0) wl[1] else 3L
    max_word_length <- if (is.numeric(wl[2]) && wl[2] >= 0)
        min(wl[2], .Machine$integer.max) else .Machine$integer.max

    m <- tdm(txt,
             isTRUE(control$removePunctuation),
             isTRUE(control$removeNumbers),
             .stopwords, .dictionary,
             as.integer(min_term_freq), as.integer(max_term_freq),
             as.integer(min_word_length), as.integer(max_word_length))

    Encoding(m$terms) <- "UTF-8"
    m <- .SimpleTripletMatrix(m$i, m$j, m$v, m$terms, x)

    ## Stemming
    ## <NOTE>
    ## Ideally tdm() could perform stemming as well but there is no easy way to
    ## access the SnowballC::wordStem() function from C++ (via Rcpp) without
    ## significant overhead (as SnowballC does not export its internal C
    ## functions).
    ##
    ## Stemming afterwards is still quite performant as we already have
    ## all terms.  However, there is some overhead involved as we need
    ## to recheck local bounds and word lengths.
    ## </NOTE>
    if (isTRUE(control$stemming)) {
        stems <- SnowballC::wordStem(m$dimnames$Terms,
                                     meta(x, "language"))
        ## Do as.factor(stems) "by hand" for performance reasons.
        uniqs <- sort(unique(stems))
        stems <- match(stems, uniqs)
        attributes(stems) <- list(levels = uniqs, class = "factor")
        m <- rollup(m, "Terms", stems)

        ## Recheck local bounds
        ## No need to check lower local bound as rollup aggregates frequencies
        m[m > max_term_freq] <- 0

        ## Recheck word lengths
        terms_length <- nchar(rownames(m))
        m <- m[min_word_length <= terms_length &
               terms_length <= max_word_length, ]
    }

    m <- filter_global_bounds(m, control$bounds$global)

    .TermDocumentMatrix(m, control$weighting)
}

TermDocumentMatrix.PCorpus <-
TermDocumentMatrix.VCorpus <-
function(x, control = list())
{
    stopifnot(is.list(control))

    tflist <- tm_parLapply(unname(content(x)), termFreq, control)

    v <- unlist(tflist)
    i <- names(v)
    terms <- sort(unique(as.character(if (is.null(control$dictionary)) i
                                      else control$dictionary)))
    i <- match(i, terms)
    j <- rep.int(seq_along(x), lengths(tflist))

    m <- .SimpleTripletMatrix(i, j, as.numeric(v), terms, x)
    m <- filter_global_bounds(m, control$bounds$global)
    .TermDocumentMatrix(m, control$weighting)
}

TermDocumentMatrix.default <-
function(x, control = list())
    TermDocumentMatrix(Corpus(VectorSource(x)), control)

DocumentTermMatrix <-
function(x, control = list())
    t(TermDocumentMatrix(x, control))

as.TermDocumentMatrix <-
function(x, ...)
    UseMethod("as.TermDocumentMatrix")
as.TermDocumentMatrix.TermDocumentMatrix <-
function(x, ...)
    x
as.TermDocumentMatrix.DocumentTermMatrix <-
function(x, ...)
    t(x)
as.TermDocumentMatrix.term_frequency <-
as.TermDocumentMatrix.textcnt <-
function(x, ...)
{
    m <- simple_triplet_matrix(i = seq_along(x),
                               j = rep_len(1L, length(x)),
                               v = as.numeric(x),
                               nrow = length(x),
                               ncol = 1,
                               dimnames =
                               list(Terms = names(x),
                                    Docs = NA_character_))

    .TermDocumentMatrix(m, weightTf)
}
as.TermDocumentMatrix.default <-
function(x, weighting, ...)
    .TermDocumentMatrix(x, weighting)

as.DocumentTermMatrix <-
function(x, ...)
    UseMethod("as.DocumentTermMatrix")
as.DocumentTermMatrix.DocumentTermMatrix <-
function(x, ...)
    x
as.DocumentTermMatrix.TermDocumentMatrix <-
function(x, ...)
    t(x)
as.DocumentTermMatrix.term_frequency <-
as.DocumentTermMatrix.textcnt <-
function(x, ...)
    t(as.TermDocumentMatrix(x))
as.DocumentTermMatrix.default <-
function(x, weighting, ...)
{
    x <- as.simple_triplet_matrix(x)
    t(.TermDocumentMatrix(t(x), weighting))
}

t.TermDocumentMatrix <-
t.DocumentTermMatrix <-
function(x)
{
    m <- NextMethod("t")
    attr(m, "weighting") <- attr(x, "weighting")
    class(m) <- if (inherits(x, "DocumentTermMatrix"))
        TermDocumentMatrix_classes
    else
        DocumentTermMatrix_classes
    m
}

termFreq <-
function(doc, control = list())
{
    stopifnot(inherits(doc, "TextDocument") || is.character(doc),
              is.list(control))

    ## Tokenize the corpus
    .tokenize <- control$tokenize
    if (is.null(.tokenize) || identical(.tokenize, "words"))
        .tokenize <- words
    else if (identical(.tokenize, "Boost"))
        .tokenize <- Boost_tokenizer
    else if (identical(.tokenize, "MC"))
        .tokenize <- MC_tokenizer
    else if (identical(.tokenize, "scan"))
        .tokenize <- scan_tokenizer
    else if (is.Span_Tokenizer(.tokenize))
        .tokenize <- as.Token_Tokenizer(.tokenize)
    if (is.function(.tokenize))
        txt <- .tokenize(doc)
    else
        stop("invalid tokenizer")

    ## Conversion to lower case
    .tolower <- control$tolower
    if (is.null(.tolower) || isTRUE(.tolower))
        .tolower <- tolower
    if (is.function(.tolower))
        txt <- .tolower(txt)

    ## Punctuation removal
    .removePunctuation <- control$removePunctuation
    if (isTRUE(.removePunctuation))
        .removePunctuation <- removePunctuation
    else if (is.list(.removePunctuation))
        .removePunctuation <-
            function(x) do.call(removePunctuation,
                                c(list(x), control$removePunctuation))

    ## Number removal
    .removeNumbers <- control$removeNumbers
    if (isTRUE(.removeNumbers))
        .removeNumbers <- removeNumbers

    .language <- control$language
    if (inherits(doc, "TextDocument"))
        .language <- meta(doc, "language")
    if (is.null(.language))
        .language <- "en"

    ## Stopword filtering
    .stopwords <- control$stopwords
    if (isTRUE(.stopwords))
        .stopwords <- function(x) x[is.na(match(x, stopwords(.language)))]
    else if (is.character(.stopwords))
        .stopwords <- function(x) x[is.na(match(x, control$stopwords))]

    ## Stemming
    .stemming <- control$stemming
    if (isTRUE(.stemming))
        .stemming <- function(x) SnowballC::wordStem(x, .language)

    ## Default order for options which support reordering
    or <- c("removePunctuation", "removeNumbers", "stopwords", "stemming")

    ## Process control options in specified order
    nc <- names(control)
    n <- nc[!is.na(match(nc, or))]
    for (name in sprintf(".%s", c(n, setdiff(or, n)))) {
        g <- get(name)
        if (is.function(g))
            txt <- g(txt)
    }

    ## If dictionary is set tabulate against it
    dictionary <- control$dictionary
    tab <- .table(if (is.null(dictionary))
                      txt
                  else
                      txt[!is.na(match(txt, dictionary))])

    ## Ensure local bounds
    bl <- control$bounds$local
    if (length(bl) == 2L && is.numeric(bl))
        tab <- tab[(tab >= bl[1]) & (tab <= bl[2]), drop = FALSE]

    ## Filter out too short or too long terms
    nc <- nchar(names(tab), type = "chars")
    wl <- control$wordLengths
    lb <- if (is.numeric(wl[1])) wl[1] else 3
    ub <- if (is.numeric(wl[2])) wl[2] else Inf
    tab <- tab[(nc >= lb) & (nc <= ub), drop = FALSE]

    class(tab) <- c("term_frequency", class(tab))
    tab
}

print.TermDocumentMatrix <-
print.DocumentTermMatrix <-
function(x, ...)
{
    format <- c("term", "document")
    if (inherits(x, "DocumentTermMatrix"))
        format <- rev(format)
    writeLines(sprintf("<<%s (%ss: %d, %ss: %d)>>",
                       class(x)[1], format[1L], nrow(x), format[2L], ncol(x)))
    writeLines(sprintf("Non-/sparse entries: %d/%.0f",
                       length(x$v), prod(dim(x)) - length(x$v)))
    sparsity <- if (!prod(dim(x))) 100
        else round( (1 - length(x$v) / prod(dim(x))) * 100)
    writeLines(sprintf("Sparsity           : %s%%", sparsity))
    writeLines(sprintf("Maximal term length: %s",
                       max(nchar(Terms(x), type = "chars"), 0)))
    writeLines(sprintf("Weighting          : %s (%s)",
                       attr(x, "weighting")[1L], attr(x, "weighting")[2L]))
    invisible(x)
}

inspect.TermDocumentMatrix <-
inspect.DocumentTermMatrix <-
function(x)
{
    print(x)
    cat("Sample             :\n")
    print(as.matrix(sample.TermDocumentMatrix(x)))
}

`[.TermDocumentMatrix` <-
`[.DocumentTermMatrix` <-
function(x, i, j, ..., drop)
{
    m <- NextMethod("[")
    attr(m, "weighting") <- attr(x, "weighting")
    class(m) <- if (inherits(x, "DocumentTermMatrix"))
        DocumentTermMatrix_classes
    else
        TermDocumentMatrix_classes
    m
}

`dimnames<-.DocumentTermMatrix` <-
function(x, value)
{
    x <- NextMethod("dimnames<-")
    dnx <- x$dimnames
    if (!is.null(dnx))
        names(dnx) <- c("Docs", "Terms")
    x$dimnames <- dnx
    x
}

`dimnames<-.TermDocumentMatrix` <-
function(x, value)
{
    x <- NextMethod("dimnames<-")
    dnx <- x$dimnames
    if (!is.null(dnx))
        names(dnx) <- c("Terms", "Docs")
    x$dimnames <- dnx
    x
}

nDocs <-
function(x)
    UseMethod("nDocs")

nTerms <-
function(x)
    UseMethod("nTerms")

nDocs.DocumentTermMatrix <-
nTerms.TermDocumentMatrix <-
function(x)
    x$nrow

nDocs.TermDocumentMatrix <-
nTerms.DocumentTermMatrix <-
function(x)
    x$ncol

Docs <-
function(x)
    UseMethod("Docs")

Terms <-
function(x)
    UseMethod("Terms")

Docs.DocumentTermMatrix <-
Terms.TermDocumentMatrix <-
function(x)
{
    s <- x$dimnames[[1L]]
    if (is.null(s))
        s <- rep.int(NA_character_, x$nrow)
    s
}

Docs.TermDocumentMatrix <-
Terms.DocumentTermMatrix <-
function(x)
{
    s <- x$dimnames[[2L]]
    if (is.null(s))
        s <- rep.int(NA_character_, x$ncol)
    s
}

c.term_frequency <-
function(..., recursive = FALSE)
{
    do.call("c", lapply(list(...), as.TermDocumentMatrix))
}

c.TermDocumentMatrix <-
function(..., recursive = FALSE)
{
    m <- lapply(list(...), as.TermDocumentMatrix)

    if (length(m) == 1L)
        return(m[[1L]])

    weighting <- attr(m[[1L]], "weighting")

    allTermsNonUnique <- unlist(lapply(m, function(x) Terms(x)[x$i]))
    allTerms <- unique(allTermsNonUnique)
    allDocs <- unlist(lapply(m, Docs))

    cs <- cumsum(lapply(m, nDocs))
    cs <- c(0, cs[-length(cs)])
    j <- lapply(m, "[[", "j")

    m <- simple_triplet_matrix(i = match(allTermsNonUnique, allTerms),
                               j = unlist(j) + rep.int(cs, lengths(j)),
                               v = unlist(lapply(m, "[[", "v")),
                               nrow = length(allTerms),
                               ncol = length(allDocs),
                               dimnames =
                               list(Terms = allTerms,
                                    Docs = allDocs))
    ## <NOTE>
    ## - We assume that all arguments have the same weighting
    ## - Even if all matrices have the same input weighting it might be
    ##   necessary to take additional steps (e.g., normalization for tf-idf or
    ##   check for (0,1)-range for binary tf)
    ## </NOTE>
    .TermDocumentMatrix(m, weighting)
}

c.DocumentTermMatrix <-
function(..., recursive = FALSE)
{
    t(do.call("c", lapply(list(...), as.TermDocumentMatrix)))
}

findFreqTerms <-
function(x, lowfreq = 0, highfreq = Inf)
{
    stopifnot(inherits(x, c("DocumentTermMatrix", "TermDocumentMatrix")),
              is.numeric(lowfreq), is.numeric(highfreq))

    if (inherits(x, "DocumentTermMatrix")) x <- t(x)
    rs <- row_sums(x)
    names(rs[rs >= lowfreq & rs <= highfreq])
}

findAssocs <-
function(x, terms, corlimit)
    UseMethod("findAssocs", x)
findAssocs.TermDocumentMatrix <-
function(x, terms, corlimit)
    findAssocs(t(x), terms, corlimit)
findAssocs.DocumentTermMatrix <-
function(x, terms, corlimit)
{
    stopifnot(is.character(terms), is.numeric(corlimit),
              corlimit >= 0, corlimit <= 1)

    j <- match(unique(terms), Terms(x), nomatch = 0L)
    suppressWarnings(
        findAssocs(crossapply_simple_triplet_matrix(x[, j], x[, -j], cor),
                   terms, rep_len(corlimit, length(terms))))
}
findAssocs.matrix <-
function(x, terms, corlimit)
{
    stopifnot(is.numeric(x))

    i <- match(terms, rownames(x), nomatch = 0L)
    names(i) <- terms
    Map(function(i, cl) {
        xi <- x[i, ]
        t <- sort(round(xi[which(xi >= cl)], 2), TRUE)
        if (!length(t))
            names(t) <- NULL
        t
        },
        i, corlimit)
}

removeSparseTerms <-
function(x, sparse)
{
    stopifnot(inherits(x, c("DocumentTermMatrix", "TermDocumentMatrix")),
              is.numeric(sparse), sparse > 0, sparse < 1)

    m <- if (inherits(x, "DocumentTermMatrix")) t(x) else x
    t <- table(m$i) > m$ncol * (1 - sparse)
    termIndex <- as.numeric(names(t[t]))
    if (inherits(x, "DocumentTermMatrix")) x[, termIndex] else x[termIndex, ]
}

sample.TermDocumentMatrix <-
function(x, size = 10)
{
    stopifnot(inherits(x, c("DocumentTermMatrix", "TermDocumentMatrix")),
              is.numeric(size), size >= 0)

    if (length(x$v) == 0L)
        return(x)

    m <- if (inherits(x, "DocumentTermMatrix")) t(x) else x
    terms <- sort(names(sort(row_sums(m), decreasing = TRUE)
                        [0:min(size, nTerms(m))]))
    docs <- sort(names(sort(col_sums(m), decreasing = TRUE)
                       [0:min(size, nDocs(m))]))
    if (inherits(x, "DocumentTermMatrix")) x[docs, terms] else x[terms, docs]
}

CategorizedDocumentTermMatrix <-
function(x, c)
{
    if (inherits(x, "TermDocumentMatrix"))
        x <- t(x)
    else if (!inherits(x, "DocumentTermMatrix"))
        stop("wrong class")

    if (length(c) != nDocs(x))
        stop("invalid category ids")

    attr(x, "Category") <- c

    class(x) <- c("CategorizedDocumentTermMatrix",
                  DocumentTermMatrix_classes)

    x
}

findMostFreqTerms <-
function(x, n = 6L, ...)
    UseMethod("findMostFreqTerms")

findMostFreqTerms.term_frequency <-
function(x, n = 6L, ...)
{
    y <- x[order(x, decreasing = TRUE)[seq_len(n)]]
    y[y > 0]
}

findMostFreqTerms.DocumentTermMatrix <-
function(x, n = 6L, INDEX = NULL, ...)
{
    terms <- Terms(x)
    if (!is.null(INDEX))
        x <- rollup(x, 1L, INDEX)
    f <- factor(x$i, seq_len(x$nrow))
    js <- split(x$j, f)
    vs <- split(x$v, f)
    y <- Map(function(j, v, n) {
                 p <- order(v, decreasing = TRUE)[seq_len(n)]
                 v <- v[p]
                 names(v) <- terms[j[p]]
                 v
             },
             js, vs, pmin(lengths(vs), n))
    names(y) <- x$dimnames[[1L]]
    y
}

findMostFreqTerms.TermDocumentMatrix <-
function(x, n = 6L, INDEX = NULL, ...)
{
    terms <- Terms(x)
    if (!is.null(INDEX))
        x <- rollup(x, 2L, INDEX)
    f <- factor(x$j, seq_len(x$ncol))
    is <- split(x$i, f)
    vs <- split(x$v, f)
    y <- Map(function(i, v, n) {
                 p <- order(v, decreasing = TRUE)[seq_len(n)]
                 v <- v[p]
                 names(v) <- terms[i[p]]
                 v
             },
             is, vs, pmin(lengths(vs), n))
    names(y) <- x$dimnames[[2L]]
    y
}

context("Term-document matrices")

test_that("construction works", {
  vs <- VectorSource(c("one two two three three three",
                       "This is a short text with a few words"))
  scorpus <- Corpus(vs)
  vcorpus <- VCorpus(vs)
  ms <- TermDocumentMatrix(scorpus)
  mv <- TermDocumentMatrix(vcorpus)
  terms <- c("few", "one", "short", "text", "this",
             "three", "two", "with", "words")
  docs <- c("1", "2")
  expect_equal(sort(Terms(ms)), terms)
  expect_equal(sort(Terms(mv)), terms)
  expect_equal(Docs(ms), docs)
  expect_equal(Docs(mv), docs)
  m <- matrix(c(0, 1, 0, 0, 0, 3, 2, 0, 0, 1, 0, 1, 1, 1, 0, 0, 1, 1), ncol = 2,
              dimnames = list("Terms" = terms, "Docs" = docs))
  expect_equal(as.matrix(ms[order(Terms(ms)), ]), m)
  expect_equal(as.matrix(mv), m)
})

test_that("construction with control arguments works", {
  vs <- VectorSource("one two two three three three")
  scorpus <- Corpus(vs)
  vcorpus <- VCorpus(vs)
  docs <- "1"
  ctrl <- list(dictionary = c("three", "two", "zero"))
  ms <- TermDocumentMatrix(scorpus, ctrl)
  mv <- TermDocumentMatrix(vcorpus, ctrl)
  m <- matrix(c(3, 2, 0),
              dimnames = list("Terms" = ctrl$dictionary, "Docs" = docs))
  expect_equal(as.matrix(ms[order(Terms(ms)), ]), m)
  expect_equal(as.matrix(mv), m)
})

test_that("zero matrix works", {
  vs <- VectorSource("one two three")
  scorpus <- Corpus(vs)
  vcorpus <- VCorpus(vs)
  ctrl <- list(dictionary = "four", wordLengths = c(1, Inf))
  ms <- TermDocumentMatrix(scorpus, ctrl)
  mv <- TermDocumentMatrix(vcorpus, ctrl)
  m <- matrix(0, dimnames = list("Terms" = ctrl$dictionary, "Docs" = "1"))
  expect_equal(as.matrix(ms), m)
  expect_equal(as.matrix(mv), m)
})

test_that("empty matrix works", {
  docs <- "1"
  ds <- DataframeSource(data.frame(doc_id = docs, text = NA))
  scorpus <- Corpus(ds)
  vcorpus <- VCorpus(ds)
  ms <- TermDocumentMatrix(scorpus)
  mv <- TermDocumentMatrix(vcorpus)
  m <- matrix(numeric(), dimnames = list("Terms" = character(), "Docs" = docs))
  expect_equal(as.matrix(ms), m)
  expect_equal(as.matrix(mv), m)
})

tm_term_score <-
function(x, terms, FUN)
    UseMethod("tm_term_score", x)

tm_term_score.term_frequency <-
function(x, terms, FUN = function(x) sum(x, na.rm = TRUE))
    FUN(x[match(terms, names(x), nomatch = 0L)])

tm_term_score.PlainTextDocument <-
function(x, terms, FUN = function(x) sum(x, na.rm = TRUE))
    tm_term_score(termFreq(x, control = list(tolower = FALSE,
                                             removePunctuation = TRUE,
                                             wordLengths = c(1, Inf))),
                  terms, FUN)

tm_term_score.TermDocumentMatrix <-
function(x, terms, FUN = col_sums)
    FUN(x[match(terms, Terms(x), nomatch = 0L), ])

tm_term_score.DocumentTermMatrix <-
function(x, terms, FUN = row_sums)
    FUN(x[, match(terms, Terms(x), nomatch = 0L)])

getTokenizers <-
function()
    c("Boost_tokenizer", "MC_tokenizer", "scan_tokenizer")

## <http://www.boost.org>
Boost_tokenizer <-
Token_Tokenizer(function(x)
{
    y <- Boost_Tokenizer(as.character(x))
    Encoding(y) <- "UTF-8"
    y
})

## <http://www.cs.utexas.edu/users/dml/software/mc/>
MC_tokenizer <-
Token_Tokenizer(function(x)
{
    x <- as.character(x)
    if(!length(x))
        return(character())
    ASCII_letters <- "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
    id <- sprintf("[%s]+", ASCII_letters)
    http <- sprintf("(https?://%s(\\.%s)*)", id, id)
    email <- sprintf("(%s@%s(\\.%s)*)", id, id, id)
    http_or_email <- sprintf("%s|%s", http, email)

    y <- c(unlist(regmatches(x, gregexpr(http_or_email, x)),
                  FALSE, FALSE),
           unlist(strsplit(gsub(http_or_email, "", x),
                           sprintf("[^%s]", ASCII_letters)),
                  FALSE, FALSE))
    y[nzchar(y)]
})

scan_tokenizer <-
Token_Tokenizer(function(x)
{
    .Call(`_tm_scan`, as.character(x), 0L)
})

stopwords <- {
    function(kind = "en") {
        kind <- as.character(kind)

        resolved <- map_IETF_Snowball(kind)
        base <- if (is.na(resolved))
            kind
        else if (identical(resolved, "porter"))
            "english"
        else
            resolved
        s <- system.file("stopwords", paste0(base, ".dat"),
                         package = "tm")
        if (identical(s, ""))
            stop(paste("no stopwords available for '", base, "'", sep = ""))
        readLines(s, encoding = "UTF-8")
    }
}

## Helper functions

.print_via_format <-
function(x, ...)
{
    writeLines(format(x, ...))
    invisible(x)
}

## Efficient alternative to table() proposed by Kurt Hornik
.table <- function(x) {
    u <- sort(unique(x))
    if(!length(u)) return(integer())
    v <- tabulate(match(x, u))
    names(v) <- u
    v
}

.xml_content <- function(doc, spec) {
    switch(spec[[1]],
           node = xml_text(xml_find_all(doc, spec[[2]])),
           "function" = spec[[2]](doc),
           unevaluated = spec[[2]])
}

IETF_Snowball_map <-
list("danish" = c("da", "dan"),
     "dutch" = c("nl", "nld", "dut"),
     "english" = c("en", "eng"),
     "finnish" = c("fi", "fin"),
     "french" = c("fr", "fra", "fre"),
     "german" = c("de", "deu", "ger"),
     "hungarian" = c("hu", "hun"),
     "italian" = c("it", "ita"),
     "norwegian" = c("no", "nor"),
     "portuguese" = c("pt", "por"),
     "romanian" = c("ro", "ron", "rum"),
     "russian" = c("ru", "rus"),
     "spanish" = c("es", "esl", "spa"),
     "swedish" = c("sv", "swe"),
     ## Have stopwords but no SnowballC stemmer ...
     "catalan" = c("ca", "cat"),
     ## Have SnowballC stemmer but no stopwords ...
     "turkish" = c("tr", "tur")
     )

# Map IETF language tags to languages used by the Snowball stemmer project
# http://en.wikipedia.org/wiki/IETF_language_tag
map_IETF_Snowball <-
local({
    codes <- unlist(IETF_Snowball_map, use.names = FALSE)
    names <- rep.int(names(IETF_Snowball_map), lengths(IETF_Snowball_map))

    function(code) {
        code <- as.character(code)

        if (identical(code, "") || identical(code, character(0)) || is.na(code))
            return("porter")

        names[charmatch(gsub("-.*", "", code), codes)]
    }
})

# Author: Ingo Feinerer

WeightFunction <- function(x, name, acronym) {
    class(x) <- c("WeightFunction", "function")
    attr(x, "name") <- name
    attr(x, "acronym") <- acronym
    x
}

# Actual TermDocumentMatrix weighting functions
weightTf <-
    WeightFunction(function(m) {
        attr(m, "weighting") <- c("term frequency", "tf")
        m
    }, "term frequency", "tf")

weightTfIdf <-
    WeightFunction(function(m, normalize = TRUE) {
        isDTM <- inherits(m, "DocumentTermMatrix")
        if (isDTM) m <- t(m)
        if (normalize) {
            cs <- col_sums(m)
            if (any(cs == 0))
                warning("empty document(s): ",
                        paste(Docs(m)[cs == 0], collapse = " "))
            names(cs) <- seq_len(nDocs(m))
            m$v <- m$v / cs[m$j]
        }
        rs <- row_sums(m > 0)
        if (any(rs == 0))
            warning("unreferenced term(s): ",
                    paste(Terms(m)[rs == 0], collapse = " "))
        lnrs <- log2(nDocs(m) / rs)
        lnrs[!is.finite(lnrs)] <- 0
        m <- m * lnrs
        attr(m, "weighting") <-
            c(sprintf("%s%s",
                      "term frequency - inverse document frequency",
                      if (normalize) " (normalized)" else ""),
              "tf-idf")
        if (isDTM) t(m) else m
    }, "term frequency - inverse document frequency", "tf-idf")

weightSMART <-
    WeightFunction(function(m, spec = "nnn", control = list()) {
        stopifnot(inherits(m, c("DocumentTermMatrix", "TermDocumentMatrix")),
                  is.character(spec), nchar(spec) == 3L, is.list(control))

        term_frequency <-
            match.arg(substr(spec, 1L, 1L),
                      c("n", "l", "a", "b", "L"))
        document_frequency <-
            match.arg(substr(spec, 2L, 2L),
                      c("n", "t", "p"))
        normalization <-
            match.arg(substr(spec, 3L, 3L),
                      c("n", "c", "u", "b"))

        isDTM <- inherits(m, "DocumentTermMatrix")
        if (isDTM) m <- t(m)

        if (normalization == "b") {
            ## Need to compute the character lengths of the documents
            ## before starting the weighting.
            charlengths <-
                tapply(nchar(Terms(m))[m$i] * m$v, m$j, sum)
        }

        ## Term frequency
        m$v <- switch(term_frequency,
                      ## natural
                      n = m$v,
                      ## logarithm
                      l = 1 + log2(m$v),
                      ## augmented
                      a = {
                          s <- tapply(m$v, m$j, max)
                          0.5 + (0.5 * m$v) / s[as.character(m$j)]
                      },
                      ## boolean
                      b = as.numeric(m$v > 0),
                      ## log ave
                      L = {
                          s <- tapply(m$v, m$j, mean)
                          ((1 + log2(m$v)) / (1 + log2(s[as.character(m$j)])))
                      })

        ## Document frequency
        rs <- row_sums(m > 0)
        if (any(rs == 0))
            warning("unreferenced term(s): ",
                    paste(Terms(m)[rs == 0], collapse = " "))
        df <- switch(document_frequency,
                     ## natural
                     n = 1,
                     ## idf
                     t = log2(nDocs(m) / rs),
                     ## prob idf
                     p = max(0, log2((nDocs(m) - rs) / rs)))
        df[!is.finite(df)] <- 0

        ## Normalization
        cs <- col_sums(m)
        if (any(cs == 0))
            warning("empty document(s): ",
                    paste(Docs(m)[cs == 0], collapse = " "))
        norm <- switch(normalization,
                       ## none
                       n = rep.int(1, nDocs(m)),
                       ## cosine
                       c = sqrt(col_sums(m ^ 2)),
                       ## pivoted unique
                       u = {
                           if (is.null(pivot <- control$pivot))
                               stop("invalid control argument pivot")
                           if (is.null(slope <- control$slope))
                               stop("invalid control argument slope")
                           (slope * sqrt(col_sums(m ^ 2)) +
                            (1 - slope) * pivot)
                       },
                       ## byte size
                       b = {
                           if (is.null(alpha <- control$alpha))
                               stop("invalid control argument alpha")
                           norm <- double(nDocs(m))
                           norm[match(names(charlengths),
                                      seq_along(norm))] <-
                                          charlengths ^ alpha
                           norm
                       })

        m <- m * df
        m$v <- m$v / norm[m$j]
        attr(m, "weighting") <- c(paste("SMART", spec), "SMART")

        if (isDTM) t(m) else m
    }, "SMART", "SMART")

weightBin <-
    WeightFunction(function(m) {
        m$v <- rep_len(1L, length(m$v))
        attr(m, "weighting") <- c("binary", "bin")
        m
    }, "binary", "bin")

# Author: Ingo Feinerer

TextDocumentMeta <-
function(author, datetimestamp, description, heading, id, language, origin, ...,
         meta = NULL)
{
    if (is.null(meta))
        meta <- list(author = author, datetimestamp = datetimestamp,
                     description = description, heading = heading, id = id,
                     language = language, origin = origin, ...)

    stopifnot(is.list(meta))
    if (!is.null(meta$author) && !inherits(meta$author, "person"))
        meta$author <- as.character(meta$author)
    if (!is.null(meta$datetimestamp) && !inherits(meta$datetimestamp, "POSIXt"))
        meta$datetimestamp <- as.character(meta$datetimestamp)
    if (!is.null(meta$description))
       meta$description <- as.character(meta$description)
    if (!is.null(meta$heading))
       meta$heading <- as.character(meta$heading)
    if (!is.null(meta$id))
       meta$id <- as.character(meta$id)
    if (!is.null(meta$language))
       meta$language <- as.character(meta$language)
    if (!is.null(meta$origin))
       meta$origin <- as.character(meta$origin)

    class(meta) <- "TextDocumentMeta"
    meta
}

print.TextDocumentMeta <-
function(x, ...)
{
    cat(sprintf("  %s: %s",
                format(names(x), justify = "left"),
                sapply(x, as.character)),
        sep = "\n")
    invisible(x)
}

CorpusMeta <-
function(..., meta = NULL)
{
    if (is.null(meta))
        meta <- list(...)

    stopifnot(is.list(meta))

    class(meta) <- "CorpusMeta"
    meta
}

meta.SimpleCorpus <-
function(x, tag = NULL, type = c("indexed", "corpus"), ...)
{
    if (identical(tag, "id")) {
        n <- names(content(x))
        return(if (is.null(n)) as.character(seq_along(x)) else n)
    }

    if (!is.null(tag) && missing(type))
        type <- if (tag %in% names(x$meta)) "corpus" else "indexed"
    type <- match.arg(type)
    if (identical(type, "indexed"))
        if (is.null(tag)) x$dmeta else x$dmeta[tag]
    else if (identical(type, "corpus"))
        if (is.null(tag)) x$meta else x$meta[[tag]]
    else
        stop("invalid type")
}

meta.VCorpus <- meta.PCorpus <-
function(x, tag = NULL, type = c("indexed", "corpus", "local"), ...)
{
    if (!is.null(tag) && missing(type)) {
        type <- if (tag %in% names(x$dmeta)) "indexed"
        else if (tag %in% names(x$meta)) "corpus"
        else "local"
    }
    type <- match.arg(type)
    if (identical(type, "indexed"))
        if (is.null(tag)) x$dmeta else x$dmeta[tag]
    else if (identical(type, "corpus"))
        if (is.null(tag)) x$meta else x$meta[[tag]]
    else if (identical(type, "local"))
        lapply(x, meta, tag)
    else
        stop("invalid type")
}

`meta<-.SimpleCorpus` <-
function(x, tag, type = c("indexed", "corpus"), ..., value)
{
    type <- match.arg(type)
    if (identical(type, "indexed"))
        x$dmeta[, tag] <- value
    else if (type == "corpus")
        x$meta[[tag]] <- value
    else
        stop("invalid type")
    x
}

`meta<-.VCorpus` <- `meta<-.PCorpus` <-
function(x, tag, type = c("indexed", "corpus", "local"), ..., value)
{
    type <- match.arg(type)
    if (identical(type, "indexed"))
        x$dmeta[, tag] <- value
    else if (type == "corpus")
        x$meta[[tag]] <- value
    else if (identical(type, "local")) {
        for (i in seq_along(x))
            meta(x[[i]], tag) <- value[i]
    } else
        stop("invalid type")
    x
}

# Simple Dublin Core to tm metadata mapping
# http://en.wikipedia.org/wiki/Dublin_core#Simple_Dublin_Core
Dublin_Core_tm_map <-
list("contributor" = "contributor",
     "coverage" = "coverage",
     "creator" = "author",
     "date" = "datetimestamp",
     "description" = "description",
     "format" = "format",
     "identifier" = "id",
     "language" = "language",
     "publisher" = "publisher",
     "relation" = "relation",
     "rights" = "rights",
     "source" = "source", # or better "origin"?
     "subject" = "subject",
     "title" = "heading",
     "type" = "type"
     )

DublinCore <-
function(x, tag = NULL)
{
    tmm <- unlist(Dublin_Core_tm_map, use.names = FALSE)
    dcm <- names(Dublin_Core_tm_map)

    if (is.null(tag)) {
        m <- lapply(tmm, function(t) meta(x, t))
        names(m) <- dcm
        class(m) <- "TextDocumentMeta"
        m
    } else
        meta(x, tmm[charmatch(tolower(tag), dcm)])
}

`DublinCore<-` <-
function(x, tag, value)
{
    tmm <- unlist(Dublin_Core_tm_map, use.names = FALSE)
    dcm <- names(Dublin_Core_tm_map)

    meta(x, tmm[charmatch(tolower(tag), dcm)]) <- value
    x
}

## Readers and writers (eventually?) for foreign document-term matrix
## format files.

## CLUTO: as we do not know the weighting, there is no high-level DTM
## reader.  If the weighting is weightTf, one can do
##   as.DocumentTermMatrix(read_stm_CLUTO(file), weightTf)
## as CLUTO always has rows as documents and cols as terms.

## MC: a simple reader for now, could certainly use more effort to name
## the weightings more properly.

read_dtm_MC <-
function(file, scalingtype = NULL)
{
    m <- read_stm_MC(file, scalingtype)
    s <- attr(m, "scalingtype")
    as.DocumentTermMatrix(m, rep.int(s, 2L))
}

## <FIXME>
## To write a decent writer we would need to be able to turn weighting
## information into MC scaling information, which may not even be
## possible.  Alternatively, we could always use 'txx', or use this in
## case we cannot map ...
## </FIXME>

## Data files for the Blei et al LDA and CTM codes are in a List of List
## format, with lines
##   n j1: x1 j2: x2 ... jn: xn
## (see http://www.cs.princeton.edu/~blei/lda-c/).
## As they are used for topic models, they *always* contain raw term
## frequencies.

read_dtm_Blei_et_al <-
function(file, vocab = NULL)
{
    x <- scan(file, character(), quiet = TRUE)
    ind <- grepl(":", x, fixed = TRUE)
    counts <- x[!ind]
    i <- rep.int(seq_along(counts), counts)
    x <- strsplit(x[ind], ":", fixed = TRUE)
    j <- as.integer(unlist(lapply(x, `[`, 1L))) + 1L
    x <- as.numeric(unlist(lapply(x, `[`, 2L)))
    m <- simple_triplet_matrix(i, j, x)
    if (!is.null(vocab))
        colnames(m) <- readLines(vocab)
    as.DocumentTermMatrix(m, weightTf)
}

# Author: Ingo Feinerer

stemCompletion <-
function(x, dictionary,
         type = c("prevalent", "first", "longest",
                  "none", "random", "shortest"))
{
    if (inherits(dictionary, "Corpus"))
        dictionary <- unlist(lapply(dictionary, words))

    type <- match.arg(type)
    possibleCompletions <- lapply(x, function(w) grep(sprintf("^%s", w),
                                                      dictionary,
                                                      value = TRUE))
    switch(type,
           first = {
               setNames(sapply(possibleCompletions, "[", 1), x)
           },
           longest = {
               ordering <-
                   lapply(possibleCompletions,
                          function(x) order(nchar(x), decreasing = TRUE))
               possibleCompletions <-
                   mapply(function(x, id) x[id], possibleCompletions,
                          ordering, SIMPLIFY = FALSE)
               setNames(sapply(possibleCompletions, "[", 1), x)
           },
           none = {
               setNames(x, x)
           },
           prevalent = {
               possibleCompletions <-
                   lapply(possibleCompletions,
                          function(x) sort(table(x), decreasing = TRUE))
               n <- names(sapply(possibleCompletions, "[", 1))
               setNames(if (length(n)) n else rep_len(NA, length(x)), x)
           },
           random = {
               setNames(sapply(possibleCompletions, function(x) {
                   if (length(x)) sample(x, 1) else NA
               }), x)
           },
           shortest = {
               ordering <- lapply(possibleCompletions,
                                  function(x) order(nchar(x)))
               possibleCompletions <-
                   mapply(function(x, id) x[id], possibleCompletions,
                          ordering, SIMPLIFY = FALSE)
               setNames(sapply(possibleCompletions, "[", 1), x)
           }
           )
}

pdf_info_via_xpdf <-
function(file, options = NULL)
{
    outfile <- tempfile("pdfinfo")
    on.exit(unlink(outfile))
    status <- system2("pdfinfo", c(options, shQuote(normalizePath(file))),
                      stdout = outfile)
    ## Could check the status ...
    ## This does not work ...
    ##   info <- as.list(read.dcf(outfile)[1L, ])

    tags <- c("Title", "Subject", "Keywords", "Author", "Creator",
              "Producer", "CreationDate", "ModDate", "Tagged", "Form",
              "Pages", "Encrypted", "Page size", "File size",
              "Optimized", "PDF version")
    re <- sprintf("^(%s)",
                  paste(sprintf("%-16s", sprintf("%s:", tags)),
                        collapse = "|"))
    lines <- readLines(outfile, warn = FALSE)
    ind <- grepl(re, lines)
    tags <- sub(": *", "", substring(lines[ind], 1L, 16L))
    info <- split(sub(re, "", lines), cumsum(ind))
    names(info) <- tags
    fmt <- "%a %b %d %X %Y"
    if (!is.null(d <- info$CreationDate))
        info$CreationDate <- strptime(d, fmt)
    if (!is.null(d <- info$ModDate))
        info$ModDate <- strptime(d, fmt)
    if (!is.null(p <- info$Pages))
        info$Pages <- as.integer(p)
    info
}

pdf_info_via_gs <-
function(file)
{
    file <- normalizePath(file)

    gs_cmd <- tools::find_gs_cmd()

    out <- system2(gs_cmd,
                   c("-dNODISPLAY -q",
                     sprintf("-sFile=%s", shQuote(file)),
                     system.file("ghostscript", "pdf_info.ps",
                                 package = "tm")),
                   stdout = TRUE)
    out <- out[cumsum(out == "") == 2L][-1L]
    val <- sub("^[^:]+:[[:space:]]*", "", out)
    names(val) <- sub(":.*", "", out)
    val <- as.list(val)
    if (!is.null(d <- val$CreationDate))
        val$CreationDate <- PDF_Date_to_POSIXt(d)
    if (!is.null(d <- val$ModDate))
        val$ModDate <- PDF_Date_to_POSIXt(d)

    val
}

PDF_Date_to_POSIXt <-
function(s)
{
    ## Strip optional 'D:' prefix.
    s <- sub("^D:", "", s)
    ## Strip apostrophes in offset spec.
    s <- gsub("'", "", s)
    if (nchar(s) <= 14L) {
        s <- sprintf("%s%s", s,
                     substring("    0101000000", nchar(s) + 1L, 14L))
        strptime(s, "%Y%m%d%H%M%S")
    } else if (substring(s, 15L, 15L) == "Z") {
        strptime(substring(s, 1L, 14L), "%Y%m%d%H%M%S")
    } else {
        strptime(s, "%Y%m%d%H%M%S%z")
    }
}

pdf_text_via_gs <-
function(file)
{
    file <- normalizePath(file)

    gs_cmd <- tools::find_gs_cmd()

    tf <- tempfile("pdf")
    on.exit(unlink(tf))

    ## The current mechanism is first converting PDF to Postscript using
    ## the ps2write device, and then extract text using the ps2ascii.ps
    ## program.  This fails for some files (e.g.,
    ## /data/rsync/PKGS/AlleleRetain/inst/doc/AlleleRetain_User_Guide.pdf
    ## which Ghostscript also fails to render.  Note that rendering via
    ## gv works "fine": but this uses the pswrite device which produces
    ## bitmap (from which no text can be extracted, of course).
    ## Using the txtwrite device is simply too unstable: e.g.,
    ##   gs -dBATCH -dNOPAUSE -sDEVICE=txtwrite -dQUIET -sOutputFile=- \
    ##     /data/rsync/PKGS/AlleleRetain/inst/doc/AlleleRetain_User_Guide.pdf
    ## keeps segfaulting.
    ## An additional nuisance is that there seems no simple way to
    ## detect a ps2ascii.ps failure.

    ## Finally, note that we currently use -DSIMPLE: without this, more
    ## information would be made available, but require post-processing.

    ## Step 1.  Convert PDF to Postscript.
    res <- system2(gs_cmd,
                   c("-q -dNOPAUSE -dBATCH -P- -dSAFER -sDEVICE=ps2write",
                     sprintf("-sOutputFile=%s", tf),
                     "-c save pop -f",
                     shQuote(file)))
    ## Step 2.  Extract text.
    txt <- system2(gs_cmd,
                   c("-q -dNODISPLAY -P- -dSAFER -dDELAYBIND -dWRITESYSTEMDICT -dSIMPLE",
                     "-c save -f ps2ascii.ps",
                     tf,
                     "-c quit"),
                   stdout = TRUE)
    ## Argh.  How can we catch errors?
    ## The return values are always 0 ...
    if (any(grepl("Error handled by opdfread.ps", txt))) {
        stop(paste(c("Ghostscript failed, with output:", txt),
                   collapse = "\n"))
    }

    strsplit(paste(txt, collapse = "\n"), "\f")[[1L]]
}

## Author: Ingo Feinerer
## Sources

getSources <-
function()
   c("DataframeSource", "DirSource", "URISource", "VectorSource", "XMLSource",
     "ZipSource")

SimpleSource <-
function(encoding = "",
         length = 0,
         position = 0,
         reader = readPlain,
         ...,
         class)
{
    if (!is.character(encoding))
        stop("invalid encoding")
    if (!is.numeric(length) || (length < 0))
        stop("invalid length entry denoting the number of elements")
    if (!is.numeric(position))
        stop("invalid position")
    if (!is.function(reader))
        stop("invalid default reader")

    s <- list(encoding = encoding, length = length,
              position = position, reader = reader, ...)
    class(s) <- unique(c(class, "SimpleSource", "Source"))
    s
}

# A data frame where each row is interpreted as document
DataframeSource <-
function(x)
{
    stopifnot(all(!is.na(match(c("doc_id", "text"), names(x)))))

    SimpleSource(length = nrow(x), reader = readDataframe,
                 content = x, class = "DataframeSource")
}

# A directory with files interpreted as documents
DirSource <-
function(directory = ".", encoding = "", pattern = NULL,
         recursive = FALSE, ignore.case = FALSE, mode = "text")
{
    if (!identical(mode, "text") &&
        !identical(mode, "binary") &&
        !identical(mode, ""))
        stop(sprintf("invalid mode '%s'", mode))

    d <- dir(directory, full.names = TRUE, pattern = pattern,
             recursive = recursive, ignore.case = ignore.case)

    if (!length(d))
        stop("empty directory")

    isfile <- !file.info(d)[["isdir"]]
    if (any(is.na(isfile)))
        stop("non-existent or non-readable file(s): ",
             paste(d[is.na(isfile)], collapse = " "))

    SimpleSource(encoding = encoding, length = sum(isfile),
                 mode = mode, filelist = d[isfile], class = "DirSource")
}

# Documents identified by a Uniform Resource Identifier
URISource <-
function(x, encoding = "", mode = "text")
{
    if (!identical(mode, "text") &&
        !identical(mode, "binary") &&
        !identical(mode, ""))
        stop(sprintf("invalid mode '%s'", mode))

    SimpleSource(encoding = encoding, length = length(x), mode = mode, uri = x,
                 class = "URISource")
}

# A vector where each component is interpreted as document
VectorSource <-
function(x)
    SimpleSource(length = length(x), content = x, class = "VectorSource")

XMLSource <-
function(x, parser = xml_contents, reader)
{
    xmldoc <- read_xml(x)
    content <- parser(xmldoc)

    SimpleSource(length = length(content), reader = reader, content = content,
                 uri = x, class = "XMLSource")
}

# A ZIP file with its compressed files interpreted as documents
ZipSource <-
function(zipfile, pattern = NULL, recursive = FALSE, ignore.case = FALSE,
         mode = "text")
{
    if (!identical(mode, "text") &&
        !identical(mode, "binary") &&
        !identical(mode, ""))
        stop(sprintf("invalid mode '%s'", mode))

    SimpleSource(exdir = NULL,
                 files = NULL,
                 mode = mode,
                 pattern = pattern,
                 recursive = recursive,
                 ignore.case = ignore.case,
                 zipfile = zipfile,
                 class = "ZipSource")
}

# tau:::read_all_bytes
read_all_bytes <-
function(con, chunksize = 2 ^ 16)
{
    if (is.character(con)) {
        return(readBin(con, raw(), file.info(con)$size))
    }

    if (!isOpen(con)) {
        open(con, "rb")
        on.exit(close(con))
    }

    bytes <- list()
    repeat {
        chunk <- readBin(con, raw(), chunksize)
        bytes <- c(bytes, list(chunk))
        if (length(chunk) < chunksize) break
    }

    unlist(bytes)
}

readContent <-
function(x, encoding, mode)
{
    if (identical(mode, "text"))
        iconv(readLines(x, warn = FALSE), encoding, "UTF-8", "byte")
    else if (identical(mode, "binary"))
        read_all_bytes(x)
    else if (identical(mode, ""))
        NULL
    else
        stop("invalid mode")
}

open.SimpleSource <-
close.SimpleSource <-
function(con, ...)
    con
open.ZipSource <-
function(con, ...)
{
    x <- con
    exdir <- tempfile("ZipSource")
    dir.create(exdir, mode = "0700")

    destfile <- x$zipfile

    if (!file.exists(destfile)) {
        destfile <- tempfile()
        download.file(x$zipfile, destfile)
        on.exit(file.remove(destfile))
    }

    files <- unzip(destfile, list = TRUE)
    ## Directories have length 0
    files <- files[files$Length > 0, "Name"]
    ## Idea: Subdirectories contain file separators
    if (!x$recursive)
        files <- files[!grepl(.Platform$file.sep, files, fixed = TRUE)]
    ## Idea: pattern and ignore.case refer to the file name (like basename)
    ## Cf. also ?dir
    if (!is.null(x$pattern))
        files <- files[grepl(x$pattern, files, ignore.case = x$ignore.case)]

    unzip(destfile, files, exdir = exdir)

    x$exdir <- exdir
    x$files <- files
    x$length <- length(files)
    x
}

close.ZipSource <-
function(con, ...)
{
    x <- con
    if (!is.null(x$exdir)) {
        unlink(x$exdir, recursive = TRUE)
        x$exdir <- NULL
        x$files <- NULL
        x$length <- 0
    }
    x
}

eoi <-
function(x)
    UseMethod("eoi", x)
eoi.SimpleSource <-
function(x)
    x$length <= x$position

getElem <-
function(x)
    UseMethod("getElem", x)
getElem.DataframeSource <-
function(x)
    list(content = x$content[x$position, ],
         uri = NULL)
getElem.DirSource <-
function(x)
{
    filename <- x$filelist[x$position]
    list(content = readContent(filename, x$encoding, x$mode),
         uri = paste0("file://", filename))
}
getElem.URISource <-
function(x)
    list(content = readContent(x$uri[x$position], x$encoding, x$mode),
         uri = x$uri[x$position])
getElem.VectorSource <-
function(x)
    list(content = x$content[x$position],
         uri = NULL)
getElem.XMLSource <-
function(x)
    list(content = x$content[[x$position]],
         uri = x$uri)
getElem.ZipSource <-
function(x)
{
       path <- file.path(x$exdir, x$files[x$position])
       list(content = readContent(path, x$encoding, x$mode),
            uri = paste0("file://", path))
}

getMeta <-
function(x)
    UseMethod("getMeta", x)
getMeta.DataframeSource <-
function(x)
    list(cmeta = NULL,
         dmeta = x$content[, is.na(match(names(x$content),
                                         c("doc_id", "text"))),
                           drop = FALSE])


length.SimpleSource <-
function(x)
    x$length

pGetElem <-
function(x)
    UseMethod("pGetElem", x)

pGetElem.DataframeSource <-
function(x)
    tm_parLapply(seq_len(x$length),
                 function(y)
                     list(content = x$content[y, ],
                          uri = NULL))
`[.DataframeSource` <- function(x, i, j, ...) x$content[i, j, ...]
`[[.DataframeSource` <- function(x, ...) x$content[[...]]

pGetElem.DirSource <-
function(x)
    tm_parLapply(x$filelist,
                 function(f)
                     list(content = readContent(f, x$encoding, x$mode),
                          uri = paste0("file://", f)))
`[.DirSource` <- function(x, i, ...) x$filelist[i, ...]
`[[.DirSource` <- function(x, i, ...) x$filelist[[i, ...]]

pGetElem.URISource <-
function(x)
    tm_parLapply(x$uri,
                 function(uri)
                     list(content = readContent(uri, x$encoding, x$mode),
                          uri = uri))
`[.URISource` <- function(x, i, ...) x$uri[i, ...]
`[[.URISource` <- function(x, i, ...) x$uri[[i, ...]]

pGetElem.VectorSource <-
function(x)
    tm_parLapply(x$content,
                 function(y) list(content = y,
                                  uri = NULL))
`[.VectorSource` <- function(x, i, ...) x$content[i, ...]
`[[.VectorSource` <- function(x, i, ...) x$content[[i, ...]]

pGetElem.ZipSource <-
function(x)
    tm_parLapply(file.path(x$exdir, x$files),
                 function(f)
                     list(content = readContent(f, x$encoding, x$mode),
                          uri = paste0("file://", f)))

reader <-
function(x)
    UseMethod("reader", x)
reader.SimpleSource <-
function(x)
    x$reader

stepNext <-
function(x)
    UseMethod("stepNext", x)
stepNext.SimpleSource <-
function(x)
{
    x$position <- x$position + 1
    x
}

## Author: Ingo Feinerer
## Readers

FunctionGenerator <-
function(x)
{
    class(x) <- c("FunctionGenerator", "function")
    x
}

getReaders <-
function()
    c("readDataframe", "readDOC", "readPDF", "readPlain", "readRCV1",
      "readRCV1asPlain", "readReut21578XML", "readReut21578XMLasPlain",
      "readTagged", "readXML")

prepareReader <-
function(readerControl, reader = NULL, ...)
{
    if (is.null(readerControl$reader))
        readerControl$reader <- reader
    if (inherits(readerControl$reader, "FunctionGenerator"))
        readerControl$reader <- readerControl$reader(...)
    if (is.null(readerControl$language))
        readerControl$language <- "en"
    readerControl
}

processURI <-
function(uri)
{
    uri <- as.character(uri)
    if (identical(substr(uri, 1, 7), "file://"))
        uri <- substr(uri, 8, nchar(uri))
    uri
}

readDataframe <-
function(elem, language, id) {
    PlainTextDocument(elem$content[, "text"],
                      id = elem$content[, "doc_id"],
                      language = language)
}

# readDOC needs antiword installed to be able to extract the text
readDOC <-
function(engine = c("antiword", "executable"),
         AntiwordOptions = "")
{
    stopifnot(is.character(engine), is.character(AntiwordOptions))

    engine <- match.arg(engine)

    antiword <-
        switch(engine,
               antiword = antiword::antiword,
               executable = function(x)
                   system2("antiword",
                           c(AntiwordOptions, shQuote(normalizePath(x))),
                           stdout = TRUE))

    if (!is.function(antiword))
        stop("invalid function for DOC extraction")

    function(elem, language, id) {
        uri <- processURI(elem$uri)
        content <- antiword(uri)
        PlainTextDocument(content, id = basename(elem$uri), language = language)
    }
}
class(readDOC) <- c("FunctionGenerator", "function")

readPDF <-
function(engine = c("pdftools", "xpdf", "Rpoppler",
                    "ghostscript", "Rcampdf", "custom"),
         control = list(info = NULL, text = NULL))
{
    stopifnot(is.character(engine), is.list(control))

    engine <- match.arg(engine)

    pdf_info <-
        switch(engine,
               pdftools = function(x) {
                   i <- pdftools::pdf_info(x)
                   c(i$keys, list(CreationDate = i$created))
               },
               xpdf = function(x) pdf_info_via_xpdf(x, control$info),
               Rpoppler = Rpoppler::PDF_info,
               ghostscript = pdf_info_via_gs,
               Rcampdf = Rcampdf::pdf_info,
               custom = control$info)

    pdf_text <-
        switch(engine,
               pdftools = pdftools::pdf_text,
               xpdf = function(x) system2("pdftotext",
                                          c(control$text, shQuote(x), "-"),
                                          stdout = TRUE),
               Rpoppler = Rpoppler::PDF_text,
               ghostscript = pdf_text_via_gs,
               Rcampdf = Rcampdf::pdf_text,
               custom = control$text)

    if (!is.function(pdf_info) || !is.function(pdf_text))
        stop("invalid function for PDF extraction")

    function(elem, language, id) {
        uri <- processURI(elem$uri)
        meta <- pdf_info(uri)
        content <- pdf_text(uri)
        PlainTextDocument(content, meta$Author, meta$CreationDate, meta$Subject,
                          meta$Title, basename(elem$uri), language,
                          meta$Creator)
     }
}
class(readPDF) <- c("FunctionGenerator", "function")

readPlain <-
function(elem, language, id) {
    if (!is.null(elem$uri))
        id <- basename(elem$uri)
    PlainTextDocument(elem$content, id = id, language = language)
}

readXML <-
function(spec, doc)
{
    stopifnot(is.list(spec), inherits(doc, "TextDocument"))

    function(elem, language, id) {
        content <- elem$content
        node <- if(inherits(content, "xml_node"))
            content
        else if(is.character(content)) 
            read_xml(paste(elem$content, collapse = "\n"))
        else
            read_xml(content)
        content(doc) <- if ("content" %in% names(spec))
            .xml_content(node, spec[["content"]])
        else
            node
        for (n in setdiff(names(spec), "content"))
            meta(doc, n) <- .xml_content(node, spec[[n]])
        if (!is.null(elem$uri))
            id <- basename(elem$uri)
        if (!length(meta(doc, "id")))
            meta(doc, "id") <- as.character(id)
        if (!length(meta(doc, "language")))
            meta(doc, "language") <- as.character(language)
        doc
    }
}
class(readXML) <- c("FunctionGenerator", "function")

RCV1Spec <-
    list(author = list("unevaluated", ""),
         datetimestamp = list("function", function(node)
           as.POSIXlt(xml_text(xml_find_all(node, "@date")),
                      tz = "GMT")),
         description = list("unevaluated", ""),
         heading = list("node", "title"),
         id = list("node", "@itemid"),
         origin = list("unevaluated", "Reuters Corpus Volume 1"),
         publisher = list("node",
           "metadata/dc[@element='dc.publisher']/@value"),
         topics = list("node",
           "metadata/codes[@class='bip:topics:1.0']/code/@code"),
         industries = list("node",
           "metadata/codes[@class='bip:industries:1.0']/code/@code"),
         countries = list("node",
           "metadata/codes[@class='bip:countries:1.0']/code/@code"))
readRCV1 <-
    readXML(spec = RCV1Spec,
            doc = XMLTextDocument())
readRCV1asPlain <-
    readXML(spec = c(RCV1Spec,
                     list(content = list("node", "text"))),
            doc = PlainTextDocument())

Reut21578XMLSpec <-
    list(author = list("node", "TEXT/AUTHOR"),
         datetimestamp = list("function", function(node)
           strptime(xml_text(xml_find_all(node, "DATE")),
                    format = "%d-%B-%Y %H:%M:%S",
                    tz = "GMT")),
         description = list("unevaluated", ""),
         heading = list("node", "TEXT/TITLE"),
         id = list("node", "@NEWID"),
         topics = list("node", "@TOPICS"),
         lewissplit = list("node", "@LEWISSPLIT"),
         cgisplit = list("node", "@CGISPLIT"),
         oldid = list("node", "@OLDID"),
         origin = list("unevaluated", "Reuters-21578 XML"),
         topics_cat = list("node", "TOPICS/D"),
         places = list("node", "PLACES/D"),
         people = list("node", "PEOPLE/D"),
         orgs = list("node", "ORGS/D"),
         exchanges = list("node", "EXCHANGES/D"))
readReut21578XML <-
    readXML(spec = Reut21578XMLSpec,
            doc = XMLTextDocument())
readReut21578XMLasPlain <-
    readXML(spec = c(Reut21578XMLSpec,
                     list(content = list("node", "TEXT/BODY"))),
            doc = PlainTextDocument())

readTagged <-
function(...)
{
    args <- list(...)
    function(elem, language, id) {
        if (!is.null(elem$content)) {
            con <- textConnection(elem$content)
            on.exit(close(con))
        } else
            con <- elem$uri

        if (!is.null(elem$uri))
            id <- basename(elem$uri)

        a <- c(list(con = con, meta = list(id = id, language = language)), args)

        do.call(TaggedTextDocument, a)
    }
}
class(readTagged) <- c("FunctionGenerator", "function")

# Author: Ingo Feinerer
# Transformations

tm_map <-
function(x, FUN, ...)
    UseMethod("tm_map", x)
tm_map.VCorpus <-
function(x, FUN, ..., lazy = FALSE)
{
    # Lazy mapping
    if (lazy) {
        fun <- function(x) FUN(x, ...)
        if (is.null(x$lazy))
            x$lazy <- list(index = rep_len(TRUE, length(x)), maps = list(fun))
        else
            x$lazy$maps <- c(x$lazy$maps, list(fun))
    } else
        x$content <- tm_parLapply(content(x), FUN, ...)
    x
}
tm_map.SimpleCorpus <-
function(x, FUN, ...)
{
    if (inherits(FUN, "content_transformer"))
        FUN <- get("FUN", envir = environment(FUN))

    n <- names(content(x))
    x$content <- FUN(content(x), ...)
    if (length(content(x)) != length(n))
        warning("transformation drops documents")
    else
        names(x$content) <- n
    x
}
tm_map.PCorpus <-
function(x, FUN, ...)
{
    db <- filehash::dbInit(x$dbcontrol[["dbName"]], x$dbcontrol[["dbType"]])
    for (i in seq_along(x))
        db[[x$content[[i]]]] <- FUN(x[[i]], ...)
    filehash::dbReorganize(db)
    x
}

# Materialize lazy mappings
materialize <-
function(x, range = seq_along(x))
{
    if (!is.null(x$lazy)) {
       i <- (seq_along(x) %in% range) & x$lazy$index
       if (any(i)) {
           x$content[i] <-
               tm_parLapply(x$content[i],
                            function(d) tm_reduce(d, x$lazy$maps))
           x$lazy$index[i] <- FALSE
       }

       # Clean up if everything is materialized
       if (!any(x$lazy$index))
           x["lazy"] <- list(NULL)
    }
    x
}

tm_reduce <-
function(x, tmFuns, ...)
    Reduce(function(f, ...) f(...), tmFuns, x, right = TRUE)

getTransformations <-
function()
    c("removeNumbers", "removePunctuation", "removeWords", "stemDocument",
      "stripWhitespace")

content_transformer <-
function(FUN)
{
    f <- function(x, ...) {
        content(x) <- FUN(content(x), ...)
        x
    }
    class(f) <- c("content_transformer", "function")
    f
}

removeNumbers <-
function(x, ...)
    UseMethod("removeNumbers")
removeNumbers.character <-
function(x, ucp = FALSE, ...)
{
    if (ucp)
        gsub("\\p{Nd}+", "", x, perl = TRUE)
    else
        .Call(`_tm_remove_chars`, x, 1L)
}
removeNumbers.PlainTextDocument <-
    content_transformer(removeNumbers.character)

removePunctuation <-
function(x, ...)
    UseMethod("removePunctuation")
removePunctuation.character <-
function(x,
         preserve_intra_word_contractions = FALSE,
         preserve_intra_word_dashes = FALSE,
         ucp = FALSE,
         ...)
{
    # Assume there are no ASCII 0x01 (SOH) or ASCII 0x02 (STX) characters.
    if (preserve_intra_word_contractions)
        x <- gsub("(\\w)'(\\w)", "\\1\1\\2", x, perl = TRUE)
    if (preserve_intra_word_dashes)
        x <- gsub("(\\w)-(\\w)", "\\1\2\\2", x, perl = TRUE)

    if (ucp)
        x <- gsub("\\p{P}+", "", x, perl = TRUE)
    else
        x <- .Call(`_tm_remove_chars`, x, 0L)

    if (preserve_intra_word_contractions)
        x <- gsub("\1", "'", x, fixed = TRUE)
    if (preserve_intra_word_dashes)
        x <- gsub("\2", "-", x, fixed = TRUE)

    x
}
removePunctuation.PlainTextDocument <-
    content_transformer(removePunctuation.character)

removeWords <-
function(x, words)
    UseMethod("removeWords", x)
# Improvements by Kurt Hornik
removeWords.character <-
function(x, words)
    gsub(sprintf("(*UCP)\\b(%s)\\b",
                 paste(sort(words, decreasing = TRUE), collapse = "|")),
         "", x, perl = TRUE)
removeWords.PlainTextDocument <-
    content_transformer(removeWords.character)

stemDocument <-
function(x, language = "english")
    UseMethod("stemDocument", x)
stemDocument.character <-
function(x, language = "english")
{
    s <- unlist(lapply(x, function(line)
                              paste(SnowballC::wordStem(words(line),
                                                        as.character(language)),
                                    collapse = " ")))
    if (is.character(s)) s else ""
}
stemDocument.PlainTextDocument <-
function(x, language = meta(x, "language"))
{
    language <- as.character(language)
    if (identical(language, "") ||
        identical(language, character(0)) ||
        is.na(language))
        language <- "english"

    content_transformer(stemDocument.character)(x)
}

stripWhitespace <-
function(x)
    UseMethod("stripWhitespace", x)
stripWhitespace.character <-
function(x)
    gsub("[[:space:]]+", " ", x)
stripWhitespace.PlainTextDocument <-
    content_transformer(stripWhitespace.character)

context("Transformations")

test_that("removePunctuation works in latin1 locale", {
  if (nzchar(suppressWarnings(Sys.setlocale("LC_CTYPE", "en_US.iso88591")))) {
    id <- c(73L, 108L, 32L, 115L, 39L, 101L, 120L, 112L, 114L, 105L, 109L, 97L,
            105L, 116L,  32L, 101L, 110L, 32L, 117L, 110L, 32L, 108L, 97L, 110L,
            103L,  97L, 103L, 101L,  32L,  99L, 104L, 226L, 116L, 105L, 233L)
    iu <- intToUtf8(id)
    il <- iconv(iu, from = "UTF-8", to = "latin1")
    td <- id[-5L]
    tu <- intToUtf8(td)
    tl <- iconv(tu, from = "UTF-8", to = "latin1")

    expect_equal(removePunctuation(iu), tu)
    expect_equal(removePunctuation(il), tl)
  } else
    skip("latin1 locale not available")
})

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

tdm <- function(strings, remove_puncts, remove_digits, stopwords, dictionary, min_term_freq, max_term_freq, min_word_length, max_word_length) {
    .Call(`_tm_tdm`, strings, remove_puncts, remove_digits, stopwords, dictionary, min_term_freq, max_term_freq, min_word_length, max_word_length)
}

Boost_Tokenizer <- function(strings) {
    .Call(`_tm_Boost_Tokenizer`, strings)
}

plot.TermDocumentMatrix <- plot.DocumentTermMatrix <-
function(x,
         terms = sample(Terms(x), 20),
         corThreshold = 0.7,
         weighting = FALSE,
         attrs = list(graph = list(rankdir = "BT"),
                      node = list(shape = "rectangle", fixedsize = FALSE)),
         ...)
{
    if (system.file(package = "Rgraphviz") == "")
        stop("Plotting requires package 'Rgraphviz'.")

    m <- if (inherits(x, "TermDocumentMatrix")) t(x) else x
    m <- as.matrix(m[, terms])
    c <- cor(m)
    c[c < corThreshold] <- 0
    c[is.na(c)] <- 0
    diag(c) <- 0
    p <- Rgraphviz::plot(methods::as(c, "graphNEL"), attrs = attrs, ...)
    if (weighting) {
        i <- 1
        lw <- round(c[lower.tri(c) & c >= corThreshold] * 10)
        for (ae in Rgraphviz::AgEdge(p)) {
            Rgraphviz::lines(ae, lwd = lw[i], len = 1)
            i <- i + 1
        }
    }
    invisible(p)
}

## Plotting functions for Zipf's and Heaps'law contributed by Kurt Hornik

## See http://en.wikipedia.org/wiki/Zipf%27s_law
Zipf_plot <-
function(x, type = "l", ...)
{
    if (inherits(x, "TermDocumentMatrix"))
        x <- t(x)
    y <- log(sort(col_sums(x), decreasing = TRUE))
    x <- log(seq_along(y))
    m <- lm(y ~ x)
    dots <- list(...)
    if (is.null(dots$xlab)) dots$xlab <- "log(rank)"
    if (is.null(dots$ylab)) dots$ylab <- "log(frequency)"
    do.call(plot, c(list(x, y, type = type), dots))
    abline(m)
    ## <NOTE>
    ## Perhaps this should (invisibly) return the fitted linear model
    ## instead of just the coefficients?
    coef(m)
    ## </NOTE>
}
## http://en.wikipedia.org/wiki/Heaps%27_law
## http://en.wikipedia.org/wiki/Text_corpus
## cum_vocabulary_size <-
## function(m)
## {
##     ## Should work in general, but it very slow for large simple triplet
##     ## matrices ...
##     s <- double(nrow(m))
##     v <- double(ncol(m))
##     for(i in seq_along(s)) {
##         v <- pmax(v, c(m[i, ]))
##         s[i] <- sum(v > 0)
##     }
##     s
## }
cum_vocabulary_size <-
function(m)
{
    ## Only works for simple triplet matrices.
    i <- sapply(split(m$i, m$j), min)
    tab <- table(i)
    v <- double(nrow(m))
    v[as.numeric(names(tab))] <- tab
    cumsum(v)
}

Heaps_plot <-
function(x, type = "l", ...)
{
    if (inherits(x, "TermDocumentMatrix"))
        x <- t(x)
    y <- log(cum_vocabulary_size(x))
    x <- log(cumsum(row_sums(x)))
    m <- lm(y ~ x)
    dots <- list(...)
    if (is.null(dots$xlab)) dots$xlab <- "log(T)"
    if (is.null(dots$ylab)) dots$ylab <- "log(V)"
    do.call(plot, c(list(x, y, type = type), dots))
    abline(m)
    ## <NOTE>
    ## Perhaps this should (invisibly) return the fitted linear model
    ## instead of just the coefficients?
    coef(m)
    ## </NOTE>
}

# Author: Ingo Feinerer
# Filters

tm_filter <-
function(x, FUN, ...)
    UseMethod("tm_filter", x)
tm_filter.PCorpus <-
tm_filter.SimpleCorpus <-
tm_filter.VCorpus <-
function(x, FUN, ...)
    x[tm_index(x, FUN, ...)]

tm_index <-
function(x, FUN, ...)
    UseMethod("tm_index", x)
tm_index.PCorpus <-
tm_index.SimpleCorpus <-
tm_index.VCorpus <-
function(x, FUN, ...)
    unlist(tm_parLapply(content(x), function(y) isTRUE(FUN(y, ...))))

c.TextDocument <-
function(..., recursive = FALSE)
{
    args <- list(...)
    x <- args[[1L]]

    if (length(args) == 1L)
        return(x)

    if (!all(unlist(lapply(args, inherits, class(x)))))
        stop("not all arguments are text documents")

    v <- list(content = args,
              meta = CorpusMeta(),
              dmeta = data.frame(row.names = seq_along(args)))
    class(v) <- c("VCorpus", "Corpus")
    v
}

.format_TextDocument <-
function(x, ...)
    c(sprintf("<<%s>>", class(x)[1L]),
      sprintf("Metadata:  %d", length(meta(x))))

inspect.TextDocument <-
function(x)
{
    print(x)
    cat("\n")
    writeLines(as.character(x))
    invisible(x)
}

PlainTextDocument <-
function(x = character(0),
         author = character(0),
         datetimestamp = as.POSIXlt(Sys.time(), tz = "GMT"),
         description = character(0),
         heading = character(0),
         id = character(0),
         language = character(0),
         origin = character(0),
         ...,
         meta = NULL,
         class = NULL)
{
    p <- list(content = as.character(x),
              meta = TextDocumentMeta(author, datetimestamp, description,
                                      heading, id, language, origin, ...,
                                      meta = meta))
    class(p) <- unique(c(class, "PlainTextDocument", "TextDocument"))
    p
}

as.character.PlainTextDocument <-
function(x, ...)
    content(x)

content.PlainTextDocument <-
function(x)
    x$content

`content<-.PlainTextDocument` <-
function(x, value)
{
    x$content <- as.character(value)
    x
}

format.PlainTextDocument <-
function(x, ...)
    c(.format_TextDocument(x),
      sprintf("Content:  chars: %d", sum(nchar(x$content))))

meta.PlainTextDocument <-
function(x, tag = NULL, ...)
    if (is.null(tag)) x$meta else x$meta[[tag]]

`meta<-.PlainTextDocument` <-
function(x, tag = NULL, ..., value)
{
    if (is.null(tag))
        x$meta <- value
    else
        x$meta[[tag]] <- value
    x
}

words.character <-
words.PlainTextDocument <-
function(x, ...)
    scan_tokenizer(x)

XMLTextDocument <-
function(x = xml_missing(),
         author = character(0),
         datetimestamp = as.POSIXlt(Sys.time(), tz = "GMT"),
         description = character(0),
         heading = character(0),
         id = character(0),
         language = character(0),
         origin = character(0),
         ...,
         meta = NULL)
{
    d <- list(content = x,
              meta = TextDocumentMeta(author, datetimestamp, description,
                                      heading, id, language, origin, ...,
                                      meta = meta))
    class(d) <- c("XMLTextDocument", "TextDocument")
    d
}

as.character.XMLTextDocument <-
function(x, ...)
    xml_text(content(x))

content.XMLTextDocument <-
function(x)
    x$content

`content<-.XMLTextDocument` <-
function(x, value)
{
    x$content <- value
    x
}

format.XMLTextDocument <- .format_TextDocument

meta.XMLTextDocument <- meta.PlainTextDocument

`meta<-.XMLTextDocument` <- `meta<-.PlainTextDocument`

tm_parLapply_engine <-
local({
    val <- NULL
    ## Could do some checking on new if given: should inherit from
    ## "cluster" or have formals (X, FUN, ...).
    function(new) {
        if (missing(new)) val else val <<- new
    }
})

tm_parLapply <-
function(X, FUN, ...)
{
    engine <- tm_parLapply_engine()
    if (inherits(engine, "cluster"))
        parLapply(engine, X, FUN, ...)
    else if (is.function(engine))
        engine(X, FUN, ...)
    else
        lapply(X, FUN, ...)
}

# Author: Ingo Feinerer

Corpus <-
function(x, readerControl = list(reader = reader(x), language = "en"))
{
    stopifnot(inherits(x, "Source"))

    readerControl <- prepareReader(readerControl, reader(x))

    if ( (inherits(x, "DataframeSource") || inherits(x, "DirSource") ||
          inherits(x, "VectorSource") ) &&
        identical(readerControl$reader, reader(x)))
        SimpleCorpus(x, readerControl)
    else
        VCorpus(x, readerControl)
}

PCorpus <-
function(x,
         readerControl = list(reader = reader(x), language = "en"),
         dbControl = list(dbName = "", dbType = "DB1"))
{
    stopifnot(inherits(x, "Source"))

    readerControl <- prepareReader(readerControl, reader(x))

    if (!filehash::dbCreate(dbControl$dbName, dbControl$dbType))
        stop("error in creating database")
    db <- filehash::dbInit(dbControl$dbName, dbControl$dbType)

    x <- open(x)
    tdl <- vector("list", length(x))
    counter <- 1
    while (!eoi(x)) {
        x <- stepNext(x)
        elem <- getElem(x)
        doc <- readerControl$reader(elem,
                                    readerControl$language,
                                    as.character(counter))
        filehash::dbInsert(db, meta(doc, "id"), doc)
        tdl[[counter]] <- meta(doc, "id")
        counter <- counter + 1
    }
    x <- close(x)

    cmeta <- CorpusMeta()
    dmeta <- data.frame(row.names = seq_along(tdl))
    ## Check if metadata retrieval is supported
    for(cl in class(x)) {
        if (is.function(getS3method("getMeta", cl, TRUE))) {
            m <- getMeta(x)
            if (!is.null(m$cmeta)) cmeta <- m$cmeta
            if (!is.null(m$dmeta)) dmeta <- m$dmeta
            break
        }
    }

    p <- list(content = tdl, meta = cmeta, dmeta = dmeta, dbcontrol = dbControl)
    class(p) <- c("PCorpus", "Corpus")
    p
}

SimpleCorpus <-
function(x, control = list(language = "en"))
{
    stopifnot(inherits(x, "Source"))

    if (!is.null(control$reader) && !identical(control$reader, reader(x)))
        warning("custom reader is ignored")

    content <- if (inherits(x, "VectorSource")) {
        if (is.character(x$content)) x$content else as.character(x$content)
    } else if (inherits(x, "DirSource")) {
        setNames(as.character(
                   lapply(x$filelist,
                          function(f) paste(readContent(f, x$encoding, "text"),
                                            collapse = "\n"))
                   ),
                 basename(x$filelist))
    } else if (inherits(x, "DataframeSource")) {
        setNames(as.character(x$content[, "text"]), x$content[, "doc_id"])
    } else
        stop("unsupported source type")

    dmeta <- if (inherits(x, "DataframeSource"))
        x$content[, is.na(match(names(x$content),
                                c("doc_id", "text"))),
                  drop = FALSE]
    else
        data.frame(row.names = seq_along(x))

    s <- list(content = content,
              meta = CorpusMeta(language = control$language),
              dmeta = dmeta)
    class(s) <- c("SimpleCorpus", "Corpus")
    s
}

VCorpus <-
function(x, readerControl = list(reader = reader(x), language = "en"))
{
    stopifnot(inherits(x, "Source"))

    readerControl <- prepareReader(readerControl, reader(x))

    x <- open(x)
    tdl <- vector("list", length(x))
    ## Check for parallel element access
    found <- FALSE
    for(cl in class(x)) {
        if (is.function(getS3method("pGetElem", cl, TRUE))) {
            tdl <- mapply(function(elem, id)
                              readerControl$reader(elem,
                                                   readerControl$language,
                                                   id),
                          pGetElem(x),
                          id = as.character(seq_along(x)),
                          SIMPLIFY = FALSE)
            found <- TRUE
            break
        }
    }
    if(!found) {
        counter <- 1
        while (!eoi(x)) {
            x <- stepNext(x)
            elem <- getElem(x)
            doc <- readerControl$reader(elem,
                                        readerControl$language,
                                        as.character(counter))
            tdl[[counter]] <- doc
            counter <- counter + 1
        }
    }
    x <- close(x)

    cmeta <- CorpusMeta()
    dmeta <- data.frame(row.names = seq_along(tdl))
    ## Check if metadata retrieval is supported
    for(cl in class(x)) {
        if (is.function(getS3method("getMeta", cl, TRUE))) {
            m <- getMeta(x)
            if (!is.null(m$cmeta)) cmeta <- m$cmeta
            if (!is.null(m$dmeta)) dmeta <- m$dmeta
            break
        }
    }

    v <- as.VCorpus(tdl)
    v$meta <- cmeta
    v$dmeta <- dmeta

    v
}

`[.PCorpus` <-
`[.SimpleCorpus` <-
function(x, i)
{
    if (!missing(i)) {
        x$content <- x$content[i]
        x$dmeta <- x$dmeta[i, , drop = FALSE]
    }
    x
}
`[.VCorpus` <-
function(x, i)
{
    if (!missing(i)) {
        x$content <- x$content[i]
        x$dmeta <- x$dmeta[i, , drop = FALSE]
        if (!is.null(x$lazy))
            x$lazy$index <- x$lazy$index[i]
    }
    x
}

.map_name_index <-
function(x, i)
{
    if (is.character(i))
        match(i, meta(x, "id", "local"))
    else
        i
}

`[[.PCorpus` <-
function(x, i)
{
    i <- .map_name_index(x, i)
    db <- filehash::dbInit(x$dbcontrol[["dbName"]], x$dbcontrol[["dbType"]])
    filehash::dbFetch(db, x$content[[i]])
}
`[[.SimpleCorpus` <-
function(x, i)
{
    i <- .map_name_index(x, i)
    n <- names(x$content)
    PlainTextDocument(x$content[[i]],
                      id = if (is.null(n)) i else n[i],
                      language = meta(x, "language"))
}
`[[.VCorpus` <-
function(x, i)
{
    i <- .map_name_index(x, i)
    if (!is.null(x$lazy))
        .Call(`_tm_copyCorpus`, x, materialize(x, i))
    x$content[[i]]
}

`[[<-.SimpleCorpus` <-
function(x, i, value)
{
    x$content[i] <- paste0(as.character(value), collapse = "\n")
    x
}
`[[<-.PCorpus` <-
function(x, i, value)
{
    i <- .map_name_index(x, i)
    db <- filehash::dbInit(x$dbcontrol[["dbName"]], x$dbcontrol[["dbType"]])
    db[[x$content[[i]]]] <- value
    x
}
`[[<-.VCorpus` <-
function(x, i, value)
{
    i <- .map_name_index(x, i)
    # Mark new objects as inactive for lazy mapping
    if (!is.null(x$lazy))
        x$lazy$index[i] <- FALSE
    x$content[[i]] <- value
    x
}

as.list.PCorpus <- as.list.VCorpus <-
function(x, ...)
    setNames(content(x), as.character(lapply(content(x), meta, "id")))

as.list.SimpleCorpus <-
function(x, ...)
    as.list(content(x))

as.VCorpus <-
function(x)
    UseMethod("as.VCorpus")
as.VCorpus.VCorpus <- identity
as.VCorpus.list <-
function(x)
{
    v <- list(content = x,
              meta = CorpusMeta(),
              dmeta = data.frame(row.names = seq_along(x)))
    class(v) <- c("VCorpus", "Corpus")
    v
}

outer_union <-
function(x, y, ...)
{
    if (nrow(x) > 0L)
        x[, setdiff(names(y), names(x))] <- NA
    if (nrow(y) > 0L)
        y[, setdiff(names(x), names(y))] <- NA
    res <- rbind(x, y)
    if (ncol(res) == 0L)
        res <- data.frame(row.names = seq_len(nrow(x) + nrow(y)))
    res
}

c.VCorpus <-
function(..., recursive = FALSE)
{
    args <- list(...)
    x <- args[[1L]]

    if (length(args) == 1L)
        return(x)

    if (!all(unlist(lapply(args, inherits, class(x)))))
        stop("not all arguments are of the same corpus type")

    v <- list(content = do.call("c", lapply(args, content)),
              meta = CorpusMeta(meta = do.call("c",
                lapply(args, function(a) meta(a, type = "corpus")))),
              dmeta = Reduce(outer_union, lapply(args, meta)))
    class(v) <- c("VCorpus", "Corpus")
    v
}

content.VCorpus <-
function(x)
{
    if (!is.null(x$lazy))
        .Call(`_tm_copyCorpus`, x, materialize(x))
    x$content
}

content.SimpleCorpus <-
function(x)
    x$content

content.PCorpus <-
function(x)
{
    db <- filehash::dbInit(x$dbcontrol[["dbName"]], x$dbcontrol[["dbType"]])
    filehash::dbMultiFetch(db, unlist(x$content))
}

inspect <-
function(x)
    UseMethod("inspect", x)
inspect.PCorpus <-
inspect.SimpleCorpus <-
inspect.VCorpus <-
function(x)
{
    print(x)
    cat("\n")
    print(noquote(content(x)))
    invisible(x)
}

length.PCorpus <-
length.SimpleCorpus <-
length.VCorpus <-
function(x)
    length(x$content)

names.PCorpus <-
names.SimpleCorpus <-
names.VCorpus <-
function(x)
    as.character(meta(x, "id", "local"))

`names<-.PCorpus` <- `names<-.VCorpus` <-
function(x, value)
{
    meta(x, "id", "local") <- as.character(value)
    x
}

format.PCorpus <-
format.SimpleCorpus <-
format.VCorpus <-
function(x, ...)
{
    c(sprintf("<<%s>>", class(x)[1L]),
      sprintf("Metadata:  corpus specific: %d, document level (indexed): %d",
              length(meta(x, type = "corpus")),
              ncol(meta(x, type = "indexed"))),
      sprintf("Content:  documents: %d", length(x)))
}

writeCorpus <-
function(x, path = ".", filenames = NULL)
{
    filenames <- file.path(path,
      if (is.null(filenames))
          sprintf("%s.txt", as.character(meta(x, "id", "local")))
      else filenames)

    stopifnot(length(x) == length(filenames))

    mapply(function(doc, f) writeLines(as.character(doc), f), x, filenames)

    invisible(x)
}

context("Sources")

test_that("DataframeSource works", {
  txt <- c("First document.", "Second document.")
  dm1 <- 1:2
  dm2 <- letters[1:2]
  df <- data.frame(doc_id = c("doc_1", "doc_2"), text = txt,
                   dmeta1 = dm1, dmeta2 = dm2, stringsAsFactors = FALSE)
  ds <- DataframeSource(df)
  scorpus <- Corpus(ds)
  vcorpus <- VCorpus(ds)
  expect_equal(as.character(scorpus[[2]]), as.character(vcorpus[[2]]))
  expect_equal(as.character(scorpus[[2]]), txt[2])
  expect_equal(meta(scorpus), meta(vcorpus))
  expect_equal(meta(scorpus),
               data.frame(dmeta1 = dm1, dmeta2 = dm2, stringsAsFactors = FALSE))
})

context("Tokenizers")

test_that("scan_tokenizer works with character vectors", {
  tokens <-
      c("a", "character", "vector", "consisting", "of", "multiple", "elements")
  expect_equal(scan_tokenizer(c(paste0(tokens[1:3], collapse = " "),
                                paste0(tokens[4:5], collapse = " "),
                                paste0(tokens[6:7], collapse = " "))), tokens)
})

\documentclass[a4paper]{article}

\usepackage[margin=2cm]{geometry}
\usepackage[utf8]{inputenc}
\usepackage[round]{natbib}
\usepackage{url}

\newcommand{\acronym}[1]{\textsc{#1}}
\newcommand{\class}[1]{\mbox{\textsf{#1}}}
\newcommand{\code}[1]{\mbox{\texttt{#1}}}
\newcommand{\pkg}[1]{{\normalfont\fontseries{b}\selectfont #1}}
\newcommand{\proglang}[1]{\textsf{#1}}

%% \VignetteIndexEntry{Introduction to the tm Package}

\begin{document}
<<Init,echo=FALSE,results=hide>>=
library("tm")
data("crude")
@
\title{Introduction to the \pkg{tm} Package\\Text Mining in \proglang{R}}
\author{Ingo Feinerer}
\maketitle

\section*{Introduction}
This vignette gives a short introduction to text mining in
\proglang{R} utilizing the text mining framework provided by the
\pkg{tm} package. We present methods for data import, corpus handling,
preprocessing, metadata management, and creation of term-document
matrices. Our focus is on the main aspects of getting started with
text mining in \proglang{R}---an in-depth description of the text
mining infrastructure offered by \pkg{tm} was published in the
\emph{Journal of Statistical Software}~\citep{Feinerer_etal_2008}. An
introductory article on text mining in \proglang{R} was published in
\emph{R News}~\citep{Rnews:Feinerer:2008}.

\section*{Data Import}
The main structure for managing documents in \pkg{tm} is a so-called
\class{Corpus}, representing a collection of text documents. A corpus
is an abstract concept, and there can exist several implementations in
parallel. The default implementation is the so-called \class{VCorpus}
(short for \emph{Volatile Corpus}) which realizes a semantics as known
from most \proglang{R} objects: corpora are \proglang{R} objects held
fully in memory. We denote this as volatile since once the
\proglang{R} object is destroyed, the whole corpus is gone. Such a
volatile corpus can be created via the constructor \code{VCorpus(x,
  readerControl)}. Another implementation is the \class{PCorpus} which
implements a \emph{Permanent Corpus} semantics, i.e., the documents
are physically stored outside of \proglang{R} (e.g., in a database),
corresponding \proglang{R} objects are basically only pointers to
external structures, and changes to the underlying corpus are
reflected to all \proglang{R} objects associated with it. Compared to
the volatile corpus the corpus encapsulated by a permanent corpus
object is not destroyed if the corresponding \proglang{R} object is
released.

Within the corpus constructor, \code{x} must be a \class{Source}
object which abstracts the input location. \pkg{tm} provides a set of
predefined sources, e.g., \class{DirSource}, \class{VectorSource}, or
\class{DataframeSource}, which handle a directory, a vector
interpreting each component as document, or data frame like structures
(like \acronym{CSV} files), respectively. Except \class{DirSource},
which is designed solely for directories on a file system, and
\class{VectorSource}, which only accepts (character) vectors, most
other implemented sources can take connections as input (a character
string is interpreted as file path). \code{getSources()} lists
available sources, and users can create their own sources.

The second argument \code{readerControl} of the corpus constructor has to be a
list with the named components \code{reader} and \code{language}. The first
component \code{reader} constructs a text document from elements delivered by a
source. The \pkg{tm} package ships with several readers (e.g.,
\code{readPlain()}, \code{readPDF()}, \code{readDOC()}, \ldots). See
\code{getReaders()} for an up-to-date list of available readers. Each source
has a default reader which can be overridden. E.g., for \code{DirSource} the
default just reads in the input files and interprets their content as text.
Finally, the second component \code{language} sets the texts' language
(preferably using \acronym{ISO} 639-2 codes).

In case of a permanent corpus, a third argument \code{dbControl} has
to be a list with the named components \code{dbName} giving the
filename holding the sourced out objects (i.e., the database), and
\code{dbType} holding a valid database type as supported by package
\pkg{filehash}. Activated database support reduces the memory demand,
however, access gets slower since each operation is limited by the
hard disk's read and write capabilities.

So e.g., plain text files in the directory \code{txt} containing Latin
(\code{lat}) texts by the Roman poet \emph{Ovid} can be read in with
following code:
<<Ovid,keep.source=TRUE>>=
txt <- system.file("texts", "txt", package = "tm")
(ovid <- VCorpus(DirSource(txt, encoding = "UTF-8"),
                 readerControl = list(language = "lat")))
@

For simple examples \code{VectorSource} is quite useful, as it can
create a corpus from character vectors, e.g.:
<<VectorSource,keep.source=TRUE>>=
docs <- c("This is a text.", "This another one.")
VCorpus(VectorSource(docs))
@

Finally we create a corpus for some Reuters documents as example for
later use:
<<Reuters,keep.source=TRUE>>=
reut21578 <- system.file("texts", "crude", package = "tm")
reuters <- VCorpus(DirSource(reut21578, mode = "binary"),
                   readerControl = list(reader = readReut21578XMLasPlain))
@

\section*{Data Export}
For the case you have created a corpus via manipulating other
objects in \proglang{R}, thus do not have the texts already stored on
a hard disk, and want to save the text documents to disk, you can
simply use \code{writeCorpus()}
<<eval=FALSE,keep.source=TRUE>>=
writeCorpus(ovid)
@
which writes a character representation of the documents in a corpus to
multiple files on disk.

\section*{Inspecting Corpora}
Custom \code{print()} methods are available which hide the raw amount of
information (consider a corpus could consist of several thousand documents,
like a database). \code{print()} gives a concise overview whereas more details
are displayed with \code{inspect()}.
<<>>=
inspect(ovid[1:2])
@
Individual documents can be accessed via \code{[[}, either via the
position in the corpus, or via their identifier.
<<keep.source=TRUE>>=
meta(ovid[[2]], "id")
identical(ovid[[2]], ovid[["ovid_2.txt"]])
@
A character representation of a document is available via
\code{as.character()} which is also used when inspecting a document:
<<keep.source=TRUE>>=
inspect(ovid[[2]])
lapply(ovid[1:2], as.character)
@

\section*{Transformations}
Once we have a corpus we typically want to modify the documents in it,
e.g., stemming, stopword removal, et cetera. In \pkg{tm}, all this
functionality is subsumed into the concept of a
\emph{transformation}. Transformations are done via the \code{tm\_map()}
function which applies (maps) a function to all elements of the
corpus. Basically, all transformations work on single text documents
and \code{tm\_map()} just applies them to all documents in a corpus.

\subsection*{Eliminating Extra Whitespace}
Extra whitespace is eliminated by:
<<>>=
reuters <- tm_map(reuters, stripWhitespace)
@

\subsection*{Convert to Lower Case}
Conversion to lower case by:
<<>>=
reuters <- tm_map(reuters, content_transformer(tolower))
@
We can use arbitrary character processing functions as transformations as long
as the function returns a text document. In this case we use
\code{content\_transformer()} which provides a convenience wrapper to access and
set the content of a document. Consequently most text manipulation functions
from base \proglang{R} can directly be used with this wrapper. This works for
\code{tolower()} as used here but also e.g.\ for \code{gsub()} which comes quite
handy for a broad range of text manipulation tasks.

\subsection*{Remove Stopwords}
Removal of stopwords by:
<<Stopwords>>=
reuters <- tm_map(reuters, removeWords, stopwords("english"))
@

\subsection*{Stemming}
Stemming is done by:
<<Stemming>>=
tm_map(reuters, stemDocument)
@

\section*{Filters}
Often it is of special interest to filter out documents satisfying
given properties. For this purpose the function \code{tm\_filter} is
designed. It is possible to write custom filter functions which get applied to
each document in the corpus. Alternatively, we can create indices based on
selections and subset the corpus with them. E.g., the following
statement filters out those documents having an \code{ID} equal to \code{"237"}
and the string \code{"INDONESIA SEEN AT CROSSROADS OVER ECONOMIC CHANGE"} as
their heading.
<<>>=
idx <- meta(reuters, "id") == '237' &
  meta(reuters, "heading") == 'INDONESIA SEEN AT CROSSROADS OVER ECONOMIC CHANGE'
reuters[idx]
@

\section*{Metadata Management}
Metadata is used to annotate text documents or whole corpora with
additional information. The easiest way to accomplish this with
\pkg{tm} is to use the \code{meta()} function. A text document has a
few predefined attributes like \code{author} but can be extended with
an arbitrary number of additional user-defined metadata tags. These
additional metadata tags are individually attached to a single text
document. From a corpus perspective these metadata attachments are
locally stored together with each individual text
document. Alternatively to \code{meta()} the function
\code{DublinCore()} provides a full mapping between Simple Dublin Core
metadata and \pkg{tm} metadata structures and can be similarly used
to get and set metadata information for text documents, e.g.:
<<DublinCore>>=
DublinCore(crude[[1]], "Creator") <- "Ano Nymous"
meta(crude[[1]])
@

For corpora the story is a bit more sophisticated. Corpora in \pkg{tm}
have two types of metadata: one is the metadata on the corpus level
(\code{corpus}), the other is the metadata related to the individual
documents (\code{indexed}) in form of a data frame. The latter is
often done for performance reasons (hence the named \code{indexed} for
indexing) or because the metadata has an own entity but still relates
directly to individual text documents, e.g., a classification result;
the classifications directly relate to the documents but the set of
classification levels forms an own entity. Both cases can be handled
with \code{meta()}:
<<>>=
meta(crude, tag = "test", type = "corpus") <- "test meta"
meta(crude, type = "corpus")
meta(crude, "foo") <- letters[1:20]
meta(crude)
@

\section*{Standard Operators and Functions}
Many standard operators and functions (\code{[}, \code{[<-},
\code{[[}, \code{[[<-}, \code{c()}, \code{lapply()}) are available for
corpora with semantics similar to standard \proglang{R}
routines. E.g., \code{c()} concatenates two (or more) corpora. Applied
to several text documents it returns a corpus. The metadata is
automatically updated, if corpora are concatenated (i.e., merged).

\section*{Creating Term-Document Matrices}
A common approach in text mining is to create a term-document matrix
from a corpus. In the \pkg{tm} package the classes
\class{TermDocumentMatrix} and \class{DocumentTermMatrix} (depending
on whether you want terms as rows and documents as columns, or vice
versa) employ sparse matrices for corpora. Inspecting a term-document matrix
displays a sample, whereas \code{as.matrix()} yields the full matrix in
dense format (which can be very memory consuming for large matrices).
<<>>=
dtm <- DocumentTermMatrix(reuters)
inspect(dtm)
@

\section*{Operations on Term-Document Matrices}
Besides the fact that on this matrix a huge amount of \proglang{R}
functions (like clustering, classifications, etc.) can be applied,
this package brings some shortcuts. Imagine we want to find those
terms that occur at least five times, then we can use the
\code{findFreqTerms()} function:
<<>>=
findFreqTerms(dtm, 5)
@
Or we want to find associations (i.e., terms which correlate) with at
least $0.8$ correlation for the term \code{opec}, then we use
\code{findAssocs()}:
<<>>=
findAssocs(dtm, "opec", 0.8)
@

Term-document matrices tend to get very big already for normal sized
data sets. Therefore we provide a method to remove \emph{sparse} terms,
i.e., terms occurring only in very few documents. Normally, this
reduces the matrix dramatically without losing significant relations
inherent to the matrix:
<<>>=
inspect(removeSparseTerms(dtm, 0.4))
@
This function call removes those terms which have at least a 40
percentage of sparse (i.e., terms occurring 0 times in a document)
elements.

\section*{Dictionary}
A dictionary is a (multi-)set of strings. It is often used to denote relevant
terms in text mining. We represent a dictionary with a character vector which
may be passed to the \code{DocumentTermMatrix()} constructor as a control
argument. Then the created matrix is tabulated against the dictionary, i.e.,
only terms from the dictionary appear in the matrix. This allows to restrict
the dimension of the matrix a priori and to focus on specific terms for
distinct text mining contexts, e.g.,
<<>>=
inspect(DocumentTermMatrix(reuters,
                           list(dictionary = c("prices", "crude", "oil"))))
@

\section*{Performance}
Often you do not need all the generality, modularity and full range of features
offered by \pkg{tm} as this sometimes comes at the price of performance.

\class{SimpleCorpus} provides a corpus which is optimized for the most common
usage scenario: importing plain texts from files in a directory or directly
from a vector in \proglang{R}, preprocessing and transforming the texts, and
finally exporting them to a term-document matrix. The aim is to boost
performance and minimize memory pressure. It loads all documents into memory,
and is designed for medium-sized to large data sets.

However, it operates only under the following contraints:
\begin{itemize}
  \item only \code{DirSource} and \code{VectorSource} are supported,
  \item no custom readers, i.e., each document is read in and stored as plain
    text (as a string, i.e., a character vector of length one),
  \item transformations applied via \code{tm\_map} must be able to
    process strings and return strings,
  \item no lazy transformations in \code{tm\_map},
  \item no meta data for individual documents (i.e., no \code{"local"} in
    \code{meta()}).
\end{itemize}

\bibliographystyle{abbrvnat}
\bibliography{references}

\end{document}

\documentclass[a4paper]{article}

\usepackage[margin=2cm]{geometry}
\usepackage[utf8]{inputenc}
\usepackage[round]{natbib}
\usepackage{url}

\newcommand{\acronym}[1]{\textsc{#1}}
\newcommand{\class}[1]{\mbox{\textsf{#1}}}
\newcommand{\code}[1]{\mbox{\texttt{#1}}}
\newcommand{\pkg}[1]{{\normalfont\fontseries{b}\selectfont #1}}
\newcommand{\proglang}[1]{\textsf{#1}}

%% \VignetteIndexEntry{Introduction to the tm Package}
%% \VignetteDepends{SnowballC}

\begin{document}
<<Init,echo=FALSE,results=hide>>=
library("tm")
data("crude")
@
\title{Introduction to the \pkg{tm} Package\\Text Mining in \proglang{R}}
\author{Ingo Feinerer}
\maketitle

\section*{Introduction}
This vignette gives a short introduction to text mining in
\proglang{R} utilizing the text mining framework provided by the
\pkg{tm} package. We present methods for data import, corpus handling,
preprocessing, metadata management, and creation of term-document
matrices. Our focus is on the main aspects of getting started with
text mining in \proglang{R}---an in-depth description of the text
mining infrastructure offered by \pkg{tm} was published in the
\emph{Journal of Statistical Software}~\citep{Feinerer_etal_2008}. An
introductory article on text mining in \proglang{R} was published in
\emph{R News}~\citep{Rnews:Feinerer:2008}.

\section*{Data Import}
The main structure for managing documents in \pkg{tm} is a so-called
\class{Corpus}, representing a collection of text documents. A corpus
is an abstract concept, and there can exist several implementations in
parallel. The default implementation is the so-called \class{VCorpus}
(short for \emph{Volatile Corpus}) which realizes a semantics as known
from most \proglang{R} objects: corpora are \proglang{R} objects held
fully in memory. We denote this as volatile since once the
\proglang{R} object is destroyed, the whole corpus is gone. Such a
volatile corpus can be created via the constructor \code{VCorpus(x,
  readerControl)}. Another implementation is the \class{PCorpus} which
implements a \emph{Permanent Corpus} semantics, i.e., the documents
are physically stored outside of \proglang{R} (e.g., in a database),
corresponding \proglang{R} objects are basically only pointers to
external structures, and changes to the underlying corpus are
reflected to all \proglang{R} objects associated with it. Compared to
the volatile corpus the corpus encapsulated by a permanent corpus
object is not destroyed if the corresponding \proglang{R} object is
released.

Within the corpus constructor, \code{x} must be a \class{Source}
object which abstracts the input location. \pkg{tm} provides a set of
predefined sources, e.g., \class{DirSource}, \class{VectorSource}, or
\class{DataframeSource}, which handle a directory, a vector
interpreting each component as document, or data frame like structures
(like \acronym{CSV} files), respectively. Except \class{DirSource},
which is designed solely for directories on a file system, and
\class{VectorSource}, which only accepts (character) vectors, most
other implemented sources can take connections as input (a character
string is interpreted as file path). \code{getSources()} lists
available sources, and users can create their own sources.

The second argument \code{readerControl} of the corpus constructor has to be a
list with the named components \code{reader} and \code{language}. The first
component \code{reader} constructs a text document from elements delivered by a
source. The \pkg{tm} package ships with several readers (e.g.,
\code{readPlain()}, \code{readPDF()}, \code{readDOC()}, \ldots). See
\code{getReaders()} for an up-to-date list of available readers. Each source
has a default reader which can be overridden. E.g., for \code{DirSource} the
default just reads in the input files and interprets their content as text.
Finally, the second component \code{language} sets the texts' language
(preferably using \acronym{ISO} 639-2 codes).

In case of a permanent corpus, a third argument \code{dbControl} has
to be a list with the named components \code{dbName} giving the
filename holding the sourced out objects (i.e., the database), and
\code{dbType} holding a valid database type as supported by package
\pkg{filehash}. Activated database support reduces the memory demand,
however, access gets slower since each operation is limited by the
hard disk's read and write capabilities.

So e.g., plain text files in the directory \code{txt} containing Latin
(\code{lat}) texts by the Roman poet \emph{Ovid} can be read in with
following code:
<<Ovid,keep.source=TRUE>>=
txt <- system.file("texts", "txt", package = "tm")
(ovid <- VCorpus(DirSource(txt, encoding = "UTF-8"),
                 readerControl = list(language = "lat")))
@

For simple examples \code{VectorSource} is quite useful, as it can
create a corpus from character vectors, e.g.:
<<VectorSource,keep.source=TRUE>>=
docs <- c("This is a text.", "This another one.")
VCorpus(VectorSource(docs))
@

Finally we create a corpus for some Reuters documents as example for
later use:
<<Reuters,keep.source=TRUE>>=
reut21578 <- system.file("texts", "crude", package = "tm")
reuters <- VCorpus(DirSource(reut21578, mode = "binary"),
                   readerControl = list(reader = readReut21578XMLasPlain))
@

\section*{Data Export}
For the case you have created a corpus via manipulating other
objects in \proglang{R}, thus do not have the texts already stored on
a hard disk, and want to save the text documents to disk, you can
simply use \code{writeCorpus()}
<<eval=FALSE,keep.source=TRUE>>=
writeCorpus(ovid)
@
which writes a character representation of the documents in a corpus to
multiple files on disk.

\section*{Inspecting Corpora}
Custom \code{print()} methods are available which hide the raw amount of
information (consider a corpus could consist of several thousand documents,
like a database). \code{print()} gives a concise overview whereas more details
are displayed with \code{inspect()}.
<<>>=
inspect(ovid[1:2])
@
Individual documents can be accessed via \code{[[}, either via the
position in the corpus, or via their identifier.
<<keep.source=TRUE>>=
meta(ovid[[2]], "id")
identical(ovid[[2]], ovid[["ovid_2.txt"]])
@
A character representation of a document is available via
\code{as.character()} which is also used when inspecting a document:
<<keep.source=TRUE>>=
inspect(ovid[[2]])
lapply(ovid[1:2], as.character)
@

\section*{Transformations}
Once we have a corpus we typically want to modify the documents in it,
e.g., stemming, stopword removal, et cetera. In \pkg{tm}, all this
functionality is subsumed into the concept of a
\emph{transformation}. Transformations are done via the \code{tm\_map()}
function which applies (maps) a function to all elements of the
corpus. Basically, all transformations work on single text documents
and \code{tm\_map()} just applies them to all documents in a corpus.

\subsection*{Eliminating Extra Whitespace}
Extra whitespace is eliminated by:
<<>>=
reuters <- tm_map(reuters, stripWhitespace)
@

\subsection*{Convert to Lower Case}
Conversion to lower case by:
<<>>=
reuters <- tm_map(reuters, content_transformer(tolower))
@
We can use arbitrary character processing functions as transformations as long
as the function returns a text document. In this case we use
\code{content\_transformer()} which provides a convenience wrapper to access and
set the content of a document. Consequently most text manipulation functions
from base \proglang{R} can directly be used with this wrapper. This works for
\code{tolower()} as used here but also e.g.\ for \code{gsub()} which comes quite
handy for a broad range of text manipulation tasks.

\subsection*{Remove Stopwords}
Removal of stopwords by:
<<Stopwords>>=
reuters <- tm_map(reuters, removeWords, stopwords("english"))
@

\subsection*{Stemming}
Stemming is done by:
<<Stemming>>=
tm_map(reuters, stemDocument)
@

\section*{Filters}
Often it is of special interest to filter out documents satisfying
given properties. For this purpose the function \code{tm\_filter} is
designed. It is possible to write custom filter functions which get applied to
each document in the corpus. Alternatively, we can create indices based on
selections and subset the corpus with them. E.g., the following
statement filters out those documents having an \code{ID} equal to \code{"237"}
and the string \code{"INDONESIA SEEN AT CROSSROADS OVER ECONOMIC CHANGE"} as
their heading.
<<>>=
idx <- meta(reuters, "id") == '237' &
  meta(reuters, "heading") == 'INDONESIA SEEN AT CROSSROADS OVER ECONOMIC CHANGE'
reuters[idx]
@

\section*{Metadata Management}
Metadata is used to annotate text documents or whole corpora with
additional information. The easiest way to accomplish this with
\pkg{tm} is to use the \code{meta()} function. A text document has a
few predefined attributes like \code{author} but can be extended with
an arbitrary number of additional user-defined metadata tags. These
additional metadata tags are individually attached to a single text
document. From a corpus perspective these metadata attachments are
locally stored together with each individual text
document. Alternatively to \code{meta()} the function
\code{DublinCore()} provides a full mapping between Simple Dublin Core
metadata and \pkg{tm} metadata structures and can be similarly used
to get and set metadata information for text documents, e.g.:
<<DublinCore>>=
DublinCore(crude[[1]], "Creator") <- "Ano Nymous"
meta(crude[[1]])
@

For corpora the story is a bit more sophisticated. Corpora in \pkg{tm}
have two types of metadata: one is the metadata on the corpus level
(\code{corpus}), the other is the metadata related to the individual
documents (\code{indexed}) in form of a data frame. The latter is
often done for performance reasons (hence the named \code{indexed} for
indexing) or because the metadata has an own entity but still relates
directly to individual text documents, e.g., a classification result;
the classifications directly relate to the documents but the set of
classification levels forms an own entity. Both cases can be handled
with \code{meta()}:
<<>>=
meta(crude, tag = "test", type = "corpus") <- "test meta"
meta(crude, type = "corpus")
meta(crude, "foo") <- letters[1:20]
meta(crude)
@

\section*{Standard Operators and Functions}
Many standard operators and functions (\code{[}, \code{[<-},
\code{[[}, \code{[[<-}, \code{c()}, \code{lapply()}) are available for
corpora with semantics similar to standard \proglang{R}
routines. E.g., \code{c()} concatenates two (or more) corpora. Applied
to several text documents it returns a corpus. The metadata is
automatically updated, if corpora are concatenated (i.e., merged).

\section*{Creating Term-Document Matrices}
A common approach in text mining is to create a term-document matrix
from a corpus. In the \pkg{tm} package the classes
\class{TermDocumentMatrix} and \class{DocumentTermMatrix} (depending
on whether you want terms as rows and documents as columns, or vice
versa) employ sparse matrices for corpora. Inspecting a term-document matrix
displays a sample, whereas \code{as.matrix()} yields the full matrix in
dense format (which can be very memory consuming for large matrices).
<<>>=
dtm <- DocumentTermMatrix(reuters)
inspect(dtm)
@

\section*{Operations on Term-Document Matrices}
Besides the fact that on this matrix a huge amount of \proglang{R}
functions (like clustering, classifications, etc.) can be applied,
this package brings some shortcuts. Imagine we want to find those
terms that occur at least five times, then we can use the
\code{findFreqTerms()} function:
<<>>=
findFreqTerms(dtm, 5)
@
Or we want to find associations (i.e., terms which correlate) with at
least $0.8$ correlation for the term \code{opec}, then we use
\code{findAssocs()}:
<<>>=
findAssocs(dtm, "opec", 0.8)
@

Term-document matrices tend to get very big already for normal sized
data sets. Therefore we provide a method to remove \emph{sparse} terms,
i.e., terms occurring only in very few documents. Normally, this
reduces the matrix dramatically without losing significant relations
inherent to the matrix:
<<>>=
inspect(removeSparseTerms(dtm, 0.4))
@
This function call removes those terms which have at least a 40
percentage of sparse (i.e., terms occurring 0 times in a document)
elements.

\section*{Dictionary}
A dictionary is a (multi-)set of strings. It is often used to denote relevant
terms in text mining. We represent a dictionary with a character vector which
may be passed to the \code{DocumentTermMatrix()} constructor as a control
argument. Then the created matrix is tabulated against the dictionary, i.e.,
only terms from the dictionary appear in the matrix. This allows to restrict
the dimension of the matrix a priori and to focus on specific terms for
distinct text mining contexts, e.g.,
<<>>=
inspect(DocumentTermMatrix(reuters,
                           list(dictionary = c("prices", "crude", "oil"))))
@

\section*{Performance}
Often you do not need all the generality, modularity and full range of features
offered by \pkg{tm} as this sometimes comes at the price of performance.

\class{SimpleCorpus} provides a corpus which is optimized for the most common
usage scenario: importing plain texts from files in a directory or directly
from a vector in \proglang{R}, preprocessing and transforming the texts, and
finally exporting them to a term-document matrix. The aim is to boost
performance and minimize memory pressure. It loads all documents into memory,
and is designed for medium-sized to large data sets.

However, it operates only under the following contraints:
\begin{itemize}
  \item only \code{DirSource} and \code{VectorSource} are supported,
  \item no custom readers, i.e., each document is read in and stored as plain
    text (as a string, i.e., a character vector of length one),
  \item transformations applied via \code{tm\_map} must be able to
    process strings and return strings,
  \item no lazy transformations in \code{tm\_map},
  \item no meta data for individual documents (i.e., no \code{"local"} in
    \code{meta()}).
\end{itemize}

\bibliographystyle{abbrvnat}
\bibliography{references}

\end{document}

\documentclass[a4paper]{article}

\usepackage[margin=2cm]{geometry}
\usepackage[round]{natbib}
\usepackage{url}

\newcommand{\acronym}[1]{\textsc{#1}}
\newcommand{\pkg}[1]{{\normalfont\fontseries{b}\selectfont #1}}
\newcommand{\proglang}[1]{\textsf{#1}}
\let\code\texttt

%% \VignetteIndexEntry{Extensions}

\begin{document}
<<Init,echo=FALSE,results=hide>>=
library("tm")
library("xml2")
@
\title{Extensions\\How to Handle Custom File Formats}
\author{Ingo Feinerer}
\maketitle

\section*{Introduction}
The possibility to handle custom file formats is a substantial feature in any
modern text mining infrastructure. \pkg{tm} has been designed aware of this
aspect from the beginning on, and has modular components which allow for
extensions. A general explanation of \pkg{tm}'s extension mechanism is
described by~\citet[Sec.~3.3]{Feinerer_etal_2008}, with an updated description
as follows.

\section*{Sources}
A source abstracts input locations and provides uniform methods for access.
Each source must provide implementations for following interface functions:
\begin{description}
\item[close()] closes the source and returns it,
\item[eoi()] returns \code{TRUE} if the end of input of the source is reached,
\item[getElem()] fetches the element at the current position,
\item[length()] gives the number of elements,
\item[open()] opens the source and returns it,
\item[reader()] returns a default reader for processing elements,
\item[pGetElem()] (optional) retrieves all elements in parallel at once, and
\item[stepNext()] increases the position in the source to the next element.
\end{description}
Retrieved elements must be encapsulated in a list with the named components
\code{content} holding the document and \code{uri} pointing to the origin of
the document (e.g., a file path or a \acronym{URL}; \code{NULL} if not
applicable or unavailable).

Custom sources are required to inherit from the virtual base class
\code{Source} and typically do so by extending the functionality provided by the
simple reference implementation \code{SimpleSource}.

E.g., a simple source which accepts an \proglang{R} vector as input could be
defined as
<<keep.source=TRUE>>=
VecSource <- function(x)
    SimpleSource(length = length(x), content = as.character(x),
                 class = "VecSource")
@
which overrides a few defaults (see \code{?SimpleSource} for defaults) and
stores the vector in the \code{content} component. The functions
\code{close()}, \code{eoi()}, \code{open()}, and \code{stepNext()} have
reasonable default methods already for the \code{SimpleSource} class: the
identity function for \code{open()} and \code{close()}, incrementing a position
counter for \code{stepNext()}, and comparing the current position with the
number of available elements as claimed by \code{length()} for \code{eoi()},
respectively. So we only need custom methods for element access:
<<keep.source=TRUE>>=
getElem.VecSource <-
function(x) list(content = x$content[x$position], uri = NULL)
pGetElem.VecSource <-
function(x) lapply(x$content, function(y) list(content = y, uri = NULL))
@

\section*{Readers}
Readers are functions for extracting textual content and metadata out of
elements delivered by a source and for constructing a text document. Each reader
must accept following arguments in its signature:
\begin{description}
\item[elem] a list with the named components \code{content} and \code{uri} (as
  delivered by a source via \code{getElem()} or \code{pGetElem()}),
\item[language] a string giving the language, and
\item[id] a character giving a unique identifier for the created text document.
\end{description}
The element \code{elem} is typically provided by a source whereas the language
and the identifier are normally provided by a corpus constructor (for the case
that \code{elem\$content} does not give information on these two essential
items).

In case a reader expects configuration arguments we can use a function
generator. A function generator is indicated by inheriting from class
\code{FunctionGenerator} and \code{function}. It allows us to process
additional arguments, store them in an environment, return a reader function
with the well-defined signature described above, and still be able to access
the additional arguments via lexical scoping. All corpus constructors in
package \pkg{tm} check the reader function for being a function generator and
if so apply it to yield the reader with the expected signature.

E.g., the reader function \code{readPlain()} is defined as
<<keep.source=TRUE>>=
readPlain <- function(elem, language, id)
    PlainTextDocument(elem$content, id = id, language = language)
@
For examples on readers using the function generator please have a look at
\code{?readPDF} or \code{?readPDF}.

However, for many cases, it is not necessary to define each detailed aspect of
how to extend \pkg{tm}. Typical examples are \acronym{XML} files which are very
common but can be rather easily handled via standard conforming \acronym{XML}
parsers. The aim of the remainder in this document is to give an overview on
how simpler, more user-friendly, forms of extension mechanisms can be applied
in \pkg{tm}.

\section*{Custom Data Formats}
A general situation is that you have gathered together some information into a
tabular data structure (like a data frame or a list matrix) that suffices to
describe documents in a corpus. However, you do not have a distinct file format
because you extracted the information out of various resources, e.g., as
delivered by \code{readtext()} in package \pkg{readtext}. Now you want to use
your information to build a corpus which is recognized by \pkg{tm}.

We assume that your information is put together in a data frame. E.g.,
consider the following example:
<<keep.source=TRUE>>=
df <- data.frame(doc_id   = c("doc 1"    , "doc 2"    , "doc 3"    ),
		 text     = c("content 1", "content 2", "content 3"),
                 title    = c("title 1"  , "title 2"  , "title 3"  ),
                 authors  = c("author 1" , "author 2" , "author 3" ),
                 topics   = c("topic 1"  , "topic 2"  , "topic 3"  ),
                 stringsAsFactors = FALSE)
@
We want to map the data frame rows to the relevant entries of a text document.

An entry \code{text} in the mapping will be matched to fill the actual content
of the text document, \code{doc\_id} will be used as document ID, all other
fields will be used as metadata tags. So we can construct a corpus out of the
data frame:
<<>>=
(corpus <- Corpus(DataframeSource(df)))
corpus[[1]]
meta(corpus[[1]])
@

\section*{Custom XML Sources}
Many modern file formats already come in \acronym{XML} format which
allows to extract information with any \acronym{XML} conforming
parser, e.g., as implemented in \proglang{R} by the \pkg{xml2}
package.

Now assume we have some custom \acronym{XML} format which we want to
access with \pkg{tm}. Then a viable way is to create a custom
\acronym{XML} source which can be configured with only a few
commands. E.g., have a look at the following example:
<<CustomXMLFile>>=
custom.xml <- system.file("texts", "custom.xml", package = "tm")
print(readLines(custom.xml), quote = FALSE)
@
As you see there is a top-level tag stating that there is a corpus,
and several document tags below. In fact, this structure is very
common in \acronym{XML} files found in text mining applications (e.g.,
both the Reuters-21578 and the Reuters Corpus Volume 1 data sets follow
this general scheme). In \pkg{tm} we expect a source to deliver
self-contained blocks of information to a reader function, each block
containing all information necessary such that the reader can
construct a (subclass of a) \code{TextDocument} from it.

The \code{XMLSource()} function can now be used to construct a custom
\acronym{XML} source. It has three arguments:
\begin{description}
\item[x] a character giving a uniform resource identifier,
\item[parser] a function accepting an \acronym{XML} document (as delivered by
  \code{read\_xml()} in package \pkg{xml2}) as input and returning a
  \acronym{XML} elements/nodes (each element/node will then be
  delivered to the reader as a self-contained block),
\item[reader] a reader function capable of turning \acronym{XML}
  elements/nodes as returned by the parser into a subclass of
  \code{TextDocument}.
\end{description}
E.g., a custom source which can cope with our custom \acronym{XML}
format could be:
<<mySource,keep.source=TRUE>>=
mySource <- function(x)
    XMLSource(x, parser = xml2::xml_children, reader = myXMLReader)
@
As you notice in this example we also provide a custom reader function
(\code{myXMLReader}). See the next section for details.

\section*{Custom XML Readers}
As we saw in the previous section we often need a custom reader
function to extract information out of \acronym{XML} chunks (typically
as delivered by some source). Fortunately, \pkg{tm} provides an easy
way to define custom \acronym{XML} reader functions. All you need to
do is to provide a so-called \emph{specification}.

Let us start with an example which defines a reader function for the
file format from the previous section:
<<myXMLReader,keep.source=TRUE>>=
myXMLReader <- readXML(
    spec = list(author = list("node", "writer"),
                content = list("node", "description"),
                datetimestamp = list("function",
                    function(x) as.POSIXlt(Sys.time(), tz = "GMT")),
                description = list("node", "@short"),
                heading = list("node", "caption"),
                id = list("function", function(x) tempfile()),
                origin = list("unevaluated", "My private bibliography"),
                type = list("node", "type")),
    doc = PlainTextDocument())
@

Formally, \code{readXML()} is the relevant function which constructs an reader.
The customization is done via the first argument \code{spec}, the second
provides an empty instance of the document which should be returned (augmented
with the extracted information out of the \acronym{XML} chunks). The
specification must consist of a named list of lists each containing two
character vectors. The constructed reader will map each list entry to the
content or a metadatum of the text document as specified by the named list
entry. Valid names include \code{content} to access the document's content, and
character strings which are mapped to metadata entries.

Each list entry must consist of two character vectors: the first
describes the type of the second argument, and the second is the
specification entry. Valid combinations are:
\begin{description}
\item[\code{type = "node", spec = "XPathExpression"}] the XPath (1.0) expression
  \code{spec} extracts information out of an \acronym{XML} node (as seen for
  \code{author}, \code{content}, \code{description}, \code{heading}, and
  \code{type} in our example specification).
\item[\code{type = "function", spec = function(doc) \ldots}] The function
  \code{spec} is called, passing over the \acronym{XML} document (as
  delivered by \code{read\_xml()} from package \pkg{xml2})
  as first argument (as seen for \code{datetimestamp} and \code{id}).
  As you notice in our example nobody forces us to actually use the passed over
  document, instead we can do anything we want (e.g., create a unique character
  vector via \code{tempfile()} to have a unique identification string).
\item[\code{type = "unevaluated", spec = "String"}] the character vector
  \code{spec} is returned without modification (e.g., \code{origin} in
  our specification).
\end{description}

Now that we have all we need to cope with our custom file format, we
can apply the source and reader function at any place in \pkg{tm}
where a source or reader is expected, respectively. E.g.,
<<>>=
corpus <- VCorpus(mySource(custom.xml))
@
constructs a corpus out of the information in our \acronym{XML}
file:
<<>>=
corpus[[1]]
meta(corpus[[1]])
@

\bibliographystyle{abbrvnat}
\bibliography{references}

\end{document}

\name{getTokenizers}
\alias{getTokenizers}
\title{Tokenizers}
\description{
  Predefined tokenizers.
}
\usage{
getTokenizers()
}
\value{
  A character vector with tokenizers provided by package \pkg{tm}.
}
\seealso{
  \code{\link{Boost_tokenizer}}, \code{\link{MC_tokenizer}} and
  \code{\link{scan_tokenizer}}.
}
\examples{
getTokenizers()
}

\name{Zipf_n_Heaps}
\alias{Zipf_plot}
\alias{Heaps_plot}
\title{Explore Corpus Term Frequency Characteristics}
\description{
  Explore Zipf's law and Heaps' law, two empirical laws in linguistics
  describing commonly observed characteristics of term frequency
  distributions in corpora.
}
\usage{
Zipf_plot(x, type = "l", \dots)
Heaps_plot(x, type = "l", \dots)
}
\arguments{
  \item{x}{a document-term matrix or term-document matrix with
    unweighted term frequencies.}
  \item{type}{a character string indicating the type of plot to be
    drawn, see \code{\link{plot}}.}
  \item{\dots}{further graphical parameters to be used for plotting.}
}
\details{
  Zipf's law (e.g., \url{https://en.wikipedia.org/wiki/Zipf\%27s_law})
  states that given some corpus of natural language utterances, the
  frequency of any word is inversely proportional to its rank in the
  frequency table, or, more generally, that the pmf of the term
  frequencies is of the form \eqn{c k^{-\beta}}, where \eqn{k} is the
  rank of the term (taken from the most to the least frequent one).
  We can conveniently explore the degree to which the law holds by
  plotting the logarithm of the frequency against the logarithm of the
  rank, and inspecting the goodness of fit of a linear model.

  Heaps' law (e.g., \url{https://en.wikipedia.org/wiki/Heaps\%27_law})
  states that the vocabulary size \eqn{V} (i.e., the number of different
  terms employed) grows polynomially with the text size \eqn{T} (the
  total number of terms in the texts), so that \eqn{V = c T^\beta}.
  We can conveniently explore the degree to which the law holds by
  plotting \eqn{\log(V)} against \eqn{\log(T)}, and inspecting the
  goodness of fit of a linear model.
}
\value{
  The coefficients of the fitted linear model.  As a side effect, the
  corresponding plot is produced.
}
\examples{
data("acq")
m <- DocumentTermMatrix(acq)
Zipf_plot(m)
Heaps_plot(m)
}

\name{PCorpus}
\alias{PCorpus}
\title{Permanent Corpora}
\description{
  Create permanent corpora.
}
\usage{
PCorpus(x,
        readerControl = list(reader = reader(x), language = "en"),
        dbControl = list(dbName = "", dbType = "DB1"))
}
\arguments{
  \item{x}{A \code{\link{Source}} object.}
  \item{readerControl}{a named list of control parameters for reading in content
    from \code{x}.
    \describe{
      \item{\code{reader}}{a function capable of reading in and processing the
        format delivered by \code{x}.}
      \item{\code{language}}{a character giving the language (preferably as
        \acronym{IETF} language tags, see \link[NLP]{language} in
        package \CRANpkg{NLP}).
        The default language is assumed to be English (\code{"en"}).}
    }
  }
  \item{dbControl}{a named list of control parameters for the underlying
    database storage provided by package \CRANpkg{filehash}.
    \describe{
      \item{\code{dbName}}{a character giving the filename for the database.}
      \item{\code{dbType}}{a character giving the database format (see
        \code{\link[filehash]{filehashOption}} for possible database formats).}
    }
  }
}
\value{
  An object inheriting from \code{PCorpus} and \code{Corpus}.
}
\details{
  A \emph{permanent corpus} stores documents outside of \R in a database. Since
  multiple \code{PCorpus} \R objects with the same underlying database can
  exist simultaneously in memory, changes in one get propagated to all
  corresponding objects (in contrast to the default \R semantics).
}
\seealso{
  \code{\link[tm]{Corpus}} for basic information on the corpus infrastructure
  employed by package \pkg{tm}.

  \code{\link{VCorpus}} provides an implementation with volatile storage
  semantics.
}
\examples{
txt <- system.file("texts", "txt", package = "tm")
\dontrun{
PCorpus(DirSource(txt),
        dbControl = list(dbName = "pcorpus.db", dbType = "DB1"))}
}

\name{tokenizer}
\alias{Boost_tokenizer}
\alias{MC_tokenizer}
\alias{scan_tokenizer}
\title{Tokenizers}
\description{Tokenize a document or character vector.}
\usage{
Boost_tokenizer(x)
MC_tokenizer(x)
scan_tokenizer(x)
}
\arguments{
  \item{x}{A character vector, or an object that can be coerced to character by
  \code{as.character}.}
}
\value{
  A character vector consisting of tokens obtained by tokenization of \code{x}.
}
\details{
  The quality and correctness of a tokenization algorithm highly depends
  on the context and application scenario. Relevant factors are the
  language of the underlying text and the notions of whitespace (which
  can vary with the used encoding and the language) and punctuation
  marks. Consequently, for superior results you probably need a custom
  tokenization function.

  \describe{
    \item{Boost_tokenizer}{Uses the Boost (\url{https://www.boost.org})
      Tokenizer (via \CRANpkg{Rcpp}).}
    \item{MC_tokenizer}{Implements the functionality of the tokenizer in the
      MC toolkit (\url{https://www.cs.utexas.edu/~dml/software/mc/}).}
    \item{scan_tokenizer}{Simulates \code{scan(\dots, what = "character")}.}
  }
}
\seealso{
  \code{\link{getTokenizers}} to list tokenizers provided by package \pkg{tm}.

  \code{\link[NLP]{Regexp_Tokenizer}} for tokenizers using regular expressions
    provided by package \CRANpkg{NLP}.

  \code{\link[tau]{tokenize}} for a simple regular expression based tokenizer
    provided by package \CRANpkg{tau}.

  \code{\link[tokenizers]{tokenizers}} for a collection of tokenizers provided
    by package \CRANpkg{tokenizers}.
}
\examples{
data("crude")
Boost_tokenizer(crude[[1]])
MC_tokenizer(crude[[1]])
scan_tokenizer(crude[[1]])
strsplit_space_tokenizer <- function(x)
    unlist(strsplit(as.character(x), "[[:space:]]+"))
strsplit_space_tokenizer(crude[[1]])
}

\name{hpc}
\alias{tm_parLapply}
\alias{tm_parLapply_engine}
\title{Parallelized \sQuote{lapply}}
\description{
  Parallelize applying a function over a list or vector according to the
  registered parallelization engine.
}
\usage{
tm_parLapply(X, FUN, ...)
tm_parLapply_engine(new)
}
\arguments{
  \item{X}{A vector (atomic or list), or other objects suitable for the
    engine in use.}
  \item{FUN}{the function to be applied to each element of \code{X}.}
  \item{...}{optional arguments to \code{FUN}.}
  \item{new}{an object inheriting from class \code{cluster} as created
    by \code{\link[parallel]{makeCluster}()} from package
    \pkg{parallel}, or a function with formals \code{X}, \code{FUN} and
    \code{...}, or \code{NULL} corresponding to the default of using no
    parallelization engine.}
}
\details{
  Parallelization can be employed to speed up some of the embarrassingly
  parallel computations performed in package \pkg{tm}, specifically
  \code{\link{tm_index}()}, \code{\link{tm_map}()} on a non-lazy-mapped
  \code{\link{VCorpus}}, and \code{\link{TermDocumentMatrix}()} on a
  \code{\link{VCorpus}} or \code{\link{PCorpus}}.

  Functions \code{tm_parLapply()} and \code{tm_parLapply_engine()} can
  be used to customize parallelization according to the available
  resources.

  \code{tm_parLapply_engine()} is used for getting (with no arguments)
  or setting (with argument \code{new}) the parallelization engine
  employed (see below for examples).

  If an engine is set to an object inheriting from class \code{cluster},
  \code{tm_parLapply()} calls
  \code{\link[parallel]{parLapply}()} with this cluster and 
  the given arguments.  If set to a function, \code{tm_parLapply()}
  calls the function with the given arguments.  Otherwise, it simply
  calls \code{\link{lapply}()}.

  Hence, parallelization via
  \code{\link[parallel]{parLapply}()}
  and a default cluster registered via
  \code{\link[parallel]{setDefaultCluster}()} can be
  achieved via \preformatted{  tm_parLapply_engine(function(X, FUN, ...)
      parallel::parLapply(NULL, X, FUN, ...))}
  or re-registering the cluster, say \code{cl}, using \preformatted{  tm_parLapply_engine(cl)}
  (note that since \R version 3.5.0, one can use
  \code{\link[parallel]{getDefaultCluster}()} to get
  the registered default cluster).  Using \preformatted{  tm_parLapply_engine(function(X, FUN, ...)
      parallel::parLapplyLB(NULL, X, FUN, ...))}
  or \preformatted{  tm_parLapply_engine(function(X, FUN, ...)
      parallel::parLapplyLB(cl, X, FUN, ...))}
  gives load-balancing parallelization with the registered default or
  given cluster, respectively.  To achieve parallelization via forking
  (on Unix-alike platforms), one can use the above with clusters created
  by \code{\link[parallel]{makeForkCluster}()}, or use \preformatted{  tm_parLapply_engine(parallel::mclapply)}
  or \preformatted{  tm_parLapply_engine(function(X, FUN, ...)
      parallel::mclapply(X, FUN, ..., mc.cores = n))}
  to use \code{\link[parallel]{mclapply}()} with the default or
  given number \code{n} of cores.
}
\value{
  A list the length of \code{X}, with the result of applying \code{FUN}
  together with the \code{...} arguments to each element of \code{X}.
}
\seealso{
  \code{\link[parallel]{makeCluster}()},
  \code{\link[parallel]{parLapply}()},
  \code{\link[parallel]{parLapplyLB}()}, and
  \code{\link[parallel]{mclapply}()}.
}

\name{weightBin}
\alias{weightBin}
\title{Weight Binary}
\description{
  Binary weight a term-document matrix.
}
\usage{
weightBin(m)
}
\arguments{
  \item{m}{A \code{\link{TermDocumentMatrix}} in term frequency format.}
}
\details{
  Formally this function is of class \code{WeightingFunction} with the
  additional attributes \code{name} and \code{acronym}.
}
\value{
  The weighted matrix.
}

\name{readReut21578XML}
\alias{readReut21578XML}
\alias{readReut21578XMLasPlain}
\title{Read In a Reuters-21578 XML Document}
\description{
  Read in a Reuters-21578 \acronym{XML} document.
}
\usage{
readReut21578XML(elem, language, id)
readReut21578XMLasPlain(elem, language, id)
}
\arguments{
  \item{elem}{a named list with the component \code{content} which must hold
    the document to be read in.}
  \item{language}{a string giving the language.}
  \item{id}{Not used.}
}
\value{
   An \code{\link{XMLTextDocument}} for \code{readReut21578XML}, or a
   \code{\link{PlainTextDocument}} for \code{readReut21578XMLasPlain},
   representing the text and metadata extracted from \code{elem$content}.
}
\references{
  %% Emms, Martin and Luz, Saturnino (2007).
  %% Machine Learning for Natural Language Processing.
  %% \emph{European Summer School of Logic, Language and Information,
  %%   course reader}.
  %% \url{http://ronaldo.cs.tcd.ie/esslli07/mlfornlp.pdf}
  %% \url{http://www.homepages.ed.ac.uk/sluzfil/esslli07/mlfornlp.pdf}

  \bibshow{Lewis:1987}

  %% Luz, Saturnino \emph{\acronym{XML}-encoded version of Reuters-21578}.
  %% \url{http://ronaldo.cs.tcd.ie/esslli07/data/reuters21578-xml/}
  %% \url{http://www.homepages.ed.ac.uk/sluzfil/esslli07/data/reuters21578-xml.tar.bz2}
}
\seealso{
  \code{\link{Reader}} for basic information on the reader infrastructure
  employed by package \pkg{tm}.
}

\name{stemDocument}
\alias{stemDocument}
\alias{stemDocument.character}
\alias{stemDocument.PlainTextDocument}
\title{Stem Words}
\description{
  Stem words in a text document using Porter's stemming algorithm.
}
\usage{
\method{stemDocument}{character}(x, language = "english")
\method{stemDocument}{PlainTextDocument}(x, language = meta(x, "language"))
}
\arguments{
  \item{x}{A character vector or text document.}
  \item{language}{A string giving the language for stemming.}
}
\details{
  Stemming requires that package \CRANpkg{SnowballC} is installed.
  The argument \code{language} is passed over to
  \code{\link[SnowballC]{wordStem}} as the name of the Snowball stemmer.
}
\examples{
data("crude")
inspect(crude[[1]])
if(requireNamespace("SnowballC")) {
    inspect(stemDocument(crude[[1]]))
}
}

\name{inspect}
\alias{inspect}
\alias{inspect.PCorpus}
\alias{inspect.TermDocumentMatrix}
\alias{inspect.TextDocument}
\alias{inspect.VCorpus}
\title{Inspect Objects}
\description{
  Inspect, i.e., display detailed information on a corpus, a
  term-document matrix, or a text document.
}
\usage{
\method{inspect}{PCorpus}(x)
\method{inspect}{VCorpus}(x)
\method{inspect}{TermDocumentMatrix}(x)
\method{inspect}{TextDocument}(x)
}
\arguments{
  \item{x}{Either a corpus, a term-document matrix, or a text document.}
}
\examples{
data("crude")
inspect(crude[1:3])
inspect(crude[[1]])
tdm <- TermDocumentMatrix(crude)[1:10, 1:10]
inspect(tdm)
}

\name{stopwords}
\alias{stopwords}
\title{Stopwords}
\description{
  Return various kinds of stopwords with support for different
  languages.
}
\usage{
stopwords(kind = "en")
}
\arguments{
  \item{kind}{A character string identifying the desired stopword list.}
}
\details{
  Available stopword lists are:
  \describe{
    \item{\code{catalan}}{Catalan stopwords (obtained from
      \url{http://latel.upf.edu/morgana/altres/pub/ca_stop.htm}),}
    \item{\code{romanian}}{Romanian stopwords (extracted from
      \url{http://snowball.tartarus.org/otherapps/romanian/romanian1.tgz}),}
    \item{\code{SMART}}{English stopwords from the SMART information
      retrieval system (as documented in Appendix 11 of
      \url{https://jmlr.csail.mit.edu/papers/volume5/lewis04a/})
      (which coincides with the stopword list used by the MC toolkit
      (\url{https://www.cs.utexas.edu/~dml/software/mc/})),}
  }
  and a set of stopword lists from the Snowball stemmer project in different
  languages (obtained from
  \samp{http://svn.tartarus.org/snowball/trunk/website/algorithms/*/stop.txt}).
  Supported languages are \code{danish}, \code{dutch}, \code{english},
  \code{finnish}, \code{french}, \code{german}, \code{hungarian}, \code{italian},
  \code{norwegian}, \code{portuguese}, \code{russian}, \code{spanish}, and
  \code{swedish}. Language names are case sensitive. Alternatively, their
  \acronym{IETF} language tags may be used.
%
%  Earlier \pkg{tm} versions (before 2013-06-14) used merged stopword lists from
%  \url{http://www.ranks.nl/resources/stopwords.html} and the Snowball stemmer
%  project.
}
\value{A character vector containing the requested stopwords. An error
  is raised if no stopwords are available for the requested
  \code{kind}.}
\examples{
stopwords("en")
stopwords("SMART")
stopwords("german")
}
\keyword{file}

\name{XMLSource}
\alias{XMLSource}
\title{XML Source}
\description{
  Create an \acronym{XML} source.
}
\usage{
XMLSource(x, parser = xml_contents, reader)
}
\arguments{
  \item{x}{a character giving a uniform resource identifier.}
  \item{parser}{a function accepting an \acronym{XML} document (as delivered by
    \code{\link[xml2]{read_xml}} in package \CRANpkg{xml2}) as input and returning
    \acronym{XML} elements/nodes.}
  \item{reader}{a function capable of turning \acronym{XML} elements/nodes as
    returned by \code{parser} into a subclass of \code{\link{TextDocument}}.}
}
\value{
  An object inheriting from \code{XMLSource}, \code{\link{SimpleSource}},
  and \code{\link{Source}}.
}
\seealso{
  \code{\link{Source}} for basic information on the source infrastructure
  employed by package \pkg{tm}.

  Vignette 'Extensions: How to Handle Custom File Formats', and
  \code{\link{readXML}}.
}

\name{readXML}
\alias{readXML}
\title{Read In an XML Document}
\description{
  Return a function which reads in an \acronym{XML} document. The structure of
  the \acronym{XML} document is described with a specification.
}
\usage{
readXML(spec, doc)
}
\arguments{
  \item{spec}{A named list of lists each containing two components. The
    constructed reader will map each list entry to the content or metadatum of
    the text document as specified by the named list entry. Valid names include
    \code{content} to access the document's content, and character strings which
    are mapped to metadata entries.

    Each list entry must consist of two components: the first must be a string
    describing the type of the second argument, and the second is the
    specification entry. Valid combinations are:
    \describe{
      \item{\code{type = "node", spec = "XPathExpression"}}{The XPath (1.0)
        expression \code{spec} extracts information from an \acronym{XML} node.}
      \item{\code{type = "function", spec = function(doc) \dots}}{The function
        \code{spec} is called, passing over the \acronym{XML} document (as
        delivered by \code{\link[xml2]{read_xml}} from package \CRANpkg{xml2}) as
        first argument.}
      \item{\code{type = "unevaluated", spec = "String"}}{The character vector
        \code{spec} is returned without modification.}
    }
  }
  \item{doc}{An (empty) document of some subclass of \code{TextDocument}.}
}
\details{
  Formally this function is a function generator, i.e., it returns a
  function (which reads in a text document) with a well-defined
  signature, but can access passed over arguments (e.g., the specification)
  via lexical scoping.
}
\value{
  A function with the following formals:
  \describe{
    \item{\code{elem}}{a named list with the component \code{content} which
      must hold the document to be read in.}
    \item{\code{language}}{a string giving the language.}
    \item{\code{id}}{a character giving a unique identifier for the created
      text document.}
  }
  The function returns \code{doc} augmented by the parsed information
  as described by \code{spec} out of the \acronym{XML} file in
  \code{elem$content}. The arguments \code{language} and \code{id} are used as
  fallback: \code{language} if no corresponding metadata entry is found in
  \code{elem$content}, and \code{id} if no corresponding metadata entry is found
  in \code{elem$content} and if \code{elem$uri} is null.
}
\seealso{
  \code{\link{Reader}} for basic information on the reader infrastructure
  employed by package \pkg{tm}.

  Vignette 'Extensions: How to Handle Custom File Formats', and
  \code{\link{XMLSource}}.
}

\name{findAssocs}
\alias{findAssocs}
\alias{findAssocs.DocumentTermMatrix}
\alias{findAssocs.TermDocumentMatrix}
\title{Find Associations in a Term-Document Matrix}
\description{
  Find associations in a document-term or term-document matrix.
}
\usage{
\method{findAssocs}{DocumentTermMatrix}(x, terms, corlimit)
\method{findAssocs}{TermDocumentMatrix}(x, terms, corlimit)
}
\arguments{
  \item{x}{A \code{\link{DocumentTermMatrix}} or a
    \code{\link{TermDocumentMatrix}}.}
  \item{terms}{a character vector holding terms.}
  \item{corlimit}{a numeric vector (of the same length as \code{terms};
    recycled otherwise) for the (inclusive) lower correlation limits of each
    term in the range from zero to one.}
}
\value{A named list. Each list component is named after a term in \code{terms}
and contains a named numeric vector. Each vector holds matching terms from
\code{x} and their rounded correlations satisfying the inclusive lower
correlation limit of \code{corlimit}.}
\examples{
data("crude")
tdm <- TermDocumentMatrix(crude)
findAssocs(tdm, c("oil", "opec", "xyz"), c(0.7, 0.75, 0.1))
}

\name{plot}
\alias{plot.TermDocumentMatrix}
\title{Visualize a Term-Document Matrix}
\description{
  Visualize correlations between terms of a term-document matrix.
}
\usage{
\method{plot}{TermDocumentMatrix}(x,
     terms = sample(Terms(x), 20),
     corThreshold = 0.7,
     weighting = FALSE,
     attrs = list(graph = list(rankdir = "BT"),
                  node = list(shape = "rectangle",
                              fixedsize = FALSE)),
     \dots)
}
\arguments{
  \item{x}{A term-document matrix.}
  \item{terms}{Terms to be plotted. Defaults to 20 randomly chosen terms
    of the term-document matrix.}
  \item{corThreshold}{Do not plot correlations below this
    threshold. Defaults to \code{0.7}.}
  \item{weighting}{Define whether the line width corresponds to the
    correlation.}
  \item{attrs}{Argument passed to the plot method for class
    \code{\link[graph]{graphNEL}}.}
  \item{\dots}{Other arguments passed to the
    \code{\link[graph]{graphNEL}} plot method.}
}
\details{
  Visualization requires that Bioconductor software package
  \pkg{Rgraphviz} is installed.
}
\examples{
\dontrun{data(crude)
tdm <- TermDocumentMatrix(crude,
                          control = list(removePunctuation = TRUE,
                                         removeNumbers = TRUE,
                                         stopwords = TRUE))
plot(tdm, corThreshold = 0.2, weighting = TRUE)}
}

\name{removeSparseTerms}
\alias{removeSparseTerms}
\title{Remove Sparse Terms from a Term-Document Matrix}
\description{
  Remove sparse terms from a document-term or term-document matrix.
}
\usage{
removeSparseTerms(x, sparse)
}
\arguments{
  \item{x}{A \code{\link{DocumentTermMatrix}} or a
    \code{\link{TermDocumentMatrix}}.}
  \item{sparse}{A numeric for the maximal allowed sparsity in the range from
    bigger zero to smaller one.}
}
\value{
  A term-document matrix where those terms from \code{x} are
  removed which have at least a \code{sparse} percentage of empty (i.e.,
  terms occurring 0 times in a document) elements. I.e., the resulting
  matrix contains only terms with a sparse factor of less than
  \code{sparse}.
}
\examples{
data("crude")
tdm <- TermDocumentMatrix(crude)
removeSparseTerms(tdm, 0.2)
}

\name{readDOC}
\alias{readDOC}
\title{Read In a MS Word Document}
\description{
  Return a function which reads in a Microsoft Word document extracting
  its text.
}
\usage{
readDOC(engine = c("antiword", "executable"), AntiwordOptions = "")
}
\arguments{
  \item{engine}{a character string for the preferred \acronym{DOC} extraction
    engine (see \bold{Details}).}
  \item{AntiwordOptions}{Options passed over to \command{antiword} executable.}
}
\details{
  Formally this function is a function generator, i.e., it returns a
  function (which reads in a text document) with a well-defined
  signature, but can access passed over arguments (e.g., options to
  \command{antiword}) via lexical scoping.

  Available \acronym{DOC} extraction engines are as follows.
  \describe{
    \item{\code{"antiword"}}{(default) Antiword utility as provided by the
      function \code{\link[antiword]{antiword}} in package \CRANpkg{antiword}.}
    \item{\code{"executable"}}{command line \command{antiword}
      executable which must be installed and accessible on your system.
      This can convert documents from Microsoft Word version 2, 6, 7,
      97, 2000, 2002 and 2003 to plain text.
      %% As of 2023-02-05, http://www.winfield.demon.nl/ does not work.
      %%   , and is available from \url{http://www.winfield.demon.nl/}.
      The character vector \code{AntiwordOptions} is passed over to the
      executable.}
  }
}
\value{
  A \code{function} with the following formals:
  \describe{
    \item{\code{elem}}{a list with the named component \code{uri} which must
      hold a valid file name.}
    \item{\code{language}}{a string giving the language.}
    \item{\code{id}}{Not used.}
  }
  The function returns a \code{\link{PlainTextDocument}} representing the text
  and metadata extracted from \code{elem$uri}.
}
\seealso{
  \code{\link{Reader}} for basic information on the reader infrastructure
  employed by package \pkg{tm}.
}

\name{DirSource}
\alias{DirSource}
\title{Directory Source}
\description{
  Create a directory source.
}
\usage{
DirSource(directory = ".",
          encoding = "",
          pattern = NULL,
          recursive = FALSE,
          ignore.case = FALSE,
          mode = "text")
}
\arguments{
  \item{directory}{A character vector of full path names; the default
    corresponds to the working directory \code{getwd()}.}
  \item{encoding}{a character string describing the current encoding. It is
    passed to \code{\link{iconv}} to convert the input to UTF-8.}
  \item{pattern}{an optional regular expression. Only file names which match
    the regular expression will be returned.}
  \item{recursive}{logical. Should the listing recurse into directories?}
  \item{ignore.case}{logical. Should pattern-matching be case-insensitive?}
  \item{mode}{a character string specifying if and how files should be read in.
    Available modes are:
    \describe{
      \item{\code{""}}{No read. In this case \code{\link{getElem}} and
        \code{\link{pGetElem}} only deliver \acronym{URI}s.}
      \item{\code{"binary"}}{Files are read in binary raw mode (via
        \code{\link{readBin}}).}
      \item{\code{"text"}}{Files are read as text (via
        \code{\link{readLines}}).}
    }
  }
}
\details{
  A \emph{directory source} acquires a list of files via \code{\link{dir}} and
  interprets each file as a document.
}
\value{
  An object inheriting from \code{DirSource}, \code{\link{SimpleSource}}, and
  \code{\link{Source}}.
}
\seealso{
  \code{\link{Source}} for basic information on the source infrastructure
  employed by package \pkg{tm}.

  \code{\link{Encoding}} and \code{\link{iconv}} on encodings.
}
\examples{
DirSource(system.file("texts", "txt", package = "tm"))
}

\name{tm_map}
\alias{tm_map}
\alias{tm_map.VCorpus}
\alias{tm_map.SimpleCorpus}
\alias{tm_map.PCorpus}
\title{Transformations on Corpora}
\description{
  Interface to apply transformation functions (also denoted as mappings)
  to corpora.
}
\usage{
\method{tm_map}{PCorpus}(x, FUN, \dots)
\method{tm_map}{SimpleCorpus}(x, FUN, \dots)
\method{tm_map}{VCorpus}(x, FUN, \dots, lazy = FALSE)
}
\arguments{
  \item{x}{A corpus.}
  \item{FUN}{a transformation function taking a text document (a character
    vector when \code{x} is a \code{SimpleCorpus}) as input and returning a text
    document (a character vector of the same length as the input vector for
    \code{SimpleCorpus}). The function \code{\link{content_transformer}} can be
    used to create a wrapper to get and set the content of text documents.}
  \item{\dots}{arguments to \code{FUN}.}
  \item{lazy}{a logical. Lazy mappings are mappings which are delayed
    until the content is accessed. It is useful for large corpora if only few
    documents will be accessed. In such a case it avoids the computationally
    expensive application of the mapping to all elements in the corpus.}
}
\value{
  A corpus with \code{FUN} applied to each document in \code{x}. In case
  of lazy mappings only internal flags are set. Access of individual documents
  triggers the execution of the corresponding transformation function.
}
\seealso{
  \code{\link{getTransformations}} for available transformations.
}
\note{
Lazy transformations change \R's standard evaluation semantics.
}
\examples{
data("crude")
## Document access triggers the stemming function
## (i.e., all other documents are not stemmed yet)
if(requireNamespace("SnowballC")) {
    tm_map(crude, stemDocument, lazy = TRUE)[[1]]
}
## Use wrapper to apply character processing function
tm_map(crude, content_transformer(tolower))
## Generate a custom transformation function which takes the heading as new content
headings <- function(x)
    PlainTextDocument(meta(x, "heading"),
                      id = meta(x, "id"),
                      language = meta(x, "language"))
inspect(tm_map(crude, headings))
}

\name{content_transformer}
\alias{content_transformer}
\title{Content Transformers}
\description{
  Create content transformers, i.e., functions which modify the content of an
  \R object.
}
\usage{
content_transformer(FUN)
}
\arguments{
  \item{FUN}{a function.}
}
\value{
  A function with two arguments:
  \describe{
    \item{\code{x}}{an \R object with implemented content getter
      (\code{\link[NLP]{content}}) and setter (\code{\link[NLP]{content<-}})
      functions.}
    \item{\code{\dots}}{arguments passed over to \code{FUN}.}
  }
}
\seealso{
  \code{\link{tm_map}} for an interface to apply transformations to corpora.
}
\examples{
data("crude")
crude[[1]]
(f <- content_transformer(function(x, pattern) gsub(pattern, "", x)))
tm_map(crude, f, "[[:digit:]]+")[[1]]
}

\name{Reader}
\alias{FunctionGenerator}
\alias{Reader}
\alias{getReaders}
\title{Readers}
\description{
  Creating readers.
}
\usage{
getReaders()
}
\details{
  \emph{Readers} are functions for extracting textual content and metadata out
  of elements delivered by a \code{\link{Source}}, and for constructing a
  \code{\link{TextDocument}}. A reader must accept following arguments in
  its signature:
  \describe{
    \item{\code{elem}}{a named list with the components \code{content} and
      \code{uri} (as delivered by a \code{\link{Source}} via
      \code{\link{getElem}} or \code{\link{pGetElem}}).}
    \item{\code{language}}{a character string giving the language.}
    \item{\code{id}}{a character giving a unique identifier for the created text
      document.}
  }
  The element \code{elem} is typically provided by a source whereas the language
  and the identifier are normally provided by a corpus constructor (for the case
  that \code{elem$content} does not give information on these two essential
  items).

  In case a reader expects configuration arguments we can use a function
  generator. A function generator is indicated by inheriting from class
  \code{FunctionGenerator} and \code{function}. It allows us to process
  additional arguments, store them in an environment, return a reader function
  with the well-defined signature described above, and still be able to access
  the additional arguments via lexical scoping. All corpus constructors in
  package \pkg{tm} check the reader function for being a function generator and
  if so apply it to yield the reader with the expected signature.
}
\value{
  For \code{getReaders()}, a character vector with readers provided by package
  \pkg{tm}.
}
\seealso{
  \code{\link{readDOC}}, \code{\link{readPDF}}, \code{\link{readPlain}},
  \code{\link{readRCV1}}, \code{\link{readRCV1asPlain}},
  \code{\link{readReut21578XML}}, \code{\link{readReut21578XMLasPlain}},
  and \code{\link{readXML}}.
}

\name{readRCV1}
\alias{readRCV1}
\alias{readRCV1asPlain}
\title{Read In a Reuters Corpus Volume 1 Document}
\description{
  Read in a Reuters Corpus Volume 1 \acronym{XML} document.
}
\usage{
readRCV1(elem, language, id)
readRCV1asPlain(elem, language, id)
}
\arguments{
  \item{elem}{a named list with the component \code{content} which must hold
    the document to be read in.}
  \item{language}{a string giving the language.}
  \item{id}{Not used.}
}
\value{
  An \code{\link{XMLTextDocument}} for \code{readRCV1}, or a
  \code{\link{PlainTextDocument}} for \code{readRCV1asPlain}, representing the
  text and metadata extracted from \code{elem$content}.
}
\seealso{
  \code{\link{Reader}} for basic information on the reader infrastructure
  employed by package \pkg{tm}.
}
\references{
  \bibshow{Lewis+Yang+Rose:2004}
}
\examples{
f <- system.file("texts", "rcv1_2330.xml", package = "tm")
f_bin <- readBin(f, raw(), file.size(f))
rcv1 <- readRCV1(elem = list(content = f_bin), language = "en", id = "id1")
content(rcv1)
meta(rcv1)
}

\name{tm_combine}
\alias{c.VCorpus}
\alias{c.TextDocument}
\alias{c.TermDocumentMatrix}
\alias{c.term_frequency}
\title{Combine Corpora, Documents, Term-Document Matrices, and Term Frequency Vectors}
\description{
  Combine several corpora into a single one, combine multiple
  documents into a corpus, combine multiple term-document matrices
  into a single one, or combine multiple term frequency vectors into a
  single term-document matrix.
}
\usage{
\method{c}{VCorpus}(\dots, recursive = FALSE)
\method{c}{TextDocument}(\dots, recursive = FALSE)
\method{c}{TermDocumentMatrix}(\dots, recursive = FALSE)
\method{c}{term_frequency}(\dots, recursive = FALSE)
}
\arguments{
  \item{\dots}{Corpora, text documents, term-document matrices, or term
    frequency vectors.}
  \item{recursive}{Not used.}
}
\seealso{
  \code{\link{VCorpus}}, \code{\link{TextDocument}},
  \code{\link{TermDocumentMatrix}}, and \code{\link{termFreq}}.
}
\examples{
data("acq")
data("crude")
meta(acq, "comment", type = "corpus") <- "Acquisitions"
meta(crude, "comment", type = "corpus") <- "Crude oil"
meta(acq, "acqLabels") <- 1:50
meta(acq, "jointLabels") <- 1:50
meta(crude, "crudeLabels") <- letters[1:20]
meta(crude, "jointLabels") <- 1:20
c(acq, crude)
meta(c(acq, crude), type = "corpus")
meta(c(acq, crude))
c(acq[[30]], crude[[10]])
c(TermDocumentMatrix(acq), TermDocumentMatrix(crude))
}

\name{tm_reduce}
\alias{tm_reduce}
\title{Combine Transformations}
\description{
Fold multiple transformations (mappings) into a single one.
}
\usage{
tm_reduce(x, tmFuns, \dots)
}
\arguments{
  \item{x}{A corpus.}
  \item{tmFuns}{A list of \pkg{tm} transformations.}
  \item{\dots}{Arguments to the individual transformations.}
}
\value{
  A single \pkg{tm} transformation function obtained by folding \code{tmFuns}
  from right to left (via \code{Reduce(\dots, right = TRUE)}).
}
\seealso{
  \code{Reduce} for \R's internal folding/accumulation mechanism, and
  \code{\link{getTransformations}} to list available transformation
  (mapping) functions.
}
\examples{
data(crude)
crude[[1]]
skipWords <- function(x) removeWords(x, c("it", "the"))
funs <- list(stripWhitespace,
             skipWords,
             removePunctuation,
             content_transformer(tolower))
tm_map(crude, FUN = tm_reduce, tmFuns = funs)[[1]]
}

\name{findFreqTerms}
\alias{findFreqTerms}
\title{Find Frequent Terms}
\description{
  Find frequent terms in a document-term or term-document matrix.
}
\usage{
findFreqTerms(x, lowfreq = 0, highfreq = Inf)
}
\arguments{
  \item{x}{A \code{\link{DocumentTermMatrix}} or
    \code{\link{TermDocumentMatrix}}.}
  \item{lowfreq}{A numeric for the lower frequency bound.}
  \item{highfreq}{A numeric for the upper frequency bound.}
}
\value{
  A character vector of terms in \code{x} which occur more or equal often
  than \code{lowfreq} times and less or equal often than \code{highfreq}
  times.
}
\details{This method works for all numeric weightings but is probably
  most meaningful for the standard term frequency (\code{tf}) weighting
  of \code{x}.}
\examples{
data("crude")
tdm <- TermDocumentMatrix(crude)
findFreqTerms(tdm, 2, 3)
}

\name{readDataframe}
\alias{readDataframe}
\title{Read In a Text Document from a Data Frame}
\description{
  Read in a text document from a row in a data frame.
}
\usage{
readDataframe(elem, language, id)
}
\arguments{
  \item{elem}{a named list with the component \code{content} which must hold
    a data frame with rows as the documents to be read in. The names of the
    columns holding the text content and the document identifier must be
    \code{"text"} and \code{"doc_id"}, respectively.}
  \item{language}{a string giving the language.}
  \item{id}{Not used.}
}
\value{
  A \code{\link{PlainTextDocument}} representing \code{elem$content}.
}
\seealso{
  \code{\link{Reader}} for basic information on the reader infrastructure
  employed by package \pkg{tm}.
}
\examples{
docs <- data.frame(doc_id = c("doc_1", "doc_2"),
                   text = c("This is a text.", "This another one."),
                   stringsAsFactors = FALSE)
ds <- DataframeSource(docs)
elem <- getElem(stepNext(ds))
result <- readDataframe(elem, "en", NULL)
inspect(result)
meta(result)
}

\name{termFreq}
\alias{termFreq}
\title{Term Frequency Vector}
\description{
  Generate a term frequency vector from a text document.
}
\usage{
termFreq(doc, control = list())
}
\arguments{
  \item{doc}{An object inheriting from \code{\link{TextDocument}} or a
    character vector.}
  \item{control}{A list of control options which override default
    settings.

    First, following two options are processed.
    \describe{
      \item{\code{tokenize}}{A function tokenizing a \code{\link{TextDocument}}
        into single tokens, a \code{\link[NLP]{Span_Tokenizer}},
        \code{\link[NLP]{Token_Tokenizer}}, or a string matching one of the
        predefined tokenization functions:
        \describe{
          \item{\code{"Boost"}}{for \code{\link{Boost_tokenizer}}, or}
          \item{\code{"MC"}}{for \code{\link{MC_tokenizer}}, or}
          \item{\code{"scan"}}{for \code{\link{scan_tokenizer}}, or}
          \item{\code{"words"}}{for \code{\link[NLP]{words}}.}
        }
        Defaults to \code{\link[NLP]{words}}.}
      \item{\code{tolower}}{Either a logical value indicating whether
        characters should be translated to lower case or a custom function
        converting characters to lower case. Defaults to
        \code{\link{tolower}}.}
    }
    Next, a set of options which are sensitive to the order of
    occurrence in the \code{control} list. Options are processed in the
    same order as specified. User-specified options have precedence over
    the default ordering so that first all user-specified options and
    then all remaining options (with the default settings and in the
    order as listed below) are processed.
    \describe{
      \item{\code{language}}{A character giving the language (preferably as
        \acronym{IETF} language tags, see \link[NLP]{language} in package
        \CRANpkg{NLP}) to be used for \code{stopwords} and \code{stemming} if
        not provided by \code{doc}.}
      \item{\code{removePunctuation}}{A logical value indicating whether
        punctuation characters should be removed from
        \code{doc}, a custom function which performs punctuation
        removal, or a list of arguments for
        \code{\link{removePunctuation}}. Defaults to \code{FALSE}.}
      \item{\code{removeNumbers}}{A logical value indicating whether
        numbers should be removed from \code{doc} or a custom function
        for number removal. Defaults to \code{FALSE}.}
      \item{\code{stopwords}}{Either a Boolean value indicating stopword
        removal using default language specific stopword lists shipped
        with this package, a character vector holding custom
        stopwords, or a custom function for stopword removal. Defaults
        to \code{FALSE}.}
      \item{\code{stemming}}{Either a Boolean value indicating whether tokens
        should be stemmed or a custom stemming function. Defaults to
        \code{FALSE}.}
    }
    Finally, following options are processed in the given order.
    \describe{
      \item{\code{dictionary}}{A character vector to be tabulated
        against. No other terms will be listed in the result. Defaults
        to \code{NULL} which means that all terms in \code{doc} are
        listed.}
      \item{\code{bounds}}{A list with a tag \code{local} whose value
        must be an integer vector of length 2. Terms that appear less
        often in \code{doc} than the lower bound \code{bounds$local[1]}
        or more often than the upper bound \code{bounds$local[2]} are
        discarded. Defaults to \code{list(local = c(1, Inf))} (i.e., every
        token will be used).}
      \item{\code{wordLengths}}{An integer vector of length 2. Words
        shorter than the minimum word length \code{wordLengths[1]} or
        longer than the maximum word length \code{wordLengths[2]} are
        discarded. Defaults to \code{c(3, Inf)}, i.e., a minimum word
        length of 3 characters.}
    }
  }
}
\value{
  A table of class \code{c("term_frequency", "integer")} with term frequencies
  as values and tokens as names.
}
\seealso{
  \code{\link{getTokenizers}}
}
\examples{
data("crude")
termFreq(crude[[14]])
if(requireNamespace("SnowballC")) {
    strsplit_space_tokenizer <- function(x)
        unlist(strsplit(as.character(x), "[[:space:]]+"))
    ctrl <- list(tokenize = strsplit_space_tokenizer,
                 removePunctuation =
                     list(preserve_intra_word_dashes = TRUE),
                 stopwords = c("reuter", "that"),
                 stemming = TRUE,
                 wordLengths = c(4, Inf))
    termFreq(crude[[14]], control = ctrl)
}
}
\keyword{math}

\name{TextDocument}
\alias{TextDocument}
\title{Text Documents}
\description{
  Representing and computing on text documents.
}
\details{
  \emph{Text documents} are documents containing (natural language) text. The
  \pkg{tm} package employs the infrastructure provided by package \CRANpkg{NLP} and
  represents text documents via the virtual S3 class \code{TextDocument}.
  Actual S3 text document classes then extend the virtual base class (such as
  \code{\link{PlainTextDocument}}).

  All extension classes must provide an \code{\link{as.character}}
  method which extracts the natural language text in documents of the
  respective classes in a \dQuote{suitable} (not necessarily structured)
  form, as well as \code{\link[NLP]{content}} and \code{\link{meta}} methods
  for accessing the (possibly raw) document content and metadata.
}
\seealso{
  \code{\link{PlainTextDocument}}, and
  \code{\link{XMLTextDocument}}
  for the text document classes provided by package \pkg{tm}.

  \code{\link[NLP]{TextDocument}} for text documents in package \pkg{NLP}.
}

\name{Docs}
\alias{Docs}
\alias{nDocs}
\alias{nTerms}
\alias{Terms}
\title{Access Document IDs and Terms}
\description{
  Accessing document IDs, terms, and their number of a term-document matrix or
  document-term matrix.
}
\usage{
Docs(x)
nDocs(x)
nTerms(x)
Terms(x)
}
\arguments{
  \item{x}{Either a \code{\link{TermDocumentMatrix}} or
    \code{\link{DocumentTermMatrix}}.}
}
\value{
  For \code{Docs} and \code{Terms}, a character vector with document IDs and
  terms, respectively.

  For \code{nDocs} and \code{nTerms}, an integer with the number of document IDs
  and terms, respectively.
}
\examples{
data("crude")
tdm <- TermDocumentMatrix(crude)[1:10,1:20]
Docs(tdm)
nDocs(tdm)
nTerms(tdm)
Terms(tdm)
}

\name{findMostFreqTerms}
\alias{findMostFreqTerms}
\alias{findMostFreqTerms.term_frequency}
\alias{findMostFreqTerms.DocumentTermMatrix}
\alias{findMostFreqTerms.TermDocumentMatrix}
\title{Find Most Frequent Terms}
\description{
  Find most frequent terms in a document-term or term-document matrix,
  or a vector of term frequencies.
}
\usage{
findMostFreqTerms(x, n = 6L, ...)
\S3method{findMostFreqTerms}{DocumentTermMatrix}(x, n = 6L, INDEX = NULL, ...)
\S3method{findMostFreqTerms}{TermDocumentMatrix}(x, n = 6L, INDEX = NULL, ...)
}
\arguments{
  \item{x}{A \code{\link{DocumentTermMatrix}} or
    \code{\link{TermDocumentMatrix}}, or a vector of term frequencies as
    obtained by \code{\link{termFreq}()}.}
  \item{n}{A single integer giving the maximal number of terms.}
  \item{INDEX}{an object specifying a grouping of documents for rollup,
    or \code{NULL} (default) in which case each document is considered
    individually.}
  \item{...}{arguments to be passed to or from methods.}
}
\value{
  For the document-term or term-document matrix methods, a list with the
  named frequencies of the up to \code{n} most frequent terms occurring
  in each document (group).  Otherwise, a single such vector of most
  frequent terms.
}
\details{
  Only terms with positive frequencies are included in the results.
}
\examples{
data("crude")

## Term frequencies:
tf <- termFreq(crude[[14L]])
findMostFreqTerms(tf)

## Document-term matrices:
dtm <- DocumentTermMatrix(crude)
## Most frequent terms for each document:
findMostFreqTerms(dtm)
## Most frequent terms for the first 10 the second 10 documents,
## respectively:
findMostFreqTerms(dtm, INDEX = rep(1 : 2, each = 10L))
}

\name{Source}
\alias{Source}
\alias{SimpleSource}
\alias{close.SimpleSource}
\alias{eoi}
\alias{eoi.SimpleSource}
\alias{getMeta}
\alias{getMeta.DataframeSource}
\alias{getElem}
\alias{getElem.DataframeSource}
\alias{getElem.DirSource}
\alias{getElem.URISource}
\alias{getElem.VectorSource}
\alias{getElem.XMLSource}
\alias{getSources}
\alias{length.SimpleSource}
\alias{open.SimpleSource}
\alias{pGetElem}
\alias{pGetElem.DataframeSource}
\alias{pGetElem.DirSource}
\alias{pGetElem.URISource}
\alias{pGetElem.VectorSource}
\alias{reader}
\alias{reader.SimpleSource}
\alias{stepNext}
\alias{stepNext.SimpleSource}
\title{Sources}
\description{
  Creating and accessing sources.
}
\usage{
SimpleSource(encoding = "",
             length = 0,
             position = 0,
             reader = readPlain,
             \dots,
             class)
getSources()
\method{close}{SimpleSource}(con, \dots)
\method{eoi}{SimpleSource}(x)
\method{getMeta}{DataframeSource}(x)
\method{getElem}{DataframeSource}(x)
\method{getElem}{DirSource}(x)
\method{getElem}{URISource}(x)
\method{getElem}{VectorSource}(x)
\method{getElem}{XMLSource}(x)
\method{length}{SimpleSource}(x)
\method{open}{SimpleSource}(con, \dots)
\method{pGetElem}{DataframeSource}(x)
\method{pGetElem}{DirSource}(x)
\method{pGetElem}{URISource}(x)
\method{pGetElem}{VectorSource}(x)
\method{reader}{SimpleSource}(x)
\method{stepNext}{SimpleSource}(x)
}
\arguments{
  \item{x}{A \code{Source}.}
  \item{con}{A \code{Source}.}
  \item{encoding}{a character giving the encoding of the elements delivered by
    the source.}
  \item{length}{a non-negative integer denoting the number of elements delivered
    by the source. If the length is unknown in advance set it to \code{0}.}
  \item{position}{a numeric indicating the current position in the source.}
  \item{reader}{a reader function (generator).}
  \item{\dots}{For \code{SimpleSource} tag-value pairs for storing additional
    information; not used otherwise.}
  \item{class}{a character vector giving additional classes to be used for the
    created source.}
}
\details{
  \emph{Sources} abstract input locations, like a directory, a connection, or
  simply an \R vector, in order to acquire content in a uniform way. In packages
  which employ the infrastructure provided by package \pkg{tm}, such sources are
  represented via the virtual S3 class \code{Source}: such packages then provide
  S3 source classes extending the virtual base class (such as
  \code{\link{DirSource}} provided by package \pkg{tm} itself).

  All extension classes must provide implementations for the functions
  \code{close}, \code{eoi}, \code{getElem}, \code{length}, \code{open},
  \code{reader}, and \code{stepNext}. For parallel element access the
  (optional) function \code{pGetElem} must be provided as well. If
  document level metadata is available, the (optional) function \code{getMeta}
  must be implemented.

  The functions \code{open} and \code{close} open and close the source,
  respectively. \code{eoi} indicates end of input. \code{getElem} fetches the
  element at the current position, whereas \code{pGetElem} retrieves all
  elements in parallel at once. The function \code{length} gives the number of
  elements. \code{reader} returns a default reader for processing elements.
  \code{stepNext} increases the position in the source to acquire the next
  element.

  The function \code{SimpleSource} provides a simple reference implementation
  and can be used when creating custom sources.
}
\value{
  For \code{SimpleSource}, an object inheriting from \code{class},
  \code{SimpleSource}, and \code{Source}.

  For \code{getSources}, a character vector with sources provided by package
  \pkg{tm}.

  \code{open} and \code{close} return the opened and closed source,
  respectively.

  For \code{eoi}, a logical indicating if the end of input of the source is
  reached.

  For \code{getElem} a named list with the components \code{content} holding the
  document and \code{uri} giving a uniform resource identifier (e.g., a file
  path or \acronym{URL}; \code{NULL} if not applicable or unavailable). For
  \code{pGetElem} a list of such named lists.

  For \code{length}, an integer for the number of elements.

  For \code{reader}, a function for the default reader.
}
\seealso{
  \code{\link{DataframeSource}}, \code{\link{DirSource}},
  \code{\link{URISource}}, \code{\link{VectorSource}}, and
  \code{\link{XMLSource}}.
}

\name{TermDocumentMatrix}
\alias{TermDocumentMatrix}
\alias{DocumentTermMatrix}
\alias{as.TermDocumentMatrix}
\alias{as.DocumentTermMatrix}
\title{Term-Document Matrix}
\description{
  Constructs or coerces to a term-document matrix or a document-term matrix.
}
\usage{
TermDocumentMatrix(x, control = list())
DocumentTermMatrix(x, control = list())
as.TermDocumentMatrix(x, \dots)
as.DocumentTermMatrix(x, \dots)
}
\arguments{
  \item{x}{for the constructors, a corpus or an \R object from which a
    corpus can be generated via \code{Corpus(VectorSource(x))}; for the
    coercing functions, either a term-document matrix or a document-term
    matrix or a \link[slam]{simple triplet matrix} (package
    \CRANpkg{slam}) or a \link[=termFreq]{term frequency vector}.} 
  \item{control}{a named list of control options. There are local
    options which are evaluated for each document and global options
    which are evaluated once for the constructed matrix. Available local
    options are documented in \code{\link{termFreq}} and are internally
    delegated to a \code{\link{termFreq}} call.

    This is different for a \code{\link{SimpleCorpus}}. In this case all
    options are processed in a fixed order in one pass to improve performance.
    It always uses the Boost (\url{https://www.boost.org}) Tokenizer (via
    \CRANpkg{Rcpp}) and takes no custom functions as option arguments.

    Available global options are:
    \describe{
      \item{\code{bounds}}{A list with a tag \code{global} whose value
        must be an integer vector of length 2. Terms that appear in less
        documents than the lower bound \code{bounds$global[1]} or in
        more documents than the upper bound \code{bounds$global[2]} are
        discarded. Defaults to \code{list(global = c(1, Inf))} (i.e., every
        term will be used).}
      \item{\code{weighting}}{A weighting function capable of handling a
        \code{TermDocumentMatrix}. It defaults to \code{weightTf} for term
        frequency weighting. Available weighting functions shipped with
        the \pkg{tm} package are \code{\link{weightTf}},
        \code{\link{weightTfIdf}}, \code{\link{weightBin}}, and
        \code{\link{weightSMART}}.}
    }}
    \item{\dots}{the additional argument \code{weighting} (typically a
    \code{\link{WeightFunction}}) is allowed when coercing a
    simple triplet matrix to a term-document or document-term matrix.}
}
\value{
  An object of class \code{TermDocumentMatrix} or class
  \code{DocumentTermMatrix} (both inheriting from a
  \link[slam]{simple triplet matrix} in package \CRANpkg{slam})
  containing a sparse term-document matrix or document-term matrix. The
  attribute \code{weighting} contains the weighting applied to the
  matrix.
}
\seealso{
  \code{\link{termFreq}} for available local control options.
}
\examples{
data("crude")
tdm <- TermDocumentMatrix(crude,
                          control = list(removePunctuation = TRUE,
                                         stopwords = TRUE))
dtm <- DocumentTermMatrix(crude,
                          control = list(weighting =
                                         function(x)
                                         weightTfIdf(x, normalize =
                                                     FALSE),
                                         stopwords = TRUE))
inspect(tdm[202:205, 1:5])
inspect(tdm[c("price", "prices", "texas"), c("127", "144", "191", "194")])
inspect(dtm[1:5, 273:276])

if(requireNamespace("SnowballC")) {
    s <- SimpleCorpus(VectorSource(unlist(lapply(crude, as.character))))
    m <- TermDocumentMatrix(s,
                            control = list(removeNumbers = TRUE,
                                           stopwords = TRUE,
                                           stemming = TRUE))
    inspect(m[c("price", "texa"), c("127", "144", "191", "194")])
}
}

\name{readPlain}
\alias{readPlain}
\title{Read In a Text Document}
\description{
  Read in a text document without knowledge about its internal structure and
  possible available metadata.
}
\usage{
readPlain(elem, language, id)
}
\arguments{
  \item{elem}{a named list with the component \code{content} which must hold
    the document to be read in.}
  \item{language}{a string giving the language.}
  \item{id}{a character giving a unique identifier for the created text
    document.}
}
\value{
  A \code{\link{PlainTextDocument}} representing \code{elem$content}. The
  argument \code{id} is used as fallback if \code{elem$uri} is null.
}
\seealso{
  \code{\link{Reader}} for basic information on the reader infrastructure
  employed by package \pkg{tm}.
}
\examples{
docs <- c("This is a text.", "This another one.")
vs <- VectorSource(docs)
elem <- getElem(stepNext(vs))
(result <- readPlain(elem, "en", "id1"))
meta(result)
}

\name{removeWords}
\alias{removeWords}
\alias{removeWords.character}
\alias{removeWords.PlainTextDocument}
\title{Remove Words from a Text Document}
\description{
  Remove words from a text document.
}
\usage{
\method{removeWords}{character}(x, words)
\method{removeWords}{PlainTextDocument}(x, \dots)
}
\arguments{
  \item{x}{A character or text document.}
  \item{words}{A character vector giving the words to be removed.}
  \item{\dots}{passed over argument \code{words}.}
}
\value{
  The character or text document without the specified words.
}
\seealso{
  \code{\link{getTransformations}} to list available transformation (mapping)
  functions.

  \code{\link[tau]{remove_stopwords}} provided by package \CRANpkg{tau}.
}
\examples{
data("crude")
crude[[1]]
removeWords(crude[[1]], stopwords("english"))
}

\name{tm_filter}
\alias{tm_filter}
\alias{tm_filter.PCorpus}
\alias{tm_filter.SimpleCorpus}
\alias{tm_filter.VCorpus}
\alias{tm_index}
\alias{tm_index.PCorpus}
\alias{tm_index.SimpleCorpus}
\alias{tm_index.VCorpus}
\title{Filter and Index Functions on Corpora}
\description{
  Interface to apply filter and index functions to corpora.
}
\usage{
\method{tm_filter}{PCorpus}(x, FUN, \dots)
\method{tm_filter}{SimpleCorpus}(x, FUN, \dots)
\method{tm_filter}{VCorpus}(x, FUN, \dots)
\method{tm_index}{PCorpus}(x, FUN, \dots)
\method{tm_index}{SimpleCorpus}(x, FUN, \dots)
\method{tm_index}{VCorpus}(x, FUN, \dots)
}
\arguments{
  \item{x}{A corpus.}
  \item{FUN}{a filter function taking a text document or a string (if
    \code{x} is a \code{SimpleCorpus}) as input and returning the
    logical value \code{TRUE} or \code{FALSE}.}
  \item{\dots}{arguments to \code{FUN}.}
}
\value{
  \code{tm_filter} returns a corpus containing documents where
  \code{FUN} matches, whereas \code{tm_index} only returns the
  corresponding indices.
}
\examples{
data("crude")
# Full-text search
tm_filter(crude, FUN = function(x) any(grep("co[m]?pany", content(x))))
}

\name{ZipSource}
\alias{ZipSource}
\title{ZIP File Source}
\description{
  Create a ZIP file source.
}
\usage{
ZipSource(zipfile,
          pattern = NULL,
          recursive = FALSE,
          ignore.case = FALSE,
          mode = "text")
}
\arguments{
  \item{zipfile}{A character string with the full path name of a ZIP file.}
  \item{pattern}{an optional regular expression. Only file names in the ZIP
    file which match the regular expression will be returned.}
  \item{recursive}{logical. Should the listing recurse into directories?}
  \item{ignore.case}{logical. Should pattern-matching be case-insensitive?}
  \item{mode}{a character string specifying if and how files should be read in.
    Available modes are:
    \describe{
      \item{\code{""}}{No read. In this case \code{\link{getElem}} and
        \code{\link{pGetElem}} only deliver \acronym{URI}s.}
      \item{\code{"binary"}}{Files are read in binary raw mode (via
        \code{\link{readBin}}).}
      \item{\code{"text"}}{Files are read as text (via
        \code{\link{readLines}}).}
    }
  }
}
\details{
  A \emph{ZIP file source} extracts a compressed ZIP file via
  \code{\link{unzip}} and interprets each file as a document.
}
\value{
  An object inheriting from \code{ZipSource}, \code{\link{SimpleSource}}, and
  \code{\link{Source}}.
}
\seealso{
  \code{\link{Source}} for basic information on the source infrastructure
  employed by package \pkg{tm}.
}
\examples{
zipfile <- tempfile()
files <- Sys.glob(file.path(system.file("texts", "txt", package = "tm"), "*"))
zip(zipfile, files)
zipfile <- paste0(zipfile, ".zip")
Corpus(ZipSource(zipfile, recursive = TRUE))[[1]]
file.remove(zipfile)
}

\name{readTagged}
\alias{readTagged}
\title{Read In a POS-Tagged Word Text Document}
\description{
  Return a function which reads in a text document containing POS-tagged words.
}
\usage{
readTagged(\dots)
}
\arguments{
  \item{\dots}{Arguments passed to \code{\link[NLP]{TaggedTextDocument}}.}
}
\details{
  Formally this function is a function generator, i.e., it returns a
  function (which reads in a text document) with a well-defined
  signature, but can access passed over arguments (\code{\dots})
  via lexical scoping.
}
\value{
  A \code{function} with the following formals:
  \describe{
    \item{\code{elem}}{a named list with the component \code{content} which must
      hold the document to be read in or the component \code{uri} holding a
      connection object or a character string.}
    \item{\code{language}}{a string giving the language.}
    \item{\code{id}}{a character giving a unique identifier for the created
      text document.}
  }
  The function returns a \code{\link[NLP]{TaggedTextDocument}} representing the
  text and metadata extracted from \code{elem$content} or \code{elem$uri}. The
  argument \code{id} is used as fallback if \code{elem$uri} is null.
}
\seealso{
  \code{\link{Reader}} for basic information on the reader infrastructure
  employed by package \pkg{tm}.
}
\examples{
# See http://www.nltk.org/book/ch05.html or file ca01 in the Brown corpus
x <- paste("The/at grand/jj jury/nn commented/vbd on/in a/at number/nn of/in",
           "other/ap topics/nns ,/, among/in them/ppo the/at Atlanta/np and/cc",
           "Fulton/np-tl County/nn-tl purchasing/vbg departments/nns which/wdt",
           "it/pps said/vbd ``/`` are/ber well/ql operated/vbn and/cc follow/vb",
           "generally/rb accepted/vbn practices/nns which/wdt inure/vb to/in the/at",
           "best/jjt interest/nn of/in both/abx governments/nns ''/'' ./.")
vs <- VectorSource(x)
elem <- getElem(stepNext(vs))
(doc <- readTagged()(elem, language = "en", id = "id1"))
tagged_words(doc)
}

\name{DataframeSource}
\alias{DataframeSource}
\title{Data Frame Source}
\description{
  Create a data frame source.
}
\usage{
DataframeSource(x)
}
\arguments{
  \item{x}{A data frame giving the texts and metadata.}
}
\details{
  A \emph{data frame source} interprets each row of the data frame \code{x} as a
  document. The first column must be named \code{"doc_id"} and contain a unique
  string identifier for each document. The second column must be named
  \code{"text"} and contain a UTF-8 encoded string representing the
  document's content. Optional additional columns are used as document level
  metadata.
}
\value{
  An object inheriting from \code{DataframeSource}, \code{\link{SimpleSource}},
  and \code{\link{Source}}.
}
\seealso{
  \code{\link{Source}} for basic information on the source infrastructure
  employed by package \pkg{tm}, and \code{\link{meta}} for types of metadata.

  \code{\link[readtext]{readtext}} for reading in a text in multiple formats
  suitable to be processed by \code{DataframeSource}.
}
\examples{
docs <- data.frame(doc_id = c("doc_1", "doc_2"),
                   text = c("This is a text.", "This another one."),
                   dmeta1 = 1:2, dmeta2 = letters[1:2],
                   stringsAsFactors = FALSE)
(ds <- DataframeSource(docs))
x <- Corpus(ds)
inspect(x)
meta(x)
}

\name{writeCorpus}
\alias{writeCorpus}
\title{Write a Corpus to Disk}
\description{
  Write a plain text representation of a corpus to multiple files on
  disk corresponding to the individual documents in the corpus.
}
\usage{
writeCorpus(x, path = ".", filenames = NULL)
}
\arguments{
  \item{x}{A corpus.}
  \item{path}{A character listing the directory to be written into.}
  \item{filenames}{Either \code{NULL} or a character vector. In case no
    filenames are provided, filenames are automatically generated by
    using the documents' identifiers in \code{x}.}
}
\details{
  The plain text representation of the corpus is obtained by calling
  \code{as.character} on each document.
}
\examples{
data("crude")
\dontrun{writeCorpus(crude, path = ".",
            filenames = paste(seq_along(crude), ".txt", sep = ""))}
}

\name{meta}
\alias{DublinCore}
\alias{DublinCore<-}
\alias{meta}
\alias{meta.PCorpus}
\alias{meta.SimpleCorpus}
\alias{meta.VCorpus}
\alias{meta<-.PCorpus}
\alias{meta<-.SimpleCorpus}
\alias{meta<-.VCorpus}
\alias{meta.PlainTextDocument}
\alias{meta<-.PlainTextDocument}
\alias{meta.XMLTextDocument}
\alias{meta<-.XMLTextDocument}
\title{Metadata Management}
\description{
  Accessing and modifying metadata of text documents and corpora.
}
\usage{
\method{meta}{PCorpus}(x, tag = NULL, type = c("indexed", "corpus", "local"), \dots)
\method{meta}{PCorpus}(x, tag, type = c("indexed", "corpus", "local"), \dots) <- value
\method{meta}{SimpleCorpus}(x, tag = NULL, type = c("indexed", "corpus"), \dots)
\method{meta}{SimpleCorpus}(x, tag, type = c("indexed", "corpus"), \dots) <- value
\method{meta}{VCorpus}(x, tag = NULL, type = c("indexed", "corpus", "local"), \dots)
\method{meta}{VCorpus}(x, tag, type = c("indexed", "corpus", "local"), \dots) <- value
\method{meta}{PlainTextDocument}(x, tag = NULL, \dots)
\method{meta}{PlainTextDocument}(x, tag = NULL, \dots) <- value
\method{meta}{XMLTextDocument}(x, tag = NULL, \dots)
\method{meta}{XMLTextDocument}(x, tag = NULL, \dots) <- value
DublinCore(x, tag = NULL)
DublinCore(x, tag) <- value
}
\arguments{
  \item{x}{For \code{DublinCore} a \code{\link{TextDocument}}, and for
    \code{meta} a \code{\link{TextDocument}} or a \code{\link{Corpus}}.}
  \item{tag}{a character giving the name of a metadatum. No tag corresponds to
    all available metadata.}
  \item{type}{a character specifying the kind of corpus metadata (see
    \bold{Details}).}
  \item{\dots}{Not used.}
  \item{value}{replacement value.}
}
\details{
  A corpus has two types of metadata. \emph{Corpus metadata} (\code{"corpus"})
  contains corpus specific metadata in form of tag-value pairs.
  \emph{Document level metadata} (\code{"indexed"}) contains document specific
  metadata but is stored in the corpus as a data frame. Document level metadata
  is typically used for semantic reasons (e.g., classifications of documents
  form an own entity due to some high-level information like the range of
  possible values) or for performance reasons (single access instead of
  extracting metadata of each document). The latter can be seen as a from of
  indexing, hence the name \code{"indexed"}. \emph{Document metadata}
  (\code{"local"}) are tag-value pairs directly stored locally at the individual
  documents.

  \code{DublinCore} is a convenience wrapper to access and modify the metadata
  of a text document using the Simple Dublin Core schema (supporting the 15
  metadata elements from the Dublin Core Metadata Element Set
  \url{https://dublincore.org/documents/dces/}).
}
\seealso{
  \code{\link[NLP]{meta}} for metadata in package \CRANpkg{NLP}.
}
\references{
  Dublin Core Metadata Initiative.
  \url{https://dublincore.org/}
}
\examples{
data("crude")
meta(crude[[1]])
DublinCore(crude[[1]])
meta(crude[[1]], tag = "topics")
meta(crude[[1]], tag = "comment") <- "A short comment."
meta(crude[[1]], tag = "topics") <- NULL
DublinCore(crude[[1]], tag = "creator") <- "Ano Nymous"
DublinCore(crude[[1]], tag = "format") <- "XML"
DublinCore(crude[[1]])
meta(crude[[1]])
meta(crude)
meta(crude, type = "corpus")
meta(crude, "labels") <- 21:40
meta(crude)
}

\name{readPDF}
\alias{readPDF}
\title{Read In a PDF Document}
\description{
  Return a function which reads in a portable document format (\acronym{PDF})
  document extracting both its text and its metadata.
}
\usage{
readPDF(engine = c("pdftools", "xpdf", "Rpoppler",
                   "ghostscript", "Rcampdf", "custom"),
        control = list(info = NULL, text = NULL))
}
\arguments{
  \item{engine}{a character string for the preferred \acronym{PDF} extraction
    engine (see \bold{Details}).}
  \item{control}{a list of control options for the engine with the named
    components \code{info} and \code{text} (see \bold{Details}).}
}
\details{
  Formally this function is a function generator, i.e., it returns a function
  (which reads in a text document) with a well-defined signature, but can access
  passed over arguments (e.g., the preferred \acronym{PDF} extraction
  \code{engine} and \code{control} options) via lexical scoping.

  Available \acronym{PDF} extraction engines are as follows.
  \describe{
    \item{\code{"pdftools"}}{(default) Poppler \acronym{PDF} rendering library
      as provided by the functions \code{\link[pdftools]{pdf_info}} and
      \code{\link[pdftools]{pdf_text}} in package \CRANpkg{pdftools}.}
    \item{\code{"xpdf"}}{command line \command{pdfinfo} and
      \command{pdftotext} executables which must be installed and accessible on
      your system. Suitable utilities are provided by the Xpdf
      (\url{http://www.xpdfreader.com/}) \acronym{PDF} viewer or by the
      Poppler (\url{https://poppler.freedesktop.org/}) \acronym{PDF} rendering
      library.}
    \item{\code{"Rpoppler"}}{Poppler \acronym{PDF} rendering library as
      provided by the functions \code{\link[Rpoppler]{PDF_info}} and
      \code{\link[Rpoppler]{PDF_text}} in package \CRANpkg{Rpoppler}.}
    \item{\code{"ghostscript"}}{Ghostscript using \file{pdf_info.ps} and
      \file{ps2ascii.ps}.}
    \item{\code{"Rcampdf"}}{Perl CAM::PDF \acronym{PDF} manipulation library
      as provided by the functions \code{pdf_info} and \code{pdf_text}
      in package \pkg{Rcampdf}, available from the repository at
      \url{http://datacube.wu.ac.at}.}
    \item{\code{"custom"}}{custom user-provided extraction engine.}
  }

  Control parameters for engine \code{"xpdf"} are as follows.
  \describe{
    \item{\code{info}}{a character vector specifying options passed over to
      the \command{pdfinfo} executable.}
    \item{\code{text}}{a character vector specifying options passed over to
      the \command{pdftotext} executable.}
  }

  Control parameters for engine \code{"custom"} are as follows.
  \describe{
    \item{\code{info}}{a function extracting metadata from a \acronym{PDF}.
      The function must accept a file path as first argument and must return a
      named list with the components \code{Author} (as character string),
      \code{CreationDate} (of class \code{POSIXlt}), \code{Subject} (as
      character string), \code{Title} (as character string), and \code{Creator}
      (as character string).}
    \item{\code{text}}{a function extracting content from a \acronym{PDF}.
      The function must accept a file path as first argument and must return a
      character vector.}
  }
}
\value{
  A \code{function} with the following formals:
  \describe{
    \item{\code{elem}}{a named list with the component \code{uri} which must
      hold a valid file name.}
    \item{\code{language}}{a string giving the language.}
    \item{\code{id}}{Not used.}
  }
  The function returns a \code{\link{PlainTextDocument}} representing the text
  and metadata extracted from \code{elem$uri}.
}
\seealso{
  \code{\link{Reader}} for basic information on the reader infrastructure
  employed by package \pkg{tm}.
}
\examples{
uri <- paste0("file://",
              system.file(file.path("doc", "tm.pdf"), package = "tm"))
engine <- if(nzchar(system.file(package = "pdftools"))) {
    "pdftools" 
} else {
    "ghostscript"
}
reader <- readPDF(engine)
pdf <- reader(elem = list(uri = uri), language = "en", id = "id1")
cat(content(pdf)[1])
VCorpus(URISource(uri, mode = ""),
        readerControl = list(reader = readPDF(engine = "ghostscript")))
}
\keyword{file}

\name{removeNumbers}
\alias{removeNumbers}
\alias{removeNumbers.character}
\alias{removeNumbers.PlainTextDocument}
\title{Remove Numbers from a Text Document}
\description{
  Remove numbers from a text document.
}
\usage{
\method{removeNumbers}{character}(x, ucp = FALSE, \dots)
\method{removeNumbers}{PlainTextDocument}(x, \dots)
}
\arguments{
  \item{x}{a character vector or text document.}
  \item{ucp}{a logical specifying whether to use Unicode character
    properties for determining digit characters.  If \code{FALSE} 
    (default), characters in the ASCII \code{[:digit:]} class (i.e., the
    decimal digits from 0 to 9) are taken; if \code{TRUE}, the
    characters with Unicode general category \code{Nd} (Decimal_Number).}
  \item{\dots}{arguments to be passed to or from methods;
    in particular, from the \code{PlainTextDocument} method to the
    \code{character} method.}
}
\value{
  The text document without numbers.
}
\seealso{
  \code{\link{getTransformations}} to list available transformation
  (mapping) functions.

  \url{https://unicode.org/reports/tr44/#General_Category_Values}.
}
\examples{
data("crude")
crude[[1]]
removeNumbers(crude[[1]])
}

\name{foreign}
\alias{read_dtm_Blei_et_al}
\alias{read_dtm_MC}
\title{Read Document-Term Matrices}
\description{
  Read document-term matrices stored in special file formats.
}
\usage{
read_dtm_Blei_et_al(file, vocab = NULL)
read_dtm_MC(file, scalingtype = NULL)
}
\arguments{
  \item{file}{a character string with the name of the file to read.}
  \item{vocab}{a character string with the name of a vocabulary file
    (giving the terms, one per line), or \code{NULL}.}
  \item{scalingtype}{a character string specifying the type of scaling
    to be used, or \code{NULL} (default), in which case the scaling will
    be inferred from the names of the files with non-zero entries found
    (see \bold{Details}).}
}
\details{
  \code{read_dtm_Blei_et_al} reads the (List of Lists type sparse
  matrix) format employed by the Latent Dirichlet Allocation and
  Correlated Topic Model C codes by Blei et al
  (\url{http://www.cs.columbia.edu/~blei/}).

  MC is a toolkit for creating vector models from text documents (see 
  \url{https://www.cs.utexas.edu/~dml/software/mc/}).  It employs a
  variant of Compressed Column Storage (CCS) sparse matrix format,
  writing data into several files with suitable names: e.g., a file with
  \file{_dim} appended to the base file name stores the matrix
  dimensions.  The non-zero entries are stored in a file the name of
  which indicates the scaling type used: e.g., \file{_tfx_nz} indicates
  scaling by term frequency (\samp{t}), inverse document frequency
  (\samp{f}) and no normalization (\samp{x}).  See \file{README} in the
  MC sources for more information.

  \code{read_dtm_MC} reads such sparse matrix information with argument
  \code{file} giving the path with the base file name.
}
\value{
  A \link[=DocumentTermMatrix]{document-term matrix}.
}
\seealso{
  \code{\link[slam]{read_stm_MC}} in package \CRANpkg{slam}.
}
\keyword{IO}

\name{weightSMART}
\alias{weightSMART}
\title{SMART Weightings}
\encoding{UTF-8}
\description{
  Weight a term-document matrix according to a combination of weights
  specified in SMART notation.
}
\usage{
weightSMART(m, spec = "nnn", control = list())
}
\arguments{
  \item{m}{A \code{\link{TermDocumentMatrix}} in term frequency format.}
  \item{spec}{a character string consisting of three characters. The first letter
    specifies a term frequency schema, the second a document frequency
    schema, and the third a normalization schema. See \bold{Details} for
    available built-in schemata.}
  \item{control}{a list of control parameters. See \bold{Details}.}
}
\details{
  Formally this function is of class \code{WeightingFunction} with the
  additional attributes \code{name} and \code{acronym}.

  The first letter of \code{spec} specifies a weighting schema for term
  frequencies of \code{m}:

  \describe{
    \item{"n"}{(natural) \eqn{\mathit{tf}_{i,j}} counts the number of occurrences
      \eqn{n_{i,j}} of a term \eqn{t_i} in a document \eqn{d_j}. The
      input term-document matrix \code{m} is assumed to be in this
      standard term frequency format already.}
    \item{"l"}{(logarithm) is defined as \eqn{1 + \log_2(\mathit{tf}_{i,j})}.}
    \item{"a"}{(augmented) is defined as \eqn{0.5 +
      \frac{0.5 * \mathit{tf}_{i,j}}{\max_i(\mathit{tf}_{i,j})}}.}
    \item{"b"}{(boolean) is defined as 1 if \eqn{\mathit{tf}_{i,j} > 0} and 0 otherwise.}
    \item{"L"}{(log average) is defined as \eqn{\frac{1 +
      \log_2(\mathit{tf}_{i,j})}{1+\log_2(\mathrm{ave}_{i\in j}(\mathit{tf}_{i,j}))}}.}
  }

  The second letter of \code{spec} specifies a weighting schema of
  document frequencies for \code{m}:

  \describe{
    \item{"n"}{(no) is defined as 1.}
    \item{"t"}{(idf) is defined as \eqn{\log_2 \frac{N}{\mathit{df}_t}} where
    \eqn{\mathit{df}_t} denotes how often term \eqn{t} occurs in all
      documents.}
    \item{"p"}{(prob idf) is defined as \eqn{\max(0, \log_2(\frac{N - \mathit{df}_t}{\mathit{df}_t}))}.}
  }

  The third letter of \code{spec} specifies a schema for normalization
  of \code{m}:

  \describe{
    \item{"n"}{(none) is defined as 1.}
    \item{"c"}{(cosine) is defined as \eqn{\sqrt{\mathrm{col\_sums}(m ^ 2)}}.}
    \item{"u"}{(pivoted unique) is defined as \eqn{\mathit{slope} *
      \sqrt{\mathrm{col\_sums}(m ^ 2)} + (1 - \mathit{slope}) *
      \mathit{pivot}} where both \code{slope} and \code{pivot} must be set
      via named tags in the \code{control} list.}
    \item{"b"}{(byte size) is defined as
      \eqn{\frac{1}{\mathit{CharLength}^\alpha}}. The parameter
      \eqn{\alpha} must be set via the named tag \code{alpha}
      in the \code{control} list.}
  }

  The final result is defined by multiplication of the chosen term
  frequency component with the chosen document frequency component with
  the chosen normalization component.
}
\value{
  The weighted matrix.
}
\references{
  \bibshow{Manning+Raghavan+Schütze:2008}
}
\examples{
data("crude")
TermDocumentMatrix(crude,
                   control = list(removePunctuation = TRUE,
                                  stopwords = TRUE,
                                  weighting = function(x)
                                  weightSMART(x, spec = "ntc")))
}

\name{VCorpus}
\alias{VCorpus}
\alias{as.VCorpus}
\title{Volatile Corpora}
\description{
  Create volatile corpora.
}
\usage{
VCorpus(x, readerControl = list(reader = reader(x), language = "en"))
as.VCorpus(x)
}
\arguments{
  \item{x}{For \code{VCorpus} a \code{\link{Source}} object, and for
    \code{as.VCorpus} an \R object.}
  \item{readerControl}{a named list of control parameters for reading in content
    from \code{x}.
    \describe{
      \item{\code{reader}}{a function capable of reading in and processing the
        format delivered by \code{x}.}
      \item{\code{language}}{a character giving the language (preferably as
        \acronym{IETF} language tags, see \link[NLP]{language} in
        package \pkg{NLP}).
        The default language is assumed to be English (\code{"en"}).}
    }
  }
}
\value{
  An object inheriting from \code{VCorpus} and \code{Corpus}.
}
\details{
  A \emph{volatile corpus} is fully kept in memory and thus all changes only
  affect the corresponding \R object.
}
\seealso{
  \code{\link{Corpus}} for basic information on the corpus infrastructure
  employed by package \pkg{tm}.

  \code{\link{PCorpus}} provides an implementation with permanent storage
  semantics.
}
\examples{
reut21578 <- system.file("texts", "crude", package = "tm")
VCorpus(DirSource(reut21578, mode = "binary"),
        list(reader = readReut21578XMLasPlain))
}

\name{stripWhitespace}
\alias{stripWhitespace}
\alias{stripWhitespace.PlainTextDocument}
\title{Strip Whitespace from a Text Document}
\description{
  Strip extra whitespace from a text document. Multiple whitespace
  characters are collapsed to a single blank.
}
\usage{
\method{stripWhitespace}{PlainTextDocument}(x, \dots)
}
\arguments{
  \item{x}{A text document.}
  \item{\dots}{Not used.}
}
\value{
  The text document with multiple whitespace characters collapsed to a
  single blank.
}
\seealso{
  \code{\link{getTransformations}} to list available transformation (mapping)
  functions.
}
\examples{
data("crude")
crude[[1]]
stripWhitespace(crude[[1]])
}

\name{XMLTextDocument}
\alias{XMLTextDocument}
\title{XML Text Documents}
\description{
  Create \acronym{XML} text documents.
}
\usage{
XMLTextDocument(x = xml_missing(),
                author = character(0),
                datetimestamp = as.POSIXlt(Sys.time(), tz = "GMT"),
                description = character(0),
                heading = character(0),
                id = character(0),
                language = character(0),
                origin = character(0),
                \dots,
                meta = NULL)
}
\arguments{
  \item{x}{An \code{\link[xml2:read_xml]{XMLDocument}}.}
  \item{author}{a character or an object of class \code{\link{person}} giving
    the author names.}
  \item{datetimestamp}{an object of class \code{\link{POSIXt}} or a character
    string giving the creation date/time information.  If a character string,
    exactly one of the \acronym{ISO} 8601 formats defined by
    \url{https://www.w3.org/TR/NOTE-datetime} should be used.
    See \code{\link[NLP]{parse_ISO_8601_datetime}} in package \CRANpkg{NLP}
    for processing such date/time information.
  }
  \item{description}{a character giving a description.}
  \item{heading}{a character giving the title or a short heading.}
  \item{id}{a character giving a unique identifier.}
  \item{language}{a character giving the language (preferably as \acronym{IETF}
    language tags, see \link[NLP]{language} in package \CRANpkg{NLP}).}
  \item{origin}{a character giving information on the source and origin.}
  \item{\dots}{user-defined document metadata tag-value pairs.}
  \item{meta}{a named list or \code{NULL} (default) giving all metadata. If set
    all other metadata arguments are ignored.}
}
\value{
  An object inheriting from \code{XMLTextDocument} and
  \code{\link{TextDocument}}.
}
\seealso{
  \code{\link{TextDocument}} for basic information on the text document
  infrastructure employed by package \pkg{tm}.
}
\examples{
xml <- system.file("extdata", "order-doc.xml", package = "xml2")
(xtd <- XMLTextDocument(xml2::read_xml(xml),
                        heading = "XML text document",
                        id = xml,
                        language = "en"))
content(xtd)
meta(xtd)
}

\name{URISource}
\alias{URISource}
\title{Uniform Resource Identifier Source}
\description{
  Create a uniform resource identifier source.
}
\usage{
URISource(x, encoding = "", mode = "text")
}
\arguments{
  \item{x}{A character vector of uniform resource identifiers (\acronym{URI}s.}
  \item{encoding}{A character string describing the current encoding. It is
    passed to \code{\link{iconv}} to convert the input to UTF-8.}
  \item{mode}{a character string specifying if and how \acronym{URI}s should be
    read in. Available modes are:
    \describe{
      \item{\code{""}}{No read. In this case \code{\link{getElem}} and
        \code{\link{pGetElem}} only deliver \acronym{URI}s.}
      \item{\code{"binary"}}{\acronym{URI}s are read in binary raw mode (via
        \code{\link{readBin}}).}
      \item{\code{"text"}}{\acronym{URI}s are read as text (via
        \code{\link{readLines}}).}
    }
  }
}
\details{
  A \emph{uniform resource identifier source} interprets each \acronym{URI} as a
  document.
}
\value{
  An object inheriting from \code{URISource}, \code{\link{SimpleSource}},
  and \code{\link{Source}}.
}
\seealso{
  \code{\link{Source}} for basic information on the source infrastructure
  employed by package \pkg{tm}.

  \code{\link{Encoding}} and \code{\link{iconv}} on encodings.
}
\examples{
loremipsum <- system.file("texts", "loremipsum.txt", package = "tm")
ovid <- system.file("texts", "txt", "ovid_1.txt", package = "tm")
us <- URISource(sprintf("file://\%s", c(loremipsum, ovid)))
inspect(VCorpus(us))
}

\name{crude}
\docType{data}
\alias{crude}
\title{20 Exemplary News Articles from the Reuters-21578 Data Set of Topic crude}
\description{
  This data set holds 20 news articles with additional meta information from
  the Reuters-21578 data set. All documents belong to the topic \code{crude}
  dealing with crude oil.
}
\usage{data("crude")}
\format{A \code{\link{VCorpus}} of 20 text documents.}
\source{Reuters-21578 Text Categorization Collection Distribution 1.0
  (\acronym{XML} format).}
\references{
  %% Emms, Martin and Luz, Saturnino (2007).
  %% Machine Learning for Natural Language Processing.
  %% \emph{European Summer School of Logic, Language and Information,
  %%   course reader}.
  %% \url{http://ronaldo.cs.tcd.ie/esslli07/mlfornlp.pdf}
  %% \url{http://www.homepages.ed.ac.uk/sluzfil/esslli07/mlfornlp.pdf}

  \bibshow{Lewis:1987}

  %% Luz, Saturnino \emph{\acronym{XML}-encoded version of Reuters-21578}.
  %% \url{http://ronaldo.cs.tcd.ie/esslli07/data/reuters21578-xml/}
  %% \url{http://www.homepages.ed.ac.uk/sluzfil/esslli07/data/reuters21578-xml.tar.bz2}
}
\examples{
data("crude")
crude
}
\keyword{datasets}

\name{SimpleCorpus}
\alias{SimpleCorpus}
\title{Simple Corpora}
\description{
  Create simple corpora.
}
\usage{
SimpleCorpus(x, control = list(language = "en"))
}
\arguments{
  \item{x}{a \code{\link{DataframeSource}}, \code{\link{DirSource}} or
    \code{\link{VectorSource}}.}
  \item{control}{a named list of control parameters.
    \describe{
      \item{\code{language}}{a character giving the language (preferably as
        \acronym{IETF} language tags, see \link[NLP]{language} in
        package \CRANpkg{NLP}).
        The default language is assumed to be English (\code{"en"}).}
    }
  }
}
\value{
  An object inheriting from \code{SimpleCorpus} and \code{Corpus}.
}
\details{
  A \emph{simple corpus} is fully kept in memory. Compared to a \code{VCorpus},
  it is optimized for the most common usage scenario: importing plain texts from
  files in a directory or directly from a vector in \R, preprocessing and
  transforming the texts, and finally exporting them to a term-document matrix.
  It adheres to the \code{\link{Corpus}} \acronym{API}. However, it takes
  internally various shortcuts to boost performance and minimize memory
  pressure; consequently it operates only under the following contraints:
  \itemize{
    \item{only \code{DataframeSource}, \code{DirSource} and \code{VectorSource}
      are supported,}
    \item{no custom readers, i.e., each document is read in and stored as plain
      text (as a string, i.e., a character vector of length one),}
    \item{transformations applied via \code{\link{tm_map}} must be able to
      process character vectors and return character vectors (of the same
      length),}
    \item{no lazy transformations in \code{\link{tm_map}},}
    \item{no meta data for individual documents (i.e., no \code{"local"} in
      \code{\link{meta}}).}
  }
}
\seealso{
  \code{\link{Corpus}} for basic information on the corpus infrastructure
  employed by package \pkg{tm}.

  \code{\link{VCorpus}} provides an implementation with volatile storage
  semantics, and \code{\link{PCorpus}} provides an implementation with
  permanent storage semantics.
}
\examples{
txt <- system.file("texts", "txt", package = "tm")
(ovid <- SimpleCorpus(DirSource(txt, encoding = "UTF-8"),
                      control = list(language = "lat")))
}

\name{PlainTextDocument}
\alias{PlainTextDocument}
\title{Plain Text Documents}
\description{
  Create plain text documents.
}
\usage{
PlainTextDocument(x = character(0),
                  author = character(0),
                  datetimestamp = as.POSIXlt(Sys.time(), tz = "GMT"),
                  description = character(0),
                  heading = character(0),
                  id = character(0),
                  language = character(0),
                  origin = character(0),
                  \dots,
                  meta = NULL,
                  class = NULL)
}
\arguments{
  \item{x}{A character string giving the plain text content.}
  \item{author}{a character string or an object of class \code{\link{person}} giving
    the author names.}
  \item{datetimestamp}{an object of class \code{\link{POSIXt}} or a character
    string giving the creation date/time information.  If a character string,
    exactly one of the \acronym{ISO} 8601 formats defined by
    \url{https://www.w3.org/TR/NOTE-datetime} should be used.
    See \code{\link[NLP]{parse_ISO_8601_datetime}} in package \CRANpkg{NLP}
    for processing such date/time information.
  }
  \item{description}{a character string giving a description.}
  \item{heading}{a character string giving the title or a short heading.}
  \item{id}{a character string giving a unique identifier.}
  \item{language}{a character string giving the language (preferably as \acronym{IETF}
    language tags, see \link[NLP]{language} in package \CRANpkg{NLP}).}
  \item{origin}{a character string giving information on the source and origin.}
  \item{\dots}{user-defined document metadata tag-value pairs.}
  \item{meta}{a named list or \code{NULL} (default) giving all metadata. If set
    all other metadata arguments are ignored.}
  \item{class}{a character vector or \code{NULL} (default) giving
    additional classes to be used for the created plain text document.}
}
\value{
  An object inheriting from \code{class}, \code{PlainTextDocument} and
  \code{\link{TextDocument}}.
}
\seealso{
  \code{\link{TextDocument}} for basic information on the text document
  infrastructure employed by package \pkg{tm}.
}
\examples{
(ptd <- PlainTextDocument("A simple plain text document",
                          heading = "Plain text document",
                          id = basename(tempfile()),
                          language = "en"))
meta(ptd)
}

\name{stemCompletion}
\alias{stemCompletion}
\title{Complete Stems}
\description{
  Heuristically complete stemmed words.
}
\usage{
stemCompletion(x,
               dictionary,
               type = c("prevalent", "first", "longest",
                        "none", "random", "shortest"))
}
\arguments{
  \item{x}{A character vector of stems to be completed.}
  \item{dictionary}{A \code{\link{Corpus}} or character vector to be searched
    for possible completions.}
  \item{type}{A \code{character} naming the heuristics to be used:
    \describe{
      \item{\code{prevalent}}{Default. Takes the most frequent match as
      completion.}
      \item{\code{first}}{Takes the first found completion.}
      \item{\code{longest}}{Takes the longest completion in terms of
        characters.}
      \item{\code{none}}{Is the identity.}
      \item{\code{random}}{Takes some completion.}
      \item{\code{shortest}}{Takes the shortest completion in terms of
        characters.}
    }
  }
}
\value{
  A character vector with completed words.
}
\examples{
data("crude")
stemCompletion(c("compan", "entit", "suppl"), crude)
}
\references{
  \bibshow{Feinerer:2010}
}

\name{tm_term_score}
\alias{tm_term_score}
\alias{tm_term_score.DocumentTermMatrix}
\alias{tm_term_score.term_frequency}
\alias{tm_term_score.PlainTextDocument}
\alias{tm_term_score.TermDocumentMatrix}
\title{Compute Score for Matching Terms}
\description{
  Compute a score based on the number of matching terms.
}
\usage{
\method{tm_term_score}{DocumentTermMatrix}(x, terms, FUN = row_sums)
\method{tm_term_score}{PlainTextDocument}(x, terms, FUN = function(x) sum(x, na.rm = TRUE))
\method{tm_term_score}{term_frequency}(x, terms, FUN = function(x) sum(x, na.rm = TRUE))
\method{tm_term_score}{TermDocumentMatrix}(x, terms, FUN = col_sums)
}
\arguments{
  \item{x}{Either a \code{\link{PlainTextDocument}}, a term frequency as
    returned by \code{\link{termFreq}}, or a
    \code{\link{TermDocumentMatrix}}.}
  \item{terms}{A character vector of terms to be matched.}
  \item{FUN}{A function computing a score from the number of terms
    matching in \code{x}.}
}
\value{
  A score as computed by \code{FUN} from the number of matching
  \code{terms} in \code{x}.
}
\examples{
data("acq")
tm_term_score(acq[[1]], c("company", "change"))
\dontrun{## Test for positive and negative sentiments
## install.packages("tm.lexicon.GeneralInquirer", repos="http://datacube.wu.ac.at", type="source")
require("tm.lexicon.GeneralInquirer")
sapply(acq[1:10], tm_term_score, terms_in_General_Inquirer_categories("Positiv"))
sapply(acq[1:10], tm_term_score, terms_in_General_Inquirer_categories("Negativ"))
tm_term_score(TermDocumentMatrix(acq[1:10],
                                control = list(removePunctuation = TRUE)),
             terms_in_General_Inquirer_categories("Positiv"))}
}

\name{removePunctuation}
\alias{removePunctuation}
\alias{removePunctuation.character}
\alias{removePunctuation.PlainTextDocument}
\title{Remove Punctuation Marks from a Text Document}
\description{
  Remove punctuation marks from a text document.
}
\usage{
\method{removePunctuation}{character}(x,
                  preserve_intra_word_contractions = FALSE,
                  preserve_intra_word_dashes = FALSE,
                  ucp = FALSE, \dots)
\method{removePunctuation}{PlainTextDocument}(x, \dots)
}
\arguments{
  \item{x}{a character vector or text document.}
  \item{preserve_intra_word_contractions}{a logical specifying whether
    intra-word contractions should be kept.}
  \item{preserve_intra_word_dashes}{a logical specifying whether
    intra-word dashes should be kept.}
  \item{ucp}{a logical specifying whether to use Unicode character
    properties for determining punctuation characters.  If \code{FALSE} 
    (default), characters in the ASCII \code{[:punct:]} class are taken;
    if \code{TRUE}, the characters with Unicode general category
    \code{P} (Punctuation).}
  \item{\dots}{arguments to be passed to or from methods;
    in particular, from the \code{PlainTextDocument} method to the
    \code{character} method.}
}
\value{
  The character or text document \code{x} without punctuation marks
  (besides intra-word contractions (\samp{'}) and intra-word dashes
  (\samp{-}) if \code{preserve_intra_word_contractions} and
  \code{preserve_intra_word_dashes} are set, respectively).
}
\seealso{
  \code{\link{getTransformations}} to list available transformation
  (mapping) functions.

  \code{\link{regex}} shows the class \code{[:punct:]} of punctuation
  characters.

  \url{https://unicode.org/reports/tr44/#General_Category_Values}.
}
\examples{
data("crude")
inspect(crude[[14]])
inspect(removePunctuation(crude[[14]]))
inspect(removePunctuation(crude[[14]],
                          preserve_intra_word_contractions = TRUE,
                          preserve_intra_word_dashes = TRUE))
}

\name{VectorSource}
\alias{VectorSource}
\title{Vector Source}
\description{
  Create a vector source.
}
\usage{
VectorSource(x)
}
\arguments{
  \item{x}{A vector giving the texts.}
}
\details{
  A \emph{vector source} interprets each element of the vector \code{x} as a
  document.
}
\value{
  An object inheriting from \code{VectorSource}, \code{\link{SimpleSource}},
  and \code{\link{Source}}.
}
\seealso{
  \code{\link{Source}} for basic information on the source infrastructure
  employed by package \pkg{tm}.
}
\examples{
docs <- c("This is a text.", "This another one.")
(vs <- VectorSource(docs))
inspect(VCorpus(vs))
}

\name{getTransformations}
\alias{getTransformations}
\title{Transformations}
\description{
  Predefined transformations (mappings) which can be used with
  \code{\link{tm_map}}.
}
\usage{
getTransformations()
}
\value{
  A character vector with transformations provided by package \pkg{tm}.
}
\seealso{
  \code{\link{removeNumbers}}, \code{\link{removePunctuation}},
  \code{\link{removeWords}}, \code{\link{stemDocument}}, and
  \code{\link{stripWhitespace}}.

  \code{\link{content_transformer}} to create custom transformations.
}  
\examples{
getTransformations()
}

\name{WeightFunction}
\alias{WeightFunction}
\title{Weighting Function}
\description{
  Construct a weighting function for term-document matrices.
}
\usage{
WeightFunction(x, name, acronym)
}
\arguments{
  \item{x}{A function which takes a \code{\link{TermDocumentMatrix}}
    with term frequencies as input, weights the elements, and returns
    the weighted matrix.}
  \item{name}{A character naming the weighting function.}
  \item{acronym}{A character giving an acronym for the name of the
    weighting function.}
}
\value{
  An object of class \code{WeightFunction} which extends the class
  \code{function} representing a weighting function.
}
\examples{
weightCutBin <- WeightFunction(function(m, cutoff) m > cutoff,
                               "binary with cutoff", "bincut")
}

\name{weightTfIdf}
\alias{weightTfIdf}
\title{Weight by Term Frequency - Inverse Document Frequency}
\description{
  Weight a term-document matrix by term frequency - inverse document
  frequency.
}
\usage{
weightTfIdf(m, normalize = TRUE)
}
\arguments{
  \item{m}{A \code{\link{TermDocumentMatrix}} in term frequency format.}
  \item{normalize}{A Boolean value indicating whether the term
    frequencies should be normalized.}
}
\details{
  Formally this function is of class \code{WeightingFunction} with the
  additional attributes \code{name} and \code{acronym}.

  \emph{Term frequency} \eqn{\mathit{tf}_{i,j}} counts the number of
  occurrences \eqn{n_{i,j}} of a term \eqn{t_i} in a document
  \eqn{d_j}. In the case of normalization, the term frequency
  \eqn{\mathit{tf}_{i,j}} is divided by \eqn{\sum_k n_{k,j}}.

  \emph{Inverse document frequency} for a term \eqn{t_i} is defined as
  \deqn{\mathit{idf}_i = \log_2 \frac{|D|}{|\{d \mid t_i \in d\}|}} where
  \eqn{|D|} denotes the total number of documents and where \eqn{|\{d
  \mid t_i \in d\}|} is the number of documents where the term \eqn{t_i}
  appears.

  \emph{Term frequency - inverse document frequency} is now defined as
  \eqn{\mathit{tf}_{i,j} \cdot \mathit{idf}_i}.
}
\value{
  The weighted matrix.
}
\references{
  \bibshow{Salton+Buckley:1988}
}

\name{Corpus}
\alias{Corpus}
\title{Corpora}
\description{
  Representing and computing on corpora.
}
\details{
  \emph{Corpora} are collections of documents containing (natural language)
  text. In packages which employ the infrastructure provided by package
  \pkg{tm}, such corpora are represented via the virtual S3 class
  \code{Corpus}: such packages then provide S3 corpus classes extending the
  virtual base class (such as \code{\link{VCorpus}} provided by package \pkg{tm}
  itself).

  All extension classes must provide accessors to extract subsets
  (\code{\link{[}}), individual documents (\code{\link{[[}}), and metadata
  (\code{\link{meta}}). The function \code{\link{length}} must return the number
  of documents, and \code{\link{as.list}} must construct a list holding the
  documents.

  A corpus can have two types of metadata (accessible via \code{\link{meta}}).
  \emph{Corpus metadata} contains corpus specific metadata in form of tag-value
  pairs. \emph{Document level metadata} contains document specific metadata but
  is stored in the corpus as a data frame. Document level metadata is typically
  used for semantic reasons (e.g., classifications of documents form an own
  entity due to some high-level information like the range of possible values)
  or for performance reasons (single access instead of extracting metadata of
  each document).

  The function \code{Corpus} is a convenience alias to \code{SimpleCorpus} or
  \code{VCorpus}, depending on the arguments provided.
}
\seealso{
  \code{\link{SimpleCorpus}}, \code{\link{VCorpus}}, and \code{\link{PCorpus}}
  for the corpora classes provided by package \pkg{tm}.

  \code{\link[tm.plugin.dc]{DCorpus}} for a distributed corpus class provided by
  package \CRANpkg{tm.plugin.dc}.
}

\name{weightTf}
\alias{weightTf}
\title{Weight by Term Frequency}
\description{
  Weight a term-document matrix by term frequency.
}
\usage{
weightTf(m)
}
\arguments{
  \item{m}{A \code{\link{TermDocumentMatrix}} in term frequency format.}
}
\details{
  Formally this function is of class \code{WeightingFunction} with the
  additional attributes \code{name} and \code{acronym}.

  This function acts as the identity function since the input matrix is
  already in term frequency format.
}
\value{
  The weighted matrix.
}
